/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *              INTERFACE FOR THE DFG PARSER              * */
/* *                                                        * */
/* *  $Module:   DFG                                        * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 1.6 $                                         * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 17:46:57 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: dfg.h,v $ */

#ifndef _DFG_
#define _DFG_

#include "misc.h"
#include "list.h"
#include "clause.h"

typedef enum { DFG_SATISFIABLE, DFG_UNSATISFIABLE, DFG_UNKNOWNSTATE } DFG_STATE;

/* Parser functions */
LIST        dfg_DFGParser(FILE* File, LIST* Axioms, LIST* Conjectures,
			  LIST* SortDecl);
LIST        dfg_ProofParser(FILE* File);
LIST        dfg_TermParser(FILE* File);

/* Functions for accessing desription information */
const char* dfg_ProblemName(void);
const char* dfg_ProblemAuthor(void);
const char* dfg_ProblemVersion(void);
const char* dfg_ProblemLogic(void);
DFG_STATE   dfg_ProblemStatus(void);
const char* dfg_ProblemStatusString(void);
const char* dfg_ProblemDescription(void);
const char* dfg_ProblemDate(void);
NAT         dfg_DescriptionLength(void);

void        dfg_Free(void);   /* Must be called after each parser call */
void        dfg_DeleteFormulaPairList(LIST);
void        dfg_StripLabelsFromList(LIST);
void        dfg_FPrintDescription(FILE*);

void        dfg_DeleteProofList(LIST);

CLAUSE      dfg_CreateClauseFromTerm(TERM, BOOL); 

#endif
