/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *              TOP MODULE OF SPASS                       * */
/* *                                                        * */
/* *  $Module:   TOP                                        * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.65 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/07/20 20:56:36 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: dfg2otter.c,v $ */

/*** MAINLOOP *************************************************/


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "dfg.h"
#include "foldfg.h"
#include "clause.h"
#include "options.h"

#include <errno.h>

/**************************************************************/
/* Main Function                                              */
/**************************************************************/

int main(int argc, char* argv[])
{
  LIST      Clauses,AxiomList,ConjectureList, SortList;
  FILE      *Output,*Input;
  char      *Filename;
  FLAGSTORE Store;

  memory_Init(-1);
  symbol_Init(TRUE);
  stack_Init();
  term_Init();
  flag_Init();
  fol_Init(TRUE);
  clause_Init();
  opts_Init();

  Store = flag_NewStore();
  flag_InitStoreByDefaults(Store);
  
  opts_DeclareSPASSFlagsAsOptions();
  if (!opts_Read(argc, argv)) 
    return EINVAL;

  opts_SetFlags(Store);
  opts_Free();
  flag_InstallStore(Store);

  if (flag_GetFlagValue(flag_STDIN) == 0 && argc != opts_Indicator()+2) {
    printf("\n\t          dfg2otter Version 0.58");
    printf("\n\t       Usage: dfg2otter [options] <input-file> <output-fule>\n\n");
    return EINVAL;
  }

  AxiomList = ConjectureList = SortList = list_Nil();

  if (flag_GetFlagValue(flag_STDIN) == 0) {
    Input = misc_OpenFile(argv[opts_Indicator()],"r");
    Clauses = dfg_DFGParser(Input,&AxiomList,&ConjectureList, &SortList);
    misc_CloseFile(Input,argv[opts_Indicator()]);
    AxiomList = list_Nconc(AxiomList, SortList);
    dfg_DeleteFormulaPairList(AxiomList);
    dfg_DeleteFormulaPairList(ConjectureList);
    Filename = argv[opts_Indicator()+1];
    Output = misc_OpenFile(Filename,"w");
    clause_FPrintCnfOtter(Output, Clauses);
    misc_CloseFile(Output,Filename);
  }
  else {
    Clauses = dfg_DFGParser(stdin,&AxiomList,&ConjectureList, &SortList);
    AxiomList = list_Nconc(AxiomList, SortList);
    dfg_DeleteFormulaPairList(AxiomList);
    dfg_DeleteFormulaPairList(ConjectureList);
    clause_FPrintCnfOtter(stdout, Clauses);
  }

  list_DeleteWithElement(Clauses,clause_Delete);
  /*symbol_Dump();*/
  flag_FreeStore(Store);
  flag_Free();
  fol_Free();
  symbol_FreeAllSymbols();
  dfg_Free();
#ifdef DEBUG
  memory_Print();
#endif
  return 0;
}
