/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                 SCANNER FOR DFG SYNTAX                 * */
/* *                                                        * */
/* *  $Module:   DFG                                        * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 1.4 $                                         * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/08 12:19:27 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: dfgscanner.l,v $ */

%{

#include <ctype.h>    /* for isprint */
#include <errno.h>
#include "misc.h"
#include "memory.h"
#include "symbol.h"
#include "term.h"
#include "dfg.h"
#include "dfgparser.h"

extern NAT dfg_LINENUMBER;  /* defined in dfgparser.y */

static NAT dfg_CountNewlines(char*);

static __inline__ char* dfg_StringCopy(void)
{
  char *copy;
  copy = (char*) memory_Malloc(yyleng+1);
  strcpy(copy, yytext);
  return copy;
}

%}

%option noyywrap
/* Force the scanner to read the input character by character */
%option always-interactive

%%

and				return DFG_AND;
author				return DFG_AUTHOR;
axioms				return DFG_AXIOMS;
begin_problem			return DFG_BEGPROB;
box				return DFG_BOX;
clause				return DFG_CLAUSE;
cnf				return DFG_CNF;
comp				return DFG_COMP;
conjectures			return DFG_CONJECS;
conv				return DFG_CONV;
date				return DFG_DATE;
description			return DFG_DESC;
dia				return DFG_DIA;
dnf				return DFG_DNF;
domain				return DFG_DOMAIN;
domrestr			return DFG_DOMRESTR;
eml|EML				return DFG_EML;
end_of_list			return DFG_ENDLIST;
end_problem			return DFG_ENDPROB;
equal				return DFG_EQUAL;
equiv				return DFG_EQUIV;
exists				return DFG_EXISTS;
false				return DFG_FALSE;
forall				return DFG_FORALL;
formula				return DFG_FORMULA;
functions			return DFG_FUNC;
hypothesis			return DFG_HYPOTH;
id  				return DFG_IDENTITY;
div  				return DFG_DIVERSITY;
implied				return DFG_IMPLIED;
implies				return DFG_IMPLIES;
KIV				return DFG_KIV;
LEM				return DFG_LEM;
list_of_clauses			return DFG_CLSLIST;
list_of_declarations		return DFG_DECLLIST;
list_of_descriptions		return DFG_DESCLIST;
list_of_formulae		return DFG_FORMLIST;
list_of_general_settings	return DFG_GENSET;
list_of_proof			return DFG_PRFLIST;
list_of_settings		return DFG_SETTINGS;
list_of_symbols			return DFG_SYMLIST;
list_of_terms                   return DFG_TERMLIST;
logic				return DFG_LOGIC;
name				return DFG_NAME;
not				return DFG_NOT;
operators			return DFG_OPERAT;
or				return DFG_OR;
OTTER				return DFG_OTTER;
predicate			return DFG_PRED;
predicates			return DFG_PRDICAT;
prop_formula			return DFG_PROP_FORMULA;
PROTEIN				return DFG_PROTEIN;
quantifiers			return DFG_QUANTIF;
ranrestr			return DFG_RANRESTR;
range				return DFG_RANGE;
rel_formula			return DFG_REL_FORMULA;
resolution			return DFG_RESOLU;
satisfiable			return DFG_SATIS;
SATURATE			return DFG_SATURATE;
SETHEO				return DFG_SETHEO;
sorts				return DFG_SORTS;
SPASS				return DFG_SPASS;
sum				return DFG_SUM;
test 				return DFG_TEST;
list_of_special_formulae	return DFG_SPECIALFORMLIST;
status				return DFG_STATUS;
subsort				return DFG_SUBSORT;
translpairs			return DFG_TRANSLPAIRS;
true				return DFG_TRUE;
unknown				return DFG_UNKNOWN;
unsatisfiable			return DFG_UNSATIS;
version				return DFG_VERSION;
3TAP				return DFG_3TAP;
"-1"				return DFG_MINUS1;

"%"[^\n]*			/* one-line comment */
\{\*([^\*]|(\*[^\}]))*\*\}	{ dfg_lval.string = dfg_StringCopy();
				  dfg_LINENUMBER += dfg_CountNewlines(yytext);
                                  return DFG_TEXT;
                                }
[0-9]+				{ unsigned long n;
                                  errno = 0;
                                  n = strtoul(yytext, NULL, 10);
				  if (errno != 0 || n > INT_MAX) {
                                    fprintf(stderr, "\nNumber to big in line %d\n",
                                            dfg_LINENUMBER);
                                    misc_Error();
                                  }
                                  dfg_lval.number = (int) n;
                                  return DFG_NUM;
                                }
[a-zA-Z0-9_]+			{ dfg_lval.string = dfg_StringCopy();
                                  return DFG_ID;
                                }
[ \t]+				/* ignore */
"\n"				dfg_LINENUMBER++;
[-()\[\],\.]			return yytext[0];
.				{ fputs("\nIllegal character '", stderr);
				  if (isprint((int)yytext[0]))
				    putc(yytext[0], stderr);
				  else
				    fprintf(stderr, "\\x%x", (unsigned int) yytext[0]);
				  fprintf(stderr, "' in line %d\n", dfg_LINENUMBER);
                                  misc_Error();
                                }

%%

static NAT dfg_CountNewlines(char* Text)
{
  NAT result = 0;

  while (*Text != 0) {
    if (*Text++ == '\n')
      result++;
  }
  return result;
}
