/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *              EXTENDED MODAL LOGICS ROUTINES            * */
/* *                                                        * */
/* *  $Module:   EML                                        * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.4 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/11/30 19:43:23 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: eml.c,v $ */

#include "cnf.h"
#include "eml.h"
#include "foldfg.h"

SYMBOL  eml_APPLYF;
SYMBOL  eml_BOX;
SYMBOL  eml_COMP;
SYMBOL  eml_CONV;
SYMBOL  eml_DIA;
SYMBOL  eml_DIV;
SYMBOL  eml_DOMAIN;
SYMBOL  eml_DOMRESTR;
SYMBOL  eml_ID;
SYMBOL  eml_RANGE;
SYMBOL  eml_RANRESTR;
SYMBOL  eml_SUM;
SYMBOL  eml_TEST;

LIST    eml_SYMBOLS;

LIST *eml_PROPINDEXTOFOSYMBOL;
SYMBOL eml_FLUTEDSYMBOLS[symbol__MAXSIGNATURE][symbol__MAXSIGNATURE];    

static const char* eml_PREDNAMEPREFIX[6] = { "P_", "R_", "Nde_" , "P", "Nde", "S" };

static LIST    eml_PROPSYMBOLLIST;  /* Unused */

static TERM  eml_RemoveTrivialAtomsAndOps(TERM);
static TERM  eml_RemoveTrivialOperator(TERM Term);
static TERM  eml_EmbedInRelCalculusProp(TERM);
static TERM  eml_EmbedInRelCalculusRel(TERM);
static TERM  eml_RelationalTranslationProp(TERM, TERM);
static TERM  eml_RelationalTranslationRel(TERM, TERM, TERM);
static TERM  eml_FunctTranslProp(TERM, TERM);
static TERM  eml_FunctTranslRel(TERM, TERM);
static TERM  eml_FunctTranslIntoFlutedLogicProp(TERM, LIST, char*, int);
static TERM  eml_FunctTranslIntoFlutedLogicRel(TERM, LIST, char*, int);

/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *  FUNCTIONS                                             * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

void eml_Init(void)
/**************************************************************
  INPUT:   None.
  RETURNS: None.
  USAGE:   extern public.
  SUMMARY: Initializes the Eml Module.
  EFFECTS: All additional extended modal logic symbols are created.  
           (The first-order symbols are created in fol_Init.)
	   And, creates an array for the list of
	   propositional/first-order symbol pairs as well as Eml logical
	   symbols/first-order logical symbol lists.
  CAUTION: MUST BE CALLED BEFORE ANY OTHER eml-FUNCTION.
***************************************************************/
{ 

  eml_BOX	= symbol_CreateJunctor("box", 2, symbol_STATLEX);
  eml_COMP	= symbol_CreateJunctor("comp", 2, symbol_STATLEX);
  eml_CONV	= symbol_CreateJunctor("conv", 1, symbol_STATLEX);
  eml_DIA	= symbol_CreateJunctor("dia", 2, symbol_STATLEX);
  eml_DIV	= symbol_CreatePredicate("div", 0, symbol_STATLEX);
  eml_DOMAIN	= symbol_CreateJunctor("domain", 1, symbol_STATLEX);
  eml_DOMRESTR	= symbol_CreateJunctor("domrestr", 2, symbol_STATLEX);
  eml_ID	= symbol_CreatePredicate("id", 0, symbol_STATLEX);
  eml_RANGE	= symbol_CreateJunctor("range", 1, symbol_STATLEX);
  eml_RANRESTR	= symbol_CreateJunctor("ranrestr", 2, symbol_STATLEX);
  eml_SUM	= symbol_CreateJunctor("sum", 2, symbol_STATLEX);
  eml_TEST	= symbol_CreateJunctor("test", 1, symbol_STATLEX);

  eml_SYMBOLS =
  	list_Cons((POINTER)eml_BOX, list_Cons((POINTER)eml_COMP,
  	  list_Cons((POINTER)eml_CONV, list_Cons((POINTER)eml_DIA,
  	    list_Cons((POINTER)eml_DIV, list_Cons((POINTER)eml_DOMAIN, 
              list_Cons((POINTER)eml_DOMRESTR, list_Cons((POINTER)eml_ID, 
		list_Cons((POINTER)eml_RANGE, list_Cons((POINTER)eml_RANRESTR,
  	          list_Cons((POINTER)eml_SUM, list_List((POINTER)eml_TEST))))))))))));

  eml_PROPINDEXTOFOSYMBOL = (LIST*)memory_Malloc(sizeof(LIST_NODE) *
		  symbol__MAXSIGNATURE);

  eml_SetPropFoSymbolAssocList(eml_Dia(),
		list_Cons((POINTER)fol_Exist(), list_List((POINTER)fol_And())));
  eml_SetPropFoSymbolAssocList(eml_Box(), 
		list_Cons((POINTER)fol_All(), list_List((POINTER)fol_Implies())));
  eml_SetPropFoSymbolAssocList(eml_Comp(), 
		list_Cons((POINTER)fol_Exist(), list_List((POINTER)fol_And())));
  eml_SetPropFoSymbolAssocList(eml_Sum(), 
		list_Cons((POINTER)fol_All(), list_List((POINTER)fol_Or())));
  eml_SetPropFoSymbolAssocList(eml_Domain(), 
		list_List((POINTER)fol_Exist()));
  eml_SetPropFoSymbolAssocList(eml_Range(), 
		list_List((POINTER)fol_Exist()));
  eml_SetPropFoSymbolAssocList(eml_Id(), 
		list_List((POINTER)fol_Equality()));
}

void eml_InitFunctTransl(void)
/**************************************************************
  INPUT:   None.
  RETURNS: None.
  USAGE:   static intern.
  SUMMARY: Initializes the routines for the functional translation of
           basic (serial) modal logics into fluted logic.
  EFFECTS: Additonal flags are set.
  CAUTION: Prerequisite for the functional translation into fluted logic
           routines.
***************************************************************/
{ 
	/* printf("\nDoing eml_InitFunctTransl()\n"); */

  /* Dynamic memory allocation of eml_FLUTEDSYMBOLS??
   * This does not work:
  eml_FLUTEDSYMBOLS = (SYMBOL **)memory_Malloc(sizeof(SYMBOL) * symbol__MAXSIGNATURE *
		  symbol__MAXSIGNATURE);
		  */
  if (!flag_GetFlagValue(flag_EMLFFLUTED)) {
    eml_APPLYF	= symbol_CreateFunction("app", 2, symbol_STATLEX);
    fol_SYMBOLS = list_Cons((POINTER)eml_APPLYF, fol_SYMBOLS);
  }

}

void eml_Free(void)
/**************************************************************
  INPUT:   None.
  RETURNS: void.
  EFFECT:  The eml related memory used by the eml module is freed.
***************************************************************/
{
  list_Delete(eml_SYMBOLS);
  /* list_Delete(eml_PROPINDEXTOFOSYMBOL); ?? */

  /* list_Delete(eml_FLUTEDSYMBOLS); ?? */
}

LIST eml_NullaryPredList()
/**************************************************************
  INPUT:   None.
  RETURNS: The list of nullary predicate symbols in the signature
  USAGE:   not used.
  EFFECTS: Creates a table for the list of
           propositional/first-order symbol pairs.
***************************************************************/
{
  LIST   nullaryPredList, Scan;
  int 	 i;

  if (!symbol_HASSIGNATURE) {
    fprintf(stderr,"\n\t eml_NullaryPredList(): no signature!?\n");
    misc_Error();
  }

  nullaryPredList = list_Nil();
  for(i = 1; i < symbol_ACTINDEX; i++)
    if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)i) &&
	  eml_IsNullaryPredicate(symbol_IndexPredSymbol(i))) 
      nullaryPredList =
	      list_Cons((POINTER)(symbol_IndexPredSymbol(i)),
			      nullaryPredList);

#ifdef DEBUG
  printf("\n nullaryPredList: ");
    for (Scan=nullaryPredList; !list_Empty(Scan); Scan=list_Cdr(Scan)) {
      symbol_Print((SYMBOL)list_Car(Scan));
      printf(" ");
    }
  printf("\n");
#endif
  
  return nullaryPredList;
}

SYMBOL eml_CreateFoSymbol(SYMBOL PropSymbol, eml_SYMBOLTYPE SymbolType,
		int Arity, int Status)
/*******************************************************
  INPUT:   A symbol (which MUST be a propositonal symbol/nullary
	   predicate symbol), a symbol type (of the fo symbol to be
	   created), the arity of the new symbol, The status of the new
	   symbol.
  RETURNS: A first-order symbol corresponding to PropSymbol.
********************************************************/
{
  SYMBOL foSymbol;
  char   newName[10];

  if (eml_FlutedType(SymbolType)) {
    /* Note: if SymbolType in fluted then any propositional symbol can
     * be associated with more than one n-ary predicate symbol, for
     * that reason no user defined FoSymbols will be considered */
    sprintf(newName,"%s%d%s", eml_PREDNAMEPREFIX[SymbolType], Arity,
	    symbol_Name(PropSymbol));
    foSymbol = symbol_CreatePredicate(newName, Arity, Status);
    eml_SetPropFlutedSymbolAssocList(PropSymbol,Arity,foSymbol);
  }
  else {
    sprintf(newName,"%s%s", eml_PREDNAMEPREFIX[SymbolType],
	    symbol_Name(PropSymbol));
    foSymbol = symbol_CreatePredicate(newName, Arity, Status);
    eml_SetPropFoSymbolAssocList(PropSymbol,list_List((POINTER)foSymbol));
  }
  
  return foSymbol;
}

TERM eml_CreateQuantifier(TERM Term, SYMBOL Quantifier, LIST VarList,
		LIST Arguments)
/*******************************************************
  INPUT:   A term (dummy, not used), symbol (which MUST be a fol
	   quantifier), a list of variables that will be bound, and
           a list of arguments.
  RETURNS: A new quantified term.
********************************************************/
{                                                                  
  return term_Create(Quantifier,                                             
		     list_Cons(term_Create(fol_Varlist(), VarList),   
			       Arguments));  
}

static TERM  eml_RplacWithQuantifier(TERM Term, SYMBOL Quantifier, 
		LIST VarList, LIST Arguments)
/**************************************************************
  INPUT:   A formula, a symbol (quantifier), two list of terms.
  RETURNS: A quantifier replacement.
  EFFECT:  Note, a term node is created for the variable list.
  CAUTION: The term in destructively changed.
***************************************************************/
{
  term_RplacTop(Term,Quantifier);
  term_RplacArgumentList(Term, 
	list_Cons(term_Create(fol_Varlist(), VarList), Arguments));

  return Term;
}

static TERM eml_BuildSortQuantifierFor(TERM Term, SYMBOL ModalTop,
		SYMBOL RelSubTop, TERM Var, BOOL WithSort,
		LIST Arguments, POINTER (*BuildFunction)())
/*******************************************************
  INPUT:   A term (which MUST be a modal term, i.e. dia(..) or box (...)),
	   a single variable, a Boolean value indicating whether sorts or
	   required, and a list of arguments.
  RETURNS: A quantified term, possibly restricted by the appropriate sort.
  CAUTION: The term is potentially destructively changed (depends on the
           BuildFunction).
********************************************************/
{                                                                  
#ifdef DEBUG                                                          
  if (!eml_IsModalOperator(ModalTop)) {
    printf("\n\nIllegal input in eml_BuildSortQuantifierFor\n");
    misc_Error();
  }
#endif                                                             
                                                                   
  if (WithSort) {
    eml_SYMBOLTYPE symbolType;
    SYMBOL foSymbol;
#ifdef DEBUG                                                          
    if (!eml_IsNullaryPredicate(RelSubTop)) {
      printf("\n\nIllegal relational argument of modal operator, in eml_BuildSortQuantifierFor\n");
      misc_Error();
    }
#endif                                                             

    symbolType = eml_VARSORT;
    foSymbol   = eml_GetFoSymbolAssocWithPropSymbol(RelSubTop,1,symbolType);
    if (foSymbol == symbol__NULL) {
      foSymbol = eml_CreateFoSymbol(RelSubTop, symbolType, 1, symbol_STATLEX);
      /* Inheritance of precidence? */
      /* symbol_SetOrdering(foSymbol,symbol_Ordering(PropSymbol)); */
    }
    return BuildFunction(Term, eml_FoQuantAssocWithPropSymbol(ModalTop), 
		list_List(Var),   
		list_List(term_Create(eml_FoJunctorAssocWithPropSymbol(ModalTop),
		  list_Cons(term_Create(foSymbol,list_List(term_Copy(Var))), Arguments))));
    /* exists x S(x) /\ ... 
     * forall x S(x) -> ... */
  }
  else
    return BuildFunction(Term, eml_FoQuantAssocWithPropSymbol(ModalTop),
		list_List(Var), Arguments);
    /* exists x ... 
     * forall x ... */
}

static TERM  eml_RplacWithDesignatedPropAtom(TERM Term, SYMBOL RplacSymbol)
/**************************************************************
  INPUT:   A formula and a symbol.
  RETURNS: A replacement without arguments, e.g. true, false, id, etc.
  CAUTION: The term in destructively changed.
***************************************************************/
{
  term_RplacTop(Term,RplacSymbol);
  list_DeleteWithElement(term_ArgumentList(Term),term_Delete);
  term_RplacArgumentList(Term,list_Nil());

  return Term;
}

static TERM  eml_RplacWithOpAndArgs(TERM Term, SYMBOL RplacSymbol, 
		LIST Arguments)
/**************************************************************
  INPUT:   A formula, a symbol and a list of terms.
  RETURNS: A replacement with argument list.
  CAUTION: The term in destructively changed.
***************************************************************/
{
  term_RplacTop(Term,RplacSymbol);
  term_RplacArgumentList(Term,Arguments);

  return Term;
}

static TERM  eml_RemoveTrivialOperator(TERM Term)
/**************************************************************
  INPUT:   A formula.
  RETURNS: The formula after
           removal of "or", "and", "comp" and "sum" with only one argument
  CAUTION: The term in destrucively changed.
***************************************************************/
{
  if (eml_IsNaryJunctor(term_TopSymbol(Term)) &&
      list_Empty(list_Cdr(term_ArgumentList(Term)))) {
    TERM helpTerm = term_FirstArgument(Term);
    Term = eml_RplacWithOpAndArgs(Term, term_TopSymbol(helpTerm), term_ArgumentList(helpTerm));
    term_Free(helpTerm);
  }

  return Term;
}

static TERM  eml_RemoveTrivialAtomsAndOps(TERM Term)
/**************************************************************
  INPUT:   A formula.
  RETURNS: The formula after applying
           or(true,...) -> true,      and(true,phi) -> phi
           or(false,phi) -> phi,      and(false,...) -> false
                 etc plus
           dia(alpha,false) -> false, box(alpha,true) -> true
           dia(false,phi) -> false,   box(false,phi) -> true
  CAUTION: The term in destructively changed.
***************************************************************/
{
  SYMBOL Top, Subtop, RelSubtop;
  LIST   Scan;
  TERM   Subterm, RelSubterm;
  BOOL   Update;

  if (!term_IsComplex(Term))
    return Term;

  Top    = term_TopSymbol(Term);
  Update = FALSE;

  /* Term = and(...) */
  if (symbol_Equal(Top,fol_And())) {
    for (Scan=term_ArgumentList(Term); !list_Empty(Scan); Scan=list_Cdr(Scan)) {
      Subterm = eml_RemoveTrivialAtomsAndOps(list_Car(Scan));
      Subtop = term_TopSymbol(Subterm);
      if (symbol_Equal(Subtop,fol_True()))
        Update = TRUE;
      else if (symbol_Equal(Subtop,fol_False()))
        /* and(..., false, ...) -> false */ 
        return eml_RplacWithDesignatedPropAtom(Term,fol_False());
    }
    if (Update) {
      /* and(..., true, ..., true, ...) -> and(..., ..., ...) */ 
      term_RplacArgumentList(Term,list_DeleteElementIf(term_ArgumentList(Term),fol_IsTrue,term_Delete));
      if (list_Empty(term_ArgumentList(Term))) {
        term_RplacTop(Term,fol_True());
        return Term;
      }
    }
    return eml_RemoveTrivialOperator(Term); /* and(phi) -> phi */
  }
  /* Term = or(...) */
  else if (symbol_Equal(Top,fol_Or())) {
    for (Scan=term_ArgumentList(Term); !list_Empty(Scan); Scan=list_Cdr(Scan)) {
      Subterm = eml_RemoveTrivialAtomsAndOps(list_Car(Scan));
      Subtop = term_TopSymbol(Subterm);
      if (symbol_Equal(Subtop,fol_False()))
        Update = TRUE;
      else if (symbol_Equal(Subtop,fol_True()))
        /* or(..., true, ...) -> true */ 
        return eml_RplacWithDesignatedPropAtom(Term,fol_True());
    }
    if (Update) {
      term_RplacArgumentList(Term,list_DeleteElementIf(term_ArgumentList(Term),fol_IsFalse,term_Delete));
      if (list_Empty(term_ArgumentList(Term))) {
        term_RplacTop(Term,fol_False());
        return Term;
      }
    }
    return eml_RemoveTrivialOperator(Term); /* or(phi) -> phi */
  }
  /* Term = not(...) */
  else if (symbol_Equal(Top,fol_Not())) {
    Subterm = eml_RemoveTrivialAtomsAndOps(term_FirstArgument(Term));
    Subtop = term_TopSymbol(Subterm);
    if (eml_IsDesignatedPropSymbol(Subtop))
      /* not(false) -> true,  not(true) -> false,  not(div) -> id,  not(id) -> div */
      return eml_RplacWithDesignatedPropAtom(Term,eml_GetDualSymbol(Subtop));
  }
  /* Term = forall(...)  or  Term = exists(...) */
  else if (fol_IsQuantifier(Top)) {
    Subterm = eml_RemoveTrivialAtomsAndOps(term_SecondArgument(Term));
    Subtop = term_TopSymbol(Subterm);
    if (symbol_Equal(Subtop,fol_True()))
      /* forall ... true -> true,  exists ... true -> true */
      return eml_RplacWithDesignatedPropAtom(Term,fol_True());
    else if (symbol_Equal(Subtop,fol_False()))
      /* forall ... false -> false,  exists ... false -> false */
      return eml_RplacWithDesignatedPropAtom(Term,fol_False());
  }
  /* Term = dia(...) or Term = box(...) */
  else if (eml_IsModalOperator(Top)) {
    RelSubterm = eml_RemoveTrivialAtomsAndOps(term_FirstArgument(Term));
    RelSubtop = term_TopSymbol(RelSubterm);
    Subterm    = eml_RemoveTrivialAtomsAndOps(term_SecondArgument(Term));
    Subtop    = term_TopSymbol(Subterm);
    if (symbol_Equal(Top,eml_Dia()) && 
         (symbol_Equal(Subtop,fol_False()) || symbol_Equal(RelSubtop,fol_False())))
      /* dia(alpha,false) -> false,  dia(false,phi) -> false */
      return eml_RplacWithDesignatedPropAtom(Term,fol_False());
    else if (symbol_Equal(Top,eml_Box()) && 
       (symbol_Equal(Subtop,fol_True()) || symbol_Equal(RelSubtop,fol_False()))) 
      /* box(alpha,true) -> true,  box(true,phi) -> true */
      return eml_RplacWithDesignatedPropAtom(Term,fol_True());
    else if (symbol_Equal(RelSubtop, eml_Id())) {
      /* dia(id,phi) -> phi,  box(id,phi) -> phi */
      term_Delete(RelSubterm);
      list_Delete(term_ArgumentList(Term));
      Term = eml_RplacWithOpAndArgs(Term, Subtop, term_ArgumentList(Subterm));
      term_Free(Subterm);
      return Term;
    }
  }
  /* Term = domain(...) or Term = range(...) */
  else if (symbol_Equal(Top,eml_Domain()) || symbol_Equal(Top,eml_Range())) {
    RelSubterm = eml_RemoveTrivialAtomsAndOps(term_FirstArgument(Term));
    RelSubtop = term_TopSymbol(RelSubterm);
    if (symbol_Equal(RelSubtop,fol_False()))
      /* domain(false) -> false,  range(false) -> false */
      return eml_RplacWithDesignatedPropAtom(Term,fol_False());
    else if (symbol_Equal(RelSubtop, eml_Id()))
      /* domain(id) -> true,  range(id) -> true */
      return eml_RplacWithDesignatedPropAtom(Term,fol_True());
  }
  /* Term = implies(...) */
  else if (symbol_Equal(Top,fol_Implies())) {
    Subterm = eml_RemoveTrivialAtomsAndOps(term_FirstArgument(Term));
    Subtop = term_TopSymbol(Subterm);
    if (symbol_Equal(Subtop,fol_False()))
      /* implies(false,phi) -> true */
      return eml_RplacWithDesignatedPropAtom(Term,fol_True());
    else if (symbol_Equal(Subtop,fol_True())) {
      term_Delete(Subterm);
      Subterm = term_SecondArgument(Term);
      list_Delete(term_ArgumentList(Term));
      Term = eml_RplacWithOpAndArgs(Term,term_TopSymbol(Subterm),term_ArgumentList(Subterm));
      term_Free(Subterm);
      return eml_RemoveTrivialAtomsAndOps(Term);
    }
    Subterm = eml_RemoveTrivialAtomsAndOps(term_SecondArgument(Term));
    Subtop = term_TopSymbol(Subterm);
    if (symbol_Equal(Subtop,fol_True()))
      /* implies(phi,true) -> true */
      return eml_RplacWithDesignatedPropAtom(Term,fol_True());
    else if (symbol_Equal(Subtop,fol_False())) 
      Term = eml_RplacWithOpAndArgs(Term, fol_Not(),
		      list_DeleteElementIf(term_ArgumentList(Term),fol_IsFalse,term_Delete));
  }
  /* Term = equiv(...) */
  else if (symbol_Equal(Top,fol_Equiv())) {
    Subterm = eml_RemoveTrivialAtomsAndOps(term_FirstArgument(Term));
    Subtop = term_TopSymbol(Subterm);
    if (symbol_Equal(Subtop,fol_False())) {
      Term = eml_RplacWithOpAndArgs(Term, fol_Not(),
		      list_DeleteElementIf(term_ArgumentList(Term),fol_IsFalse,term_Delete));
      return eml_RemoveTrivialAtomsAndOps(Term);
    }
    else if (symbol_Equal(Subtop,fol_True())) {
      term_Delete(Subterm);
      Subterm = term_SecondArgument(Term);
      list_Delete(term_ArgumentList(Term));
      Term = eml_RplacWithOpAndArgs(Term, term_TopSymbol(Subterm),
		      term_ArgumentList(Subterm));
      term_Free(Subterm);
      return eml_RemoveTrivialAtomsAndOps(Term);
    }
    Subterm = eml_RemoveTrivialAtomsAndOps(term_SecondArgument(Term));
    Subtop = term_TopSymbol(Subterm);
    if (symbol_Equal(Subtop,fol_False())) {
      Term = eml_RplacWithOpAndArgs(Term, fol_Not(),
		list_DeleteElementIf(term_ArgumentList(Term),fol_IsFalse,term_Delete));
    }
    else if (symbol_Equal(Subtop,fol_True())) {
      term_Delete(Subterm);
      Subterm = term_FirstArgument(Term);
      list_Delete(term_ArgumentList(Term));
      Term = eml_RplacWithOpAndArgs(Term, term_TopSymbol(Subterm),
		term_ArgumentList(Subterm));
      term_Free(Subterm);
    }
  }
  /* Term = comp(...) */
  else if (symbol_Equal(Top,eml_Comp())) {
    for (Scan=term_ArgumentList(Term); !list_Empty(Scan); Scan=list_Cdr(Scan)) {
      Subterm = eml_RemoveTrivialAtomsAndOps(list_Car(Scan));
      Subtop = term_TopSymbol(Subterm);
      if (symbol_Equal(Subtop,eml_Id()))
        Update = TRUE;
      else if (symbol_Equal(Subtop,fol_False()))
	/* comp(..., false, ...) -> false */
        return eml_RplacWithDesignatedPropAtom(Term,fol_False());
    }
    if (Update) {
      /* comp(..., id, ..., id, ...) -> comp(..., ..., ...) */
      term_RplacArgumentList(Term,list_DeleteElementIf(term_ArgumentList(Term),eml_IsId,term_Delete));
      if (list_Empty(term_ArgumentList(Term))) {
        term_RplacTop(Term,eml_Id());
        return Term;
      }
    }
    return eml_RemoveTrivialOperator(Term); /* comp(alpha) -> alpha */
  }
  /* Term = sum(...) */
  else if (symbol_Equal(Top,eml_Sum())) {
    for (Scan=term_ArgumentList(Term); !list_Empty(Scan); Scan=list_Cdr(Scan)) {
      Subterm = eml_RemoveTrivialAtomsAndOps(list_Car(Scan));
      Subtop = term_TopSymbol(Subterm);
      if (symbol_Equal(Subtop,eml_Div()))
        Update = TRUE;
      else if (symbol_Equal(Subtop,fol_True()))
	/* sum(..., true, ...) -> true */
        return eml_RplacWithDesignatedPropAtom(Term,fol_True());
    }
    if (Update) {
      /* sum(..., div, ..., div, ...) -> sum(..., ..., ...) */
      term_RplacArgumentList(Term,list_DeleteElementIf(term_ArgumentList(Term),eml_IsDiv,term_Delete));
      if (list_Empty(term_ArgumentList(Term))) {
        term_RplacTop(Term,eml_Div());
        return Term;
      }
    }
    return eml_RemoveTrivialOperator(Term); /* sum(alpha) -> alpha */
  } 
  /* Term = test(...) */
  else if (symbol_Equal(Top,eml_Test())) {
    Subterm = eml_RemoveTrivialAtomsAndOps(term_FirstArgument(Term));
    Subtop = term_TopSymbol(Subterm);
    if (symbol_Equal(Subtop,fol_False()))
      /* test(false) -> false */
      return eml_RplacWithDesignatedPropAtom(Term,fol_False());
    else if (symbol_Equal(Subtop, fol_True()))
      /* test(true) -> id  */
      return eml_RplacWithDesignatedPropAtom(Term,eml_Id());
  }
  /* Term = conv(...) */
  else if (symbol_Equal(Top,eml_Conv())) {
    RelSubterm = eml_RemoveTrivialAtomsAndOps(term_FirstArgument(Term));
    RelSubtop = term_TopSymbol(RelSubterm);
    if (symbol_Equal(RelSubtop,fol_True()) ||
		symbol_Equal(RelSubtop,fol_False()) ||
		symbol_Equal(RelSubtop,eml_Id()) ||
		symbol_Equal(RelSubtop,eml_Div()))
      /* conv(false) -> false,  conv(true) -> true,  conv(id) -> id,  conv(div) -> div */
      return eml_RplacWithDesignatedPropAtom(Term,RelSubtop);
    else if (symbol_Equal(RelSubtop,eml_Test())) {
      /* conv(test(...)) -> test(...)  */
      Term = eml_RplacWithOpAndArgs(Term, RelSubtop, term_ArgumentList(RelSubterm));
      term_Free(RelSubterm);
      return Term;
    }
  }
  /* Term = domrestr(...) or Term = ranrestr(...) */
  else if (symbol_Equal(Top,eml_DomRestr()) || symbol_Equal(Top,eml_RanRestr())) {
    RelSubterm = eml_RemoveTrivialAtomsAndOps(term_FirstArgument(Term));
    RelSubtop  = term_TopSymbol(RelSubterm);
    Subterm    = eml_RemoveTrivialAtomsAndOps(term_SecondArgument(Term));
    Subtop     = term_TopSymbol(Subterm);
    if (symbol_Equal(RelSubtop,fol_False()) || symbol_Equal(Subtop,fol_False()))
      /* domrestr(false,...), domrestr(...,false) -> false,  
       * ranrestr(false,...), ranrestr(...,false) -> false */
      return eml_RplacWithDesignatedPropAtom(Term,fol_False());
    else if (symbol_Equal(Subtop, fol_True())) {
      /* domrestr(alpha,true) -> alpha,  ranrestr(alpha,true) -> alpha */
      term_Delete(Subterm);
      list_Delete(term_ArgumentList(Term));
      Term = eml_RplacWithOpAndArgs(Term, RelSubtop, term_ArgumentList(RelSubterm));
      term_Free(RelSubterm);
      return Term;
    }
  }

  return Term;
}

TERM  eml_ObviousSimplifications(TERM Term)
/**************************************************************
  INPUT:   An eml formula.
  RETURNS: The formula after performing the following simplifications:
             - eliminate true, false, id, etc 
             - remove "or" and "and" with only one argument
             - merge quantors
  CAUTION: The term is destructively changed.
  NOTE:    This does not obliviate simplification of the first-order
	   formula.  E.g. dia(alpha,true) cannot be reduced in the modal
	   context.
	   Nested expression of n-ary associative operations (and, comp,
	   ...) are reduced later.
***************************************************************/
{
  Term = eml_RemoveTrivialAtomsAndOps(Term);
  Term = cnf_SimplifyQuantors(Term);  /* strictly only needed for non-propositional formulae */

  return Term;
}

BOOL eml_IsPropositional(TERM Term)
/**************************************************************
  INPUT:   A formula.
  RETURNS: True, if formula does not contain a first order quantifier.
***************************************************************/
{
  LIST Scan;
  BOOL result;

  result = TRUE;
  if (term_IsComplex(Term)) {
    if (fol_IsQuantifier(term_TopSymbol(Term)))
      return FALSE;
    else 
      for(Scan=term_ArgumentList(Term); result && !list_Empty(Scan); Scan=list_Cdr(Scan))
        result = eml_IsPropositional(list_Car(Scan));
  }

  return result;
}

static TERM eml_RplacPropByFoAtom(TERM Atom, eml_SYMBOLTYPE SymbolType,
		int Arity, LIST Args)
/**************************************************************
  INPUT:   A propositional atom, a symbol type, a natural number, and a
           list of variables.
  RETURNS: A first-order atom with the specified arity formed with the
	   symbol of the specified type and the specified arguments.
	   The predicate name is built from the propositional name.
  EFFECT:  If the predicate symbol does not exist in the signature it is
           created and inherit its precedence setting.
  CAUTION: The atom is destructively changed.
***************************************************************/
{
  SYMBOL propSymbol, foSymbol;

  propSymbol = term_TopSymbol(Atom);

  if (symbol_Equal(propSymbol,fol_True()) || symbol_Equal(propSymbol,fol_False())) {
    return Atom;  /* Attempt to simplify ?? */
  }

  Args    = list_CopyWithElement(Args,(POINTER (*)())term_CopyIterative);
  /* term_Copy does not suffice for the path-based functional
   * translation*/

  if (symbol_Equal(propSymbol,eml_Id())) {
    symbol_SetOrdering(fol_Equality(),symbol_Ordering(eml_Id()));
    return eml_RplacWithOpAndArgs(Atom, fol_Equality(), Args);
  }
  else if (symbol_Equal(propSymbol,eml_Div())) {
    return eml_RplacWithOpAndArgs(Atom, fol_Not(),
		    list_List(term_Create(fol_Equality(),Args)));
  }

  foSymbol   = eml_GetFoSymbolAssocWithPropSymbol(propSymbol,Arity,SymbolType);

#ifdef DEBUG
  if (!eml_FlutedType(SymbolType) && (foSymbol != symbol__NULL) &&
	    (symbol_Arity(foSymbol) != Arity)) {
    fprintf(stderr, "Symbol %s declared with arity %d, but arity %d required\n",
    symbol_Name(foSymbol), symbol_Arity(foSymbol),Arity);
    misc_Error();
    return NULL;
  } 
#endif

  if (foSymbol == symbol__NULL) {
    foSymbol = eml_CreateFoSymbol(propSymbol, SymbolType, Arity, symbol_STATLEX);
    /* Precedence among the first-order symbols is inherited from that among
     * the propositional symbols */
    symbol_SetOrdering(foSymbol,symbol_Ordering(propSymbol));
  }

  /* This does not work */
  if (!flag_GetFlagValue(flag_DOCPROOF)) {
    eml_PROPSYMBOLLIST = list_Cons((POINTER)propSymbol,eml_PROPSYMBOLLIST);
  }

  return eml_RplacWithOpAndArgs(Atom, foSymbol, Args);
}

SYMBOL eml_CreateFlutedSymbol(SYMBOL PropSymbol, eml_SYMBOLTYPE SymbolType,
		char* NameSuffix, int Arity, int Status)
/*******************************************************
  INPUT:   A symbol (which MUST be a propositonal symbol/nullary
	   predicate symbol), a symbol type (of the fo symbol to be
	   created), the predefined name suffix, the arity of the new
	   symbol, the status of the new symbol.
  RETURNS: A first-order symbol corresponding to PropSymbol (special for
	   the functial translation into fluted logic with the sorts
	   encoded in the predicate names).
  REMARK:  Similar to dfg_SymbolDecl.  Prototypical.
********************************************************/
{
  char   newName[symbol__SYMBOLMAXLEN];
  SYMBOL symbol;
  
  sprintf(newName,"%s%s_%d%s", eml_PREDNAMEPREFIX[SymbolType],
		  symbol_Name(PropSymbol), Arity, NameSuffix);

  /* Check if this symbol was declared earlier */
  symbol = symbol_Lookup(newName);
  if (symbol != 0) {
    /* Symbol was declared before 
     * Check if the old and new symbol type are equal */
    /* Eventually we would want more checking to be done here, i.e.
     * we need new designated symbol types */
    if ( !symbol_IsPredicate(symbol)) {
      fprintf(stderr, "In eml_CreateFlutedSymbol: Symbol %s was already declared as ",
	      newName);
      switch (symbol_Type(symbol)) {
      case symbol_CONSTANT:
      case symbol_FUNCTION:
	fputs("function\n", stderr); break;
      case symbol_JUNCTOR:
	fputs("junctor", stderr); break;
      default:
	fputs("unknown non-predicate type\n", stderr);
      }
      misc_Error();
    }
    /* Now check the old and new arity if specified */
    if (Arity != -2 && Arity != symbol_Arity(symbol)) {
      fprintf(stderr, "In eml_CreateFlutedSymbol: Symbol %s was already declared with arity %d\n",
	      newName, symbol_Arity(symbol));
      misc_Error();
    }
  } else
    /* Symbol was not declared before */
    symbol = symbol_CreatePredicate(newName, Arity, Status);

  return  symbol;
}

static TERM eml_RplacPropByFlutedAtom(TERM Atom, eml_SYMBOLTYPE
		SymbolType, char* NameSuffix, int Arity, LIST Args)
/**************************************************************
  INPUT:   A propositional atom, a symbol type, a string, a natural
           number, and a list of variables.
  RETURNS: A first-order atom with the specified arity formed with the
	   symbol of the specified type and the specified arguments.
	   The predicate name is built from the propositional name and
	   (different from eml_RplacPropByFoAtom) the given name suffix.
  EFFECT:  A new predicate symbol is created and inherits its precedence
           setting.
  CAUTION: The atom is destructively changed.
***************************************************************/
{
  SYMBOL propSymbol, foSymbol;

  propSymbol = term_TopSymbol(Atom);

  if (symbol_Equal(propSymbol,fol_True()) || symbol_Equal(propSymbol,fol_False()))
    return Atom;

  Args    = list_CopyWithElement(Args,(POINTER (*)())term_Copy);

  if (flag_GetFlagValue(flag_EMLFFSORTS))  /* with explicit sorts */
    foSymbol = eml_CreateFlutedSymbol(propSymbol, SymbolType,
		    symbol_Name(propSymbol), Arity, symbol_STATLEX);
  else                    /* sort info encoded in predicate names */
    foSymbol = eml_CreateFlutedSymbol(propSymbol, SymbolType, NameSuffix, Arity,
		    symbol_STATLEX);
  symbol_SetOrdering(foSymbol,symbol_Ordering(propSymbol));  /* ?? */

  return eml_RplacWithOpAndArgs(Atom, foSymbol, Args);
}

static TERM eml_EmbedInRelCalculusProp(TERM Term)
/**************************************************************
  INPUT:   An Eml (propositional) formula.
  RETURNS: Its translation into a relational formula as a right
           injective formula.
  CAUTION: The term is destructively changed.
***************************************************************/
{
  SYMBOL Top;

  Top    = term_TopSymbol(Term);

  /* Term = p */
  if (symbol_IsPredicate(Top)) {
    return eml_RplacWithOpAndArgs(Term, eml_Comp(),
	     list_Cons(term_Copy(Term), 
		list_List(term_Create(fol_True(), list_Nil()))));
    /* comp(p,true) */
  }
  /* Term = not(phi), or(...), and(...), comp(...), sum(...), 
   *        test(phi), domrestr(...), ranrestr(...) */
  else  if (eml_IsRelationalJunctor(Top)) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term); !list_Empty(Scan); Scan=list_Cdr(Scan))
      eml_EmbedInRelCalculusProp(list_Car(Scan));
    return Term;
  }
  /* Term = dia(alpha,phi) */
  else  if (symbol_Equal(Top, eml_Dia())) {
    LIST args;          
    args = list_Cons(eml_EmbedInRelCalculusRel(term_FirstArgument(Term)), 
	     list_List(eml_EmbedInRelCalculusProp(term_SecondArgument(Term))));
    return eml_RplacWithOpAndArgs(Term, eml_Comp(), args);
    /* comp(alpha,tr(phi)) */
  }
  /* Term = box(alpha,phi) */
  else  if (symbol_Equal(Top, eml_Box())) {
    TERM helpTerm;          
    helpTerm = term_Create(eml_Comp(), 
		 list_Cons(eml_EmbedInRelCalculusRel(term_FirstArgument(Term)),
		   list_List(term_Create(fol_Not(), 
		     list_List(eml_EmbedInRelCalculusProp(term_SecondArgument(Term)))))));
    return eml_RplacWithOpAndArgs(Term, fol_Not(), list_List(helpTerm));
    /* not(comp(alpha,not(tr(phi)))) */
  }
  /* Term = domain(alpha)  or  Term = range(alpha) */
  else if (symbol_Equal(Top,eml_Domain()) || symbol_Equal(Top,eml_Range())) {
    LIST args;          
    term_RplacTop(Term, eml_Comp());
    if (symbol_Equal(Top,eml_Domain())) 
      args = list_Cons(eml_EmbedInRelCalculusRel(term_FirstArgument(Term)), 
		   list_List(term_Create(fol_True(), list_Nil())));
    else
      args = list_Cons(term_Create(fol_True(), list_Nil()), 
		   list_List(eml_EmbedInRelCalculusRel(term_FirstArgument(Term))));
    term_RplacArgumentList(Term,args);

    return Term;
    /* comp(alpha,true)  or  comp(true,alpha) */
  }
  else {
    fprintf(stderr, "\nNo case for %s in eml_RelationalTranslationProp()\n",
             symbol_Name(Top));
    misc_Error();
    return NULL;
  }

}

static TERM eml_EmbedInRelCalculusRel(TERM Term)
/**************************************************************
  INPUT:   An Eml (propositional) formula.
  RETURNS: Its translation into a relational formula as a right
           injective formula.
  CAUTION: The term is destructively changed.
***************************************************************/
{
  SYMBOL Top;

  Top    = term_TopSymbol(Term);

  /* Term = p */
  if (symbol_IsPredicate(Top)) {
    return Term;
  }
  /* Term = test(phi) */
  else  if (symbol_Equal(Top, eml_Test())) {
    return eml_RplacWithOpAndArgs(Term, fol_And(),
	     list_Cons(eml_EmbedInRelCalculusProp(term_FirstArgument(Term)),
	       list_List(term_Create(eml_Id(), list_Nil()))));
    /* tr(phi) /\ id */
  }
  /* Term = domrestr(alpha,phi) */
  else  if (symbol_Equal(Top, eml_DomRestr())) {
    return eml_RplacWithOpAndArgs(Term, fol_And(),
	     list_Cons(term_FirstArgument(Term),
	       list_List(eml_EmbedInRelCalculusProp(term_SecondArgument(Term)))));
    /* alpha /\ tr(phi) */
  }
  /* Term = ranrestr(alpha,phi) */
  else  if (symbol_Equal(Top, eml_RanRestr())) {
    return eml_RplacWithOpAndArgs(Term, fol_And(),
	     list_Cons(term_FirstArgument(Term),
	       list_List(term_Create(eml_Conv(),
		 list_List(eml_EmbedInRelCalculusProp(term_SecondArgument(Term)))))));
    /* alpha /\ conv(tr(phi)) */
  }
  /* Term = not(phi), or(...), and(...), comp(...), sum(...) */
  /* It is important that this test is done last */
  else  if (eml_IsRelationalJunctor(Top)) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term); !list_Empty(Scan); Scan=list_Cdr(Scan))
      eml_EmbedInRelCalculusRel(list_Car(Scan));
    return Term;
  }
  else {
    fprintf(stderr, "\nNo case for %s in eml_RelationalTranslationProp()\n",
             symbol_Name(Top));
    misc_Error();
    return NULL;
  }

}


static TERM eml_RelationalTranslationProp(TERM Term, TERM Var)
/**************************************************************
  INPUT:   An eml (propositional) formula and a variable (the free variable).
  RETURNS: Its standard (relational) translation.
  CAUTION: The term is destructively changed.
***************************************************************/
{
  SYMBOL Top;

  Top    = term_TopSymbol(Term);

  /* Term = p */
  if (symbol_IsPredicate(Top)) {
    return eml_RplacPropByFoAtom(Term,eml_PROP,1,list_List(Var));
  }
  /* Term = not(phi), Term = or(...) or Term = and(...) */
  else  if (eml_IsPropJunctor(Top)) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term); !list_Empty(Scan); Scan=list_Cdr(Scan))
      eml_RelationalTranslationProp(list_Car(Scan),Var);
    return Term;
  }
  /* Term = dia(alpha,phi)  or  Term = box(alpha,phi) */
  else if (eml_IsModalOperator(Top)) {
    TERM newVar, helpTerm;          
    newVar   = term_Create(symbol_CreateStandardVariable(), list_Nil());
    helpTerm = term_Create(eml_FoJunctorAssocWithPropSymbol(Top), 
        list_Cons(eml_RelationalTranslationRel(term_FirstArgument(Term),Var,newVar),
        list_List(eml_RelationalTranslationProp(term_SecondArgument(Term),newVar))));
    return eml_RplacWithQuantifier(Term, eml_FoQuantAssocWithPropSymbol(Top), list_List(newVar),
		    list_List(helpTerm));
    /* exists y tr(alpha)(x,y) /\ tr(phi)(y)  or
     * forall y tr(alpha)(x,y) -> tr(phi)(y)  */
  }
  /* Term = domain(alpha)  or  Term = range(alpha) */
  else if (symbol_Equal(Top,eml_Domain()) || symbol_Equal(Top,eml_Range())) {
    TERM newVar, helpTerm;          
    newVar   = term_Create(symbol_CreateStandardVariable(), list_Nil());
    if (symbol_Equal(Top,eml_Domain())) 
      helpTerm = eml_RelationalTranslationRel(term_FirstArgument(Term),Var,newVar);
    else
      helpTerm = eml_RelationalTranslationRel(term_FirstArgument(Term),newVar,Var);

    return eml_RplacWithQuantifier(Term, eml_FoQuantAssocWithPropSymbol(Top), list_List(newVar),
		    list_List(helpTerm));
    /* exists y tr(alpha)(x,y)  or  exists y tr(alpha)(y,x) */
  }
  else {
    fprintf(stderr, "\nNo case for %s in eml_RelationalTranslationProp()\n",
             symbol_Name(Top));
    misc_Error();
    return NULL;
  }

}

static TERM eml_RelationalTranslationRel(TERM Term, TERM VarX, TERM VarY)
/**************************************************************
  INPUT:   An eml (relational) formula and two variables (the free
           variables).
  RETURNS: Its standard (relational) translation.
  CAUTION: The term is destructively changed.
***************************************************************/
{
  SYMBOL Top;

  Top    = term_TopSymbol(Term);

  /* Term = r */
  if (symbol_IsPredicate(Top)) {
    return eml_RplacPropByFoAtom(Term,eml_REL,2,list_Cons(VarX,list_List(VarY)));
  }
  /* Term = not(alpha), Term = or(...) or Term = and(...) */
  else  if (eml_IsPropJunctor(Top)) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term); !list_Empty(Scan); Scan=list_Cdr(Scan))
      eml_RelationalTranslationRel(list_Car(Scan),VarX,VarY);
    return Term;
  }
  /* Term = conv(alpha) */
  else  if (symbol_Equal(Top,eml_Conv())) {
    TERM helpTerm;          
    helpTerm = eml_RelationalTranslationRel(term_FirstArgument(Term),VarY,VarX);
    Term     = eml_RplacWithOpAndArgs(Term, term_TopSymbol(helpTerm),
		    term_ArgumentList(helpTerm));
    term_Free(helpTerm);
    return Term;
    /* tr(alpha)(y,x) */
  }
  /* Term = comp(...)  or  sum(...)*/
  else if (symbol_Equal(Top,eml_Comp()) || symbol_Equal(Top,eml_Sum())) {
    TERM var, newVar, helpTerm;          
    LIST Scan, varList;
    var = VarX;
    varList = list_Nil();
    for (Scan=term_ArgumentList(Term); !list_Empty(list_Cdr(Scan)); Scan=list_Cdr(Scan)) {
      newVar   = term_Create(symbol_CreateStandardVariable(), list_Nil());
      varList  = list_Nconc(varList, list_List(newVar));
      eml_RelationalTranslationRel(list_Car(Scan),var,newVar);
      var = newVar;
    }
    eml_RelationalTranslationRel(list_Car(Scan),var,VarY);

    helpTerm = term_Create(eml_FoJunctorAssocWithPropSymbol(Top), term_ArgumentList(Term)); 
    return eml_RplacWithQuantifier(Term,
		eml_FoQuantAssocWithPropSymbol(Top), varList, list_List(helpTerm));
      /* exists z1,z2 tr(alpha1)(x,z1) /\ tr(alpha2)(z1,z2) /\ tr(alpha3)(z2,y) ...  or
       * forall z1,z2 tr(alpha1)(x,z1) \/ tr(alpha2)(z1,z2) \/ tr(alpha3)(z2,y) ... */
  }
  /* Term = test(phi) */
  else if (symbol_Equal(Top,eml_Test())) {
    TERM helpTerm;          
    helpTerm = eml_RelationalTranslationProp(term_FirstArgument(Term),VarY);
    return eml_RplacWithOpAndArgs(Term, fol_And(),
	list_Cons(term_Create(fol_Equality(), 
		list_Cons(term_Copy(VarX), list_List(term_Copy(VarY)))),
	list_List(helpTerm)));
    /* equal(x,y) /\ tr(phi)(y) */
  }
  /* Term = domrestr(alpha,phi)  or  ranrestr(alpha,phi) */
  else if (symbol_Equal(Top,eml_DomRestr()) || symbol_Equal(Top,eml_RanRestr())) {
    TERM helpTerm, helpTermRel;          
    helpTermRel = eml_RelationalTranslationRel(term_FirstArgument(Term),VarX,VarY);
    if (symbol_Equal(Top,eml_DomRestr()))
      helpTerm    = eml_RelationalTranslationProp(term_SecondArgument(Term),VarX);
    else
      helpTerm    = eml_RelationalTranslationProp(term_SecondArgument(Term),VarY);
    return eml_RplacWithOpAndArgs(Term, fol_And(),
	     list_Cons(helpTermRel, list_List(helpTerm)));
    /* tr(alpha)(x,y) /\ tr(phi)(x)  or  tr(alpha)(x,y) /\ tr(phi)(y) */
  }
  else {
    fprintf(stderr, "\nNo case for %s in eml_RelationalTranslationRel()\n",
             symbol_Name(Top));
    misc_Error();
    return NULL;
  }

}

static TERM eml_RelationalTranslation(TERM Term)
/**************************************************************
  INPUT:   An eml (propositional) formula.
  RETURNS: Its standard (relational) translation.
  CAUTION: The term is destructively changed.
***************************************************************/
{
  TERM newVarX, helpTerm;

  symbol_ResetStandardVarCounter();

  newVarX = term_Create(symbol_CreateStandardVariable(), list_Nil());

  if (flag_GetFlagValue(flag_EML2REL)) {  /* Embedding into the relational calculus */
    TERM newVarY;

    Term = eml_EmbedInRelCalculusProp(Term);
    if (flag_GetFlagValue(flag_PPROBLEM))
      printf("\n [Rel] \t");term_Print(Term);

    newVarY  = term_Create(symbol_CreateStandardVariable(), list_Nil());
    helpTerm = eml_RelationalTranslationRel(term_Copy(Term),newVarX,newVarY);

    if (flag_GetFlagValue(flag_PPROBLEM))
      printf("\n [RelTr] \t");
    return eml_RplacWithQuantifier(Term, fol_All(),
             list_Cons(newVarX,list_List(newVarY)), list_List(helpTerm));
    /* all x y tr(Term,x,y) */
  }
  else {  /* Default */
    helpTerm = eml_RelationalTranslationProp(term_Copy(Term),newVarX);

    if (flag_GetFlagValue(flag_PPROBLEM))
      printf("\n [RelTr] \t");
    return eml_RplacWithQuantifier(Term, fol_All(), list_List(newVarX), list_List(helpTerm));
    /* all x tr(Term,x) */
  }
}

static TERM eml_FunctTranslIntoFlutedLogicProp(TERM Term, LIST VarList,
		char* SortEncoding, int VarListLength)
/**************************************************************
  INPUT:   An eml (propositional) formula, a variable list and its
           length.
  RETURNS: Its functional translation into fluted logic.
  CAUTION: The term is destructively changed.
***************************************************************/
{
  SYMBOL Top;

  Top    = term_TopSymbol(Term);
  
#ifdef DEBUG
  if (VarListLength != list_Length(VarList)) {
    fprintf(stderr, "Arity VarListLength conflict\n");
    misc_Error();
    return NULL;
  }
#endif

  /* Term = p */
  if (symbol_IsPredicate(Top)) {
    return eml_RplacPropByFlutedAtom(Term, eml_FLUTEDPROP, SortEncoding,
		    VarListLength, VarList);
  }
  /* Term = not(phi), Term = or(...) or Term = and(...) */
  else  if (eml_IsPropJunctor(Top)) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term); !list_Empty(Scan); Scan=list_Cdr(Scan))
      eml_FunctTranslIntoFlutedLogicProp(list_Car(Scan), VarList,
		      SortEncoding, VarListLength);
    return Term;
  }
  /* Term = dia(r,phi)  or  Term = box(r,phi) */
  else if (eml_IsModalOperator(Top)) {
    SYMBOL relSubTop;
    TERM newVar, helpTerm;          
    LIST newVarList;
    char newSortEncoding[symbol__SYMBOLMAXLEN];

    relSubTop = term_TopSymbol(term_FirstArgument(Term));
    newVar     = term_Create(symbol_CreateStandardVariable(), list_Nil());
    newVarList = list_Nconc(list_Copy(VarList), list_List(newVar));
    sprintf(newSortEncoding,"%s%s", SortEncoding, symbol_Name(relSubTop));
    if (flag_GetFlagValue(flag_EMLFUNCNDEQ)) {
      helpTerm = eml_BuildSortQuantifierFor(Term, Top, relSubTop,
		newVar, flag_GetFlagValue(flag_EMLFFSORTS), 
		list_List(eml_FunctTranslIntoFlutedLogicProp(term_SecondArgument(Term),
			newVarList, newSortEncoding, VarListLength+1)),
		(POINTER (*)())eml_CreateQuantifier);
      return eml_RplacWithOpAndArgs(Term, eml_FoJunctorAssocWithPropSymbol(Top), 
		list_Cons(eml_FunctTranslIntoFlutedLogicRel(term_FirstArgument(Term),
			VarList, SortEncoding, VarListLength), 
		list_List(helpTerm)));
      /* nde_r(vl) /\ exists y_r tr(phi)(vl,y)  or
       * nde_r(vl) -> forall y_r tr(phi)(vl,y)  */
    }
    else {
      helpTerm = term_Create(eml_FoJunctorAssocWithPropSymbol(Top),
		list_Cons(eml_FunctTranslIntoFlutedLogicRel(term_FirstArgument(Term),
		    list_CopyWithElement(VarList,(POINTER (*)())term_Copy), 
		    SortEncoding, VarListLength), 
		  list_List(eml_FunctTranslIntoFlutedLogicProp(term_SecondArgument(Term),
		    list_CopyWithElement(newVarList,(POINTER (*)())term_Copy),
		    newSortEncoding, VarListLength+1))));
      return eml_BuildSortQuantifierFor(Term, Top, relSubTop, newVar,
		  flag_GetFlagValue(flag_EMLFFSORTS),
		  list_List(helpTerm), (POINTER (*)())eml_RplacWithQuantifier);
      /* exists y_r nde_r(vl) /\ tr(phi)(vl,y)  or
       * forall y_r nde_r(vl) -> tr(phi)(vl,y)  */
    }
  }
  else {
    fprintf(stderr, "\nNo case for %s in eml_FunctTranslIntoFlutedLogicProp()\n",
             symbol_Name(Top));
    misc_Error();
    return NULL;
  }

}

static TERM eml_FunctTranslIntoFlutedLogicRel(TERM Term, LIST VarList,
		char* SortEncoding, int VarListLength)
/**************************************************************
  INPUT:   An EML relational formula, a variable list, a string (encoding
           the sorts of the accessibility relations in the scope), and
	   the length of the variable list (corresponds to the modal
	   depth).
  RETURNS: Its functional translation into fluted logic.
  CAUTION: The term is destructively changed.
***************************************************************/
{
  SYMBOL Top;

  Top    = term_TopSymbol(Term);

  /* Term = r */
  if (symbol_IsPredicate(Top)) {
    if (flag_GetFlagValue(flag_EMLKD))
      return eml_RplacWithDesignatedPropAtom(Term, fol_True());
    else 
      return eml_RplacPropByFlutedAtom(Term, eml_FLUTEDNDE,
		      SortEncoding, VarListLength, VarList);
  }
  else {
    fprintf(stderr, "\nNo case for %s in eml_FunctTranslIntoFlutedLogicProp()\n",
             symbol_Name(Top));
    misc_Error();
    return NULL;
  }
}

static TERM eml_FunctTranslProp(TERM Term, TERM WorldPath)
/**************************************************************
  INPUT:   An eml (propositional) formula, a variable list and its
           length.
  RETURNS: Its functional translation.
  CAUTION: The term is destructively changed.
***************************************************************/
{
  SYMBOL Top;

  Top    = term_TopSymbol(Term);

  /* Term = p */
  if (symbol_IsPredicate(Top)) {
    return eml_RplacPropByFoAtom(Term,eml_PROP,1,list_List(WorldPath));
  }
  /* Term = not(phi), Term = or(...) or Term = and(...) */
  else  if (eml_IsPropJunctor(Top)) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term); !list_Empty(Scan); Scan=list_Cdr(Scan))
      eml_FunctTranslProp(list_Car(Scan), WorldPath);
    return Term;
  }
  /* Term = dia(alpha,phi)  or  Term = box(alpha,phi) */
  else if (eml_IsModalOperator(Top)) {
    SYMBOL relSubTop;
    TERM newVar, newWorldPath, helpTerm;          
    relSubTop    = term_TopSymbol(term_FirstArgument(Term));
    newVar       = term_Create(symbol_CreateStandardVariable(), list_Nil());
    newWorldPath = term_Create(eml_ApplyFunct(), list_Cons(WorldPath,
			list_List(newVar)));
    if (flag_GetFlagValue(flag_EMLFUNCNDEQ)) {
      helpTerm = eml_BuildSortQuantifierFor(Term, Top, relSubTop, newVar, TRUE, 
		list_List(eml_FunctTranslProp(term_SecondArgument(Term),
			newWorldPath)), (POINTER (*)())eml_CreateQuantifier);
      return eml_RplacWithOpAndArgs(Term, eml_FoJunctorAssocWithPropSymbol(Top), 
		list_Cons(eml_FunctTranslRel(term_FirstArgument(Term),
			WorldPath), 
		list_List(helpTerm)));
      /* nde_r(wp) /\ exists y_r tr(phi)(app(wp,y_r))  or
       * nde_r(wp) -> forall y_r tr(phi)(app(wp,y_r))  */
    }
    else {
      helpTerm = term_Create(eml_FoJunctorAssocWithPropSymbol(Top), 
	  list_Cons(eml_FunctTranslRel(term_FirstArgument(Term), WorldPath),
          list_List(eml_FunctTranslProp(term_SecondArgument(Term), newWorldPath))));
      return eml_BuildSortQuantifierFor(Term, Top, relSubTop, newVar, TRUE,
		  list_List(helpTerm), (POINTER (*)())eml_RplacWithQuantifier);
      /* exists y_r nde_r(wp) /\ tr(phi)(app(wp,y_r))  or
       * forall y_r nde_r(wp) -> tr(phi)(app(wp,y_r))  */
    }
  }
  else {
    fprintf(stderr, "\nNo case for %s in eml_FunctTranslProp()\n",
             symbol_Name(Top));
    misc_Error();
    return NULL;
  }

}

static TERM eml_FunctTranslRel(TERM Term, TERM WorldPath)
/**************************************************************
  INPUT:   An eml (propositional) formula, a variable list and its
           length.
  RETURNS: Its functional translation into fluted logic.
  CAUTION: The term is destructively changed.
***************************************************************/
{
  SYMBOL Top;

  Top    = term_TopSymbol(Term);

  /* Term = r */
  if (symbol_IsPredicate(Top)) {
    if (flag_GetFlagValue(flag_EMLKD))
      return eml_RplacWithDesignatedPropAtom(Term, fol_True());
    else 
      return eml_RplacPropByFoAtom(Term,eml_NDE,1,list_List(WorldPath));
  }
  else {
    fprintf(stderr, "\nNo case for %s in eml_FunctTranslProp()\n",
             symbol_Name(Top));
    misc_Error();
    return NULL;
  }
}


static TERM eml_FunctionalTranslation(TERM Term, BOOL IntoFlutedLogic)
/**************************************************************
  INPUT:   An eml (propositional) formula and a flag.
  RETURNS: Its functional translation.  If the flag is true the
	   translation is into fluted logic, meaning n-ary predicates
	   are generated.
  CAUTION: The term is destructively changed.
***************************************************************/
{
  symbol_ResetStandardVarCounter();

  if (IntoFlutedLogic) {
    char nameSuffix[symbol__SYMBOLMAXLEN] = "";
    
    if (flag_GetFlagValue(flag_PPROBLEM))
      printf("\n [FuncFTr] \t");
    Term = eml_FunctTranslIntoFlutedLogicProp(Term, list_Nil(), nameSuffix, 0);
    /* Check whether sharing is possible on the name suffixes,
     * e.g.whether the iterative allocation of new memory in
     * eml_FunctTranslIntoFlutedLogicProp is necessary. */
    /* memory_Free(nameSuffix, symbol__SYMBOLMAXLEN); */
    return  Term;
    /* note: without world quantifier */
  }
  else {
    TERM newVar, helpTerm;
    if (flag_GetFlagValue(flag_PPROBLEM))
      printf("\n [FuncTr] \t");
    newVar   = term_Create(symbol_CreateStandardVariable(), list_Nil());
    helpTerm = eml_FunctTranslProp(term_Copy(Term),newVar);
    return eml_RplacWithQuantifier(Term, fol_All(),
		list_List(newVar), list_List(helpTerm));
    /* all x tr(Term,x) */
  }

}


static TERM  eml_TranslateToFol(TERM Term)
/**************************************************************
  INPUT:   An eml formula.
  RETURNS: Its translation as specified by flag_EMLTR, provided
           the formula is provided, otherwise the simplification of the
           original formula.
  CAUTION: The term is destructively changed.
***************************************************************/
{
  if (!eml_IsPropositional(Term))
    fol_NormalizeVars(Term);

  Term = eml_ObviousSimplifications(Term);
  if (flag_GetFlagValue(flag_PPROBLEM))
    printf("\n [Simpl]\t");term_Print(Term);

  if (eml_IsPropositional(Term)) {
    if (flag_GetFlagValue(flag_EMLTR) == flag_EMLTRREL)
      Term = eml_RelationalTranslation(Term);
    else if (flag_GetFlagValue(flag_EMLTR) == flag_EMLTRFUNC ||
		    flag_GetFlagValue(flag_EMLTR) == flag_EMLTROPTFUNC)
      Term = eml_FunctionalTranslation(Term, flag_GetFlagValue(flag_EMLFFLUTED));
  }
  else
    printf("\n [FOL]\t");

  return Term;
}

void eml_TranlateAndConvertToCnf(LIST* AxiomList, LIST* ConjectureList, LIST* AxClauses, 
                 LIST* AllLabels,
                 HASH  TermLabelToClauselist, HASH  ClauseToTermLabellist,
                 BOOL  KeepProofSearch,
                 LIST* Symblist)
/**************************************************************
  INPUT:   A list of axiomformulae,
           a list of conjectureformulae,
           a pointer to a list in which clauses derived from axiomformulae 
           are stored,
           a pointer to a list in which clauses derived from 
           conjectureformulae are stored,
           a pointer to a list of all termlabels,
           a hasharray storing for every termlabel the list of clauses 
           derived from the term (if DocProof is set),
           a hasharray in which for every clause the list of labels 
           of the terms used for deriving the clause is stored (if DocProof 
           is set),
           a pointer to a list of symbols which have to be deleted later if
           the ProofSearch object is kept.
  RETURNS: If KeepProofSearch is TRUE, then the ProofSearch object is not
           freed but returned.
           Else, NULL is returned.
  REMARK:  Prototypical.  Not all arguments are used, as yet.
***************************************************************/
{
  LIST  Scan, Pair;
  TERM  helpTerm;
  int	flag;

  /* This does not work */
  if (!flag_GetFlagValue(flag_DOCPROOF))
    eml_PROPSYMBOLLIST = list_Nil();
  
  /* Print warnings for non sensical flag combinations */
  flag = flag_GetFlagValue(flag_EMLTR);
  if ((flag != flag_EMLTRFUNC && flag != flag_EMLTROPTFUNC)) {
    if (flag_GetFlagValue(flag_EMLFUNCNDEQ))
      printf("\nWarning: Flag %s has no effect together with %s translation method",
	flag_GetFlagName(flag_EMLFUNCNDEQ), flag_GetTranslationMethod(flag));
    if (flag_GetFlagValue(flag_EMLFFSORTS))
      printf("\nWarning: Flag %s has no effect together with %s translation method",
	flag_GetFlagName(flag_EMLFFSORTS), flag_GetTranslationMethod(flag));
    if (flag_GetFlagValue(flag_EMLFFSORTS))
      printf("\nWarning: Flag %s has no effect together with %s translation method",
	flag_GetFlagName(flag_EMLFFSORTS), flag_GetTranslationMethod(flag));
  }
  /* Set additional flags */
  if ((flag == flag_EMLTROPTFUNC)) {
    flag_SetFlagValue(flag_QUANTEXCH, 1);
  }
  if ((flag == flag_EMLTRFUNC || flag == flag_EMLTROPTFUNC) &&
      flag_GetFlagValue(flag_EMLFFSORTS)) {
    flag_SetFlagValue(flag_SORTS, 1);  /* ?? */
  }

  if (flag_GetFlagValue(flag_PPROBLEM)) {
    printf("\n----------------------TRANSLATION-START--------------------------");
    flag_PrintSpecialTranslationMethod();
    flag_PrintSpecialTranslationFlags();
    printf("\n Precedence: "); symbol_PrintPrecedence();
    printf("\nAxioms:");
  }

  /* Initialisation */
   if (flag_GetFlagValue(flag_EMLTR) == flag_EMLTRFUNC || 
	flag_GetFlagValue(flag_EMLTR) == flag_EMLTROPTFUNC)
     eml_InitFunctTransl();
      
  for(Scan=*AxiomList; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Pair = list_Car(Scan);
    helpTerm = (TERM) list_PairSecond(Pair);
    if (flag_GetFlagValue(flag_PPROBLEM)) {
      printf("\nLabel : %s", (char*) list_PairFirst(Pair)); 
      if (eml_IsPropositional(helpTerm))
        printf("\n [%s]\t", flag_Name(flag_EML));
      else
        printf("\n [FOL]\t");
      term_Print(helpTerm);
    }
    helpTerm = eml_TranslateToFol(helpTerm);
    if (flag_GetFlagValue(flag_PPROBLEM))
      term_Print(helpTerm);
  }

  if (flag_GetFlagValue(flag_PPROBLEM))
    printf("\nConjecture:");

  for(Scan=*ConjectureList; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Pair = list_Car(Scan);
    helpTerm = (TERM) list_PairSecond(Pair);
    if (flag_GetFlagValue(flag_PPROBLEM)) {
      printf("\nLabel : %s", (char*) list_PairFirst(Pair)); 
      if (eml_IsPropositional(helpTerm))
        printf("\n [%s]\t", flag_Name(flag_EML));
      else
        printf("\n [FOL]\t");
      term_Print(helpTerm);
    }
#ifdef DEBUG
    if (!symbol_Equal(term_TopSymbol(helpTerm),fol_Not())) {
      fprintf(stderr, "In eml_TranlateAndConvertToCnf(): The list of conjectures should have been negated\n");
      misc_Error();
    } 
    else 
#endif
      term_RplacFirstArgument(helpTerm,
	      eml_TranslateToFol(term_FirstArgument(helpTerm)));
    if (flag_GetFlagValue(flag_PPROBLEM))
      term_Print(helpTerm);
  }

  /* Check whether the following does the job (as it is supposed to).
   * Deleting the symbol list causes a segmentation fault later on in cnf.c.
   * Is this possbily due to a missing
   *  if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
   * ??
   * */
  /* if (!flag_GetFlagValue(flag_DOCPROOF)) {
    list_DeleteWithElement(eml_PROPSYMBOLLIST,symbol_Delete);
  } */

}
