/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *              EXTENDED MODAL LOGICS ROUTINES            * */
/* *                                                        * */
/* *  $Module:   EML                                        * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.4 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/11/30 19:44:32 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: eml.h,v $ */

#ifndef _EMLDFG_
#define _EMLDFG_

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "term.h"
#include "list.h"
#include "hasharray.h"

/**************************************************************/
/* Global Variables and Constants (Only seen by macros)       */
/**************************************************************/

extern SYMBOL  eml_APPLYF;
extern SYMBOL  eml_BOX;
extern SYMBOL  eml_COMP;
extern SYMBOL  eml_CONV;
extern SYMBOL  eml_DIA;
extern SYMBOL  eml_DIV;
extern SYMBOL  eml_DOMAIN;
extern SYMBOL  eml_DOMRESTR;
extern SYMBOL  eml_ID;
extern SYMBOL  eml_RANGE;
extern SYMBOL  eml_RANRESTR;
extern SYMBOL  eml_SUM;
extern SYMBOL  eml_TEST;

typedef enum { eml_PROP, 	eml_REL,       eml_NDE,
               eml_FLUTEDPROP, 	eml_FLUTEDNDE, eml_VARSORT } eml_SYMBOLTYPE;

extern LIST *eml_PROPINDEXTOFOSYMBOL;
extern SYMBOL eml_FLUTEDSYMBOLS[symbol__MAXSIGNATURE][symbol__MAXSIGNATURE];    
/* Records which symbols of the fluted signature have been created */

/**************************************************************/
/* Access to the first-order symbols.                         */
/**************************************************************/

static __inline__ SYMBOL eml_Box(void)
{
  return eml_BOX;
}

static __inline__ SYMBOL eml_Dia(void)
{
  return eml_DIA;
}

static __inline__ SYMBOL eml_Domain(void)
{
  return eml_DOMAIN;
}

static __inline__ SYMBOL eml_Range(void)
{
  return eml_RANGE;
}

static __inline__ SYMBOL eml_Comp(void)
{
  return eml_COMP;
}

static __inline__ SYMBOL eml_Sum(void)
{
  return eml_SUM;
}

static __inline__ SYMBOL eml_Conv(void)
{
  return eml_CONV;
}

static __inline__ SYMBOL eml_Test(void)
{
  return eml_TEST;
}

static __inline__ SYMBOL eml_Id(void)
{
  return eml_ID;
}

static __inline__ SYMBOL eml_Div(void)
{
  return eml_DIV;
}

static __inline__ SYMBOL eml_DomRestr(void)
{
  return eml_DOMRESTR;
}

static __inline__ SYMBOL eml_RanRestr(void)
{
  return eml_RANRESTR;
}

static __inline__ SYMBOL eml_ApplyFunct(void)
{
  return eml_APPLYF;
}

/**************************************************************/
/* Symbol identification and access                           */
/**************************************************************/

static __inline__ BOOL eml_IsId(TERM S)
{
  return symbol_Equal(eml_ID,term_TopSymbol(S));
}

static __inline__ BOOL eml_IsDiv(TERM S)
{
  return symbol_Equal(eml_DIV,term_TopSymbol(S));
}

static __inline__ BOOL eml_IsModalOperator(SYMBOL S)
{
  return (symbol_Equal(eml_BOX,S) || symbol_Equal(eml_DIA,S));
}

static __inline__ BOOL eml_IsRelPropSymbol(SYMBOL S)
{
  return eml_IsModalOperator(S);
}

static __inline__ BOOL eml_IsModalLiteral(TERM T)
/* unused */
{
  return fol_IsLiteral(T) || eml_IsModalOperator(term_TopSymbol(T)) ||
      (symbol_Equal(term_TopSymbol(T),fol_Not()) && 
       eml_IsModalOperator(term_TopSymbol(term_FirstArgument(T))));
}


static __inline__ BOOL eml_IsPropJunctor(SYMBOL S)
{
  return symbol_Equal(S, fol_AND) || 
      symbol_Equal(S, fol_OR) || 
      symbol_Equal(S, fol_NOT) || symbol_Equal(S, fol_IMPLIES) ||
      symbol_Equal(S, fol_EQUIV);
}

static __inline__ BOOL eml_IsDesignatedPropSymbol(SYMBOL S)
{
  return symbol_Equal(S, fol_TRUE) || 
      symbol_Equal(S, fol_FALSE) || 
      symbol_Equal(S, eml_ID) || 
      symbol_Equal(S, eml_DIV);
}

static __inline__ BOOL eml_IsNullaryPredicate(SYMBOL S)
{
  return symbol_IsPredicate(S) && !eml_IsDesignatedPropSymbol(S) &&
	  (symbol_Arity(S) == 0);
}

static __inline__ BOOL eml_IsNaryJunctor(SYMBOL S)
{
  return symbol_Equal(S, fol_AND) || 
      symbol_Equal(S, fol_OR) || 
      symbol_Equal(S, eml_COMP) || 
      symbol_Equal(S, eml_SUM);
}

static __inline__ BOOL eml_IsRelationalJunctor(SYMBOL S)
{
  return eml_IsPropJunctor(S) ||
      symbol_Equal(S, eml_COMP) || 
      symbol_Equal(S, eml_CONV) ||
      symbol_Equal(S, eml_SUM) ||
      symbol_Equal(S, eml_DOMRESTR) ||
      symbol_Equal(S, eml_RANRESTR) ||
      symbol_Equal(S, eml_TEST);
}

static __inline__ BOOL eml_IsModalJunctor(SYMBOL S)
{
  return eml_IsModalOperator(S) || eml_IsPropJunctor(S);
}

static __inline__ TERM eml_Atom(TERM Lit)
{
  if (term_TopSymbol(Lit) == fol_NOT)
    return term_FirstArgument(Lit);
  else
    return Lit;
}

static __inline__ SYMBOL eml_GetDualSymbol(SYMBOL symbol)
{
  SYMBOL dual;

  dual = symbol;

  if (symbol_Equal(symbol,fol_True()))
    dual = fol_False();
  else if (symbol_Equal(symbol,fol_False()))
    dual = fol_True();
  else if (symbol_Equal(symbol,fol_Or()))
    dual = fol_And();
  else if (symbol_Equal(symbol,fol_And()))
    dual =  fol_Or();
  else if (symbol_Equal(symbol,fol_Exist()))
    dual = fol_All();
  else if (symbol_Equal(symbol,fol_All()))
    dual = fol_Exist();
  else if (symbol_Equal(symbol,eml_Dia()))
    dual = eml_Box();
  else if (symbol_Equal(symbol,eml_Box()))
    dual = eml_Dia();
  else if (symbol_Equal(symbol,eml_Id()))
    dual = eml_Div();
  else if (symbol_Equal(symbol,eml_Div()))
    dual = eml_Id();
  else if (symbol_Equal(symbol,eml_Comp()))
    dual = eml_Sum();
  else if (symbol_Equal(symbol,eml_Sum()))
    dual = eml_Comp();

  return dual;
}

/**************************************************************/
/* Propositional/first order symbol associations              */
/**************************************************************/

static BOOL eml_VarSortType(eml_SYMBOLTYPE SymbolType)
{
  return SymbolType == eml_VARSORT;
}

static BOOL eml_FlutedType(eml_SYMBOLTYPE SymbolType)
{
  return SymbolType == eml_FLUTEDPROP || SymbolType == eml_FLUTEDNDE;
}

static __inline__ void eml_SetPropFlutedSymbolAssocList(SYMBOL PropSymbol, 
		int Arity, SYMBOL FoSymbol)
{
  eml_FLUTEDSYMBOLS[symbol_Index(PropSymbol)][Arity-1] = FoSymbol;
}

static __inline__ SYMBOL eml_GetFoSymbolAssocWithPropSymbol(SYMBOL
		PropSymbol, NAT Arity, eml_SYMBOLTYPE SymbolType)
{
  if (eml_FlutedType(SymbolType)) {
    SYMBOL symbol;
    symbol = eml_FLUTEDSYMBOLS[symbol_Index(PropSymbol)][Arity-1];
    if (symbol_IsSymbol(symbol))
      return symbol;
  }
  else if (eml_VarSortType(SymbolType)) {
    return PropSymbol;  /* Struture sharing!
			 * Alternatively, set up another array called
			 * eml_PROPINDEXTOSORTSYMBOL, say */
  } 
  else {
    LIST symbolList;
    symbolList = eml_PROPINDEXTOFOSYMBOL[symbol_Index(PropSymbol)];
    if (!list_Empty(symbolList))
      return (SYMBOL) list_First(symbolList);
  }

  return symbol__NULL;
}

static __inline__ void eml_SetPropFoSymbolAssocList(SYMBOL PropSymbol, LIST FoSymbols)
{
  eml_PROPINDEXTOFOSYMBOL[symbol_Index(PropSymbol)] =  FoSymbols;
}

static __inline__ SYMBOL eml_FoQuantAssocWithPropSymbol(SYMBOL PropSymbol)
{
  LIST symbolList;
  symbolList = eml_PROPINDEXTOFOSYMBOL[symbol_Index(PropSymbol)];
  if (list_Empty(symbolList))
    return symbol__NULL;
  else
    return (SYMBOL) list_First(symbolList);
}

static __inline__ SYMBOL eml_FoJunctorAssocWithPropSymbol(SYMBOL PropSymbol)
{
  LIST symbolList;
  symbolList = eml_PROPINDEXTOFOSYMBOL[symbol_Index(PropSymbol)];
  if (list_Empty(symbolList) || list_Empty(list_Cdr(symbolList)))
    return symbol__NULL;
  else
  return (SYMBOL) list_Second(symbolList);
}

/**************************************************************/
/* Functions                                                  */
/**************************************************************/

void eml_Init(void);
void eml_Free(void);
void eml_TranlateAndConvertToCnf(LIST*, LIST*, LIST*, LIST*, HASH, HASH, BOOL, LIST*);

#endif
