/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                 FLAGS OF SPASS                         * */
/* *                                                        * */
/* *  $Module:   FLAGS                                      * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.74 $                                          * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 17:46:59 $                                      * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: flags.c,v $ */


#include "flags.h"

/**************************************************************/
/* Global Variables                                           */
/**************************************************************/

FLAGSTORE  flag_DEFAULTVALUES;
FLAGSTORE  flag_VALUES;
char *flag_STRINGS[flag_MAXFLAG];


/**************************************************************/
/* Functions                                                  */
/**************************************************************/

void flag_Init(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  Sets all default values for known flags.
  MEMORY:  Allocates memory for the default store.
***************************************************************/
{
  /* Start with defaultvalues */
  flag_DEFAULTVALUES = flag_NewStore();
  flag_InstallStore(flag_DEFAULTVALUES);

  /* Autonomous mode */
  flag_STRINGS[flag_AUTO]         = "Auto";
  flag_VALUES[flag_AUTO]          = 1;
  
  /* Set of Support Mode */
  flag_STRINGS[flag_SOS]         = "SOS";
  flag_VALUES[flag_SOS]          = 0;

  /* If set input is considered from stdio and printed to stdout */
  flag_STRINGS[flag_STDIN]        = "Stdin";
  flag_VALUES[flag_STDIN]         = 0;

   /* If set interactive queries are possible */
  flag_STRINGS[flag_INTERACTIVE]        = "Interactive";
  flag_VALUES[flag_INTERACTIVE]         = 0;

  /* If set only Flotter CNF-translation is performed */
  flag_STRINGS[flag_FLOTTER]        = "Flotter";
  flag_VALUES[flag_FLOTTER]         = 0;

  /* Choose calculus: paramodulation */
  flag_STRINGS[flag_CPARA]        = "CPara";
  flag_VALUES[flag_CPARA]         = 0;

  /* Choose calculus: ordered paramodulation */
  flag_STRINGS[flag_CORDPARA]     = "COrdPara";
  flag_VALUES[flag_CORDPARA]      = 0;

  /* Choose calculus: superpostion with equality factoring */
  flag_STRINGS[flag_CSUPE]        = "CSupE";
  flag_VALUES[flag_CSUPE]         = 0;

  /* Choose calculus: superpostion with merging paramodulation */
  flag_STRINGS[flag_CSUPP]        = "CSupP";
  flag_VALUES[flag_CSUPP]         = 0;

  /* Allowed number of loops, -1 means no restriction */
  flag_STRINGS[flag_LOOPS]        = "Loops";
  flag_VALUES[flag_LOOPS]         = -1;

  /* Allowed number of splits, -1 means no restriction */
  flag_STRINGS[flag_SPLITS]       = "Splits";
  flag_VALUES[flag_SPLITS]        = -1;

  /* Decides the level of sort usage: if 0 then no sort information is processed,
     if 1 all negative monadic literals with a variable as its argument are processed,
     if 2 all negative monadic literals are processed */
  flag_STRINGS[flag_SORTS]        = "Sorts";
  flag_VALUES[flag_SORTS]         = 1;

  /* If set, non-constant Solem terms in the clausal form of the
   * conjecture are replaced by constants */
  flag_STRINGS[flag_QUANTEXCH]    = "QuantExch";
  flag_VALUES[flag_QUANTEXCH]     = 0;

  /* ForwardSubsumption output not activated */
  flag_STRINGS[flag_PSUB]         = "PSub";
  flag_VALUES[flag_PSUB]          = 0;

  /* Maximal memory allocation */
  flag_STRINGS[flag_MEMORY]       = "Memory";
  flag_VALUES[flag_MEMORY]        = -1;

  /* Document static soft typing  */
  flag_STRINGS[flag_DOCSST]       = "DocSST";
  flag_VALUES[flag_DOCSST]        = 0;

  /* Document dynamic soft typing  */
  flag_STRINGS[flag_DOCDST]       = "DocDST";
  flag_VALUES[flag_DOCDST]        = 0;

  /* Rewriting output not activated  */
  flag_STRINGS[flag_PREW]         = "PRew";
  flag_VALUES[flag_PREW]          = 0;

  /* Condensing output not activated  */
  flag_STRINGS[flag_PCON]         = "PCon";
  flag_VALUES[flag_PCON]          = 0;

  /* Tautology output not activated  */
  flag_STRINGS[flag_PTAUT]        = "PTaut";
  flag_VALUES[flag_PTAUT]         = 0;

  /* Output of obvious red. not activated  */
  flag_STRINGS[flag_POBV]         = "PObv";
  flag_VALUES[flag_POBV]          = 0;

  /* SortSimplification output not activated  */
  flag_STRINGS[flag_PSSI]         = "PSSi";
  flag_VALUES[flag_PSSI]          = 0;

  /* Static soft typing output not activated  */
  flag_STRINGS[flag_PSST]         = "PSST";
  flag_VALUES[flag_PSST]          = 0;

  /* Proof output not activated  */
  flag_STRINGS[flag_DOCPROOF]     = "DocProof";
  flag_VALUES[flag_DOCPROOF]      = 0;

  /* Only the parent formulae used in a proof are printed, no detailed proof  */
  flag_STRINGS[flag_DOCFORMULAEONLY]  = "DocFormulaeOnly";
  flag_VALUES[flag_DOCFORMULAEONLY]   = 0;

  /*  Mapping between formulae and clauses not activated  */
  flag_STRINGS[flag_FORMULAE2CLAUSE]     = "Formulae2Clause";
  flag_VALUES[flag_FORMULAE2CLAUSE]      = 0;

  /* ClauseReduction output not activated  */
  flag_STRINGS[flag_PCLR]         = "PClR";
  flag_VALUES[flag_PCLR]          = 0;

  /* Unit conflict output not activated  */
  flag_STRINGS[flag_PUNC]         = "PUnC";
  flag_VALUES[flag_PUNC]          = 0;

  /* Derived clauses output not activated  */
  flag_STRINGS[flag_PDER]         = "PDer";
  flag_VALUES[flag_PDER]          = 0;

  /* Given clause output not activated  */
  flag_STRINGS[flag_PGIVEN]       = "PGiven";
  flag_VALUES[flag_PGIVEN]        = 1;

  /* Kept clauses output not activated  */
  flag_STRINGS[flag_PKEPT]        = "PKept";
  flag_VALUES[flag_PKEPT]         = 0;

  /* Split backtrack emphasizing not activated */
  flag_STRINGS[flag_DOCSPLIT]     = "DocSplit";
  flag_VALUES[flag_DOCSPLIT]      = 0;

  /* Print information about input clauses */
  flag_STRINGS[flag_PPROBLEM]     = "PProblem";
  flag_VALUES[flag_PPROBLEM]      = 1;

  /* Print all derived empty clauses */
  flag_STRINGS[flag_PEMPTYCLAUSE] = "PEmptyClause";
  flag_VALUES[flag_PEMPTYCLAUSE]  = 0;

  /* Print statistic about memory, clauses */
  flag_STRINGS[flag_PSTATISTIC]   = "PStatistic";
  flag_VALUES[flag_PSTATISTIC]    = 1;

  /* Output saturated set of clauses to file, default no */
  flag_STRINGS[flag_FPMODEL]      = "FPModel";
  flag_VALUES[flag_FPMODEL]       = 0;

   /* Output proof in DFG format to file, default no */
  flag_STRINGS[flag_FPDFGPROOF]   = "FPDFGProof";
  flag_VALUES[flag_FPDFGPROOF]    = 0;

  /* Output the actual values of all SPASS flags */
  flag_STRINGS[flag_PFLAGS]       = "PFlags";
  flag_VALUES[flag_PFLAGS]        = 0;

  /* Optimized skolemization output not activated  */
  flag_STRINGS[flag_POPTSKOLEM]   = "POptSkolem";
  flag_VALUES[flag_POPTSKOLEM]    = 0;

  /* Strong skolemization output activated  */
  flag_STRINGS[flag_PSTRSKOLEM]   = "PStrSkolem";
  flag_VALUES[flag_PSTRSKOLEM]    = 0;

    /* Application of definitions  output activated  */
  flag_STRINGS[flag_PAPPLYDEFS]   = "PApplyDefs";
  flag_VALUES[flag_PAPPLYDEFS]    = 0;

  /* Diagnostic information about clauses is output */
  flag_STRINGS[flag_PDIAGNOSE]   = "PDiagnosis";
  flag_VALUES[flag_PDIAGNOSE]    = 0;

  /* Amount of time (seconds) available to SPASS, -1 means arbitrary */
  flag_STRINGS[flag_TIMELIMIT]    = "TimeLimit";
  flag_VALUES[flag_TIMELIMIT]     = -1;
  
  /* Select: 0 -> no selection, 1 -> select if multiple maximal literals, 2 -> always select */
  flag_STRINGS[flag_SELECT]       = "Select";
  flag_VALUES[flag_SELECT]        = 1;

  /* Activates the inference rule Empty Sort */
  flag_STRINGS[flag_IEMS]       = "IEmS";
  flag_VALUES[flag_IEMS]        = 1;

  /* Activates the inference rule Sort Resolution */
  flag_STRINGS[flag_ISOR]       = "ISoR";
  flag_VALUES[flag_ISOR]        = 1;

  /* Activates the inference rule Equality Resolution */
  flag_STRINGS[flag_IEQR]       = "IEqR";
  flag_VALUES[flag_IEQR]        = 1;

  /* Activates the inference rule Equality Factoring */
  flag_STRINGS[flag_IEQF]       = "IEqF";
  flag_VALUES[flag_IEQF]        = 1;

  /* Activates the inference rule Merging Paramodulation */
  flag_STRINGS[flag_IMPM]       = "IMPm";
  flag_VALUES[flag_IMPM]        = 0;

  /* Activates the inference rule Superposition Right */
  flag_STRINGS[flag_ISPR]       = "ISpR";
  flag_VALUES[flag_ISPR]        = 1;
  
   /* Inference rule Ordered Paramodulation not active */  
  flag_STRINGS[flag_IOPM]       = "IOPm";   
  flag_VALUES[flag_IOPM]        = 0;   

  /*   Inference rule Paramodulation not active */  
  flag_STRINGS[flag_ISPM]       = "ISPm";   
  flag_VALUES[flag_ISPM]        = 0;   

  /* Activates the inference rule Superposition Left */
  flag_STRINGS[flag_ISPL]       = "ISpL";
  flag_VALUES[flag_ISPL]        = 1;
  
  /* Activates the inference rule General Resolution */
  flag_STRINGS[flag_IGER]       = "IGeR";
  flag_VALUES[flag_IGER]        = 1;

  /* Dectivates the inference rule Simple Hyperresolution */
  flag_STRINGS[flag_ISHY]       = "ISHy";
  flag_VALUES[flag_ISHY]        = 0;
  
  /* Activates the inference rule Ordered Hyperresolution */
  flag_STRINGS[flag_IOHY]       = "IOHy";
  flag_VALUES[flag_IOHY]        = 0;

  /* Activates the inference rule General Factoring */
  flag_STRINGS[flag_IGEF]       = "IGeF";
  flag_VALUES[flag_IGEF]        = 1;

  /* Activates the inference rule Bounded Unit Resolution */
  flag_STRINGS[flag_IBUR]       = "IBUR";
  flag_VALUES[flag_IBUR]        = 0;

  /* Activates the inference rule Definition Application */
  flag_STRINGS[flag_IDEF]       = "IDEF";
  flag_VALUES[flag_IDEF]        = 0;

  /* Activates the inference rule  Unit Resolution */
  flag_STRINGS[flag_IUNR]       = "IUnR";
  flag_VALUES[flag_IUNR]        = 0;
  
  /* Activates Forward Rewriting */
  flag_STRINGS[flag_RFREW]      = "RFRew";
  flag_VALUES[flag_RFREW]       = 1;

  /* Activates Backward Rewriting */
  flag_STRINGS[flag_RBREW]      = "RBRew";
  flag_VALUES[flag_RBREW]       = 1;

  /* Activates Unit Conflict */
  flag_STRINGS[flag_RUNC]       = "RUnC";
  flag_VALUES[flag_RUNC]        = 1;

  /* Activates Forward Subsumption */
  flag_STRINGS[flag_RFSUB]      = "RFSub";
  flag_VALUES[flag_RFSUB]       = 1;

  /* Activates Backward Subsumption */
  flag_STRINGS[flag_RBSUB]      = "RBSub";
  flag_VALUES[flag_RBSUB]       = 1;

  /* Activates Forward Clause Reduction */
  flag_STRINGS[flag_RFCLR]      = "RFClR";
  flag_VALUES[flag_RFCLR]       = 1;

  /* Activates Backward Clause Reduction */
  flag_STRINGS[flag_RBCLR]      = "RBClR";
  flag_VALUES[flag_RBCLR]       = 1;

  /* Activates the reduction rule Obvious Reduction */
  flag_STRINGS[flag_ROBV]       = "RObv";
  flag_VALUES[flag_ROBV]        = 1;

  /* Activates the reduction rule Tautology */
  flag_STRINGS[flag_RTAUT]      = "RTaut";
  flag_VALUES[flag_RTAUT]       = 1;

  /* Activates the reduction rule Sort Simplification */
  flag_STRINGS[flag_RSSI]       = "RSSi";
  flag_VALUES[flag_RSSI]        = 1;

  /* Activates  static soft typing */
  flag_STRINGS[flag_RSST]       = "RSST";
  flag_VALUES[flag_RSST]        = 1;

  /* Activates  dynamic soft typing */
  flag_STRINGS[flag_RDST]       = "RDST";
  flag_VALUES[flag_RDST]        = 1;

  /* Activates Condensing */
  flag_STRINGS[flag_RCON]      = "RCon";
  flag_VALUES[flag_RCON]       = 1;

  /* Activates reduction of input clauses */
  flag_STRINGS[flag_RINPUT]    = "RInput";
  flag_VALUES[flag_RINPUT]     = 1;

  /* Activates application of definitions */
  flag_STRINGS[flag_APPLYDEFS]    = "ApplyDefs";
  flag_VALUES[flag_APPLYDEFS]     = 0;

  /* If true usable and worked off are completely interreduced; otherwise only worked off */
  flag_STRINGS[flag_FULLRED]   = "FullRed";
  flag_VALUES[flag_FULLRED]    = 1;

  /* Activates unit saturation of input clauses */
  flag_STRINGS[flag_SATINPUT]   = "SatInput";
  flag_VALUES[flag_SATINPUT]    = 0;

  /* Ratio between weight and depth selection of clauses from usable */
  flag_STRINGS[flag_WDRATIO]   = "WDRatio";
  flag_VALUES[flag_WDRATIO]    = 5;

  /* Weight of a function symbol; weight of clause is used to select given */
  flag_STRINGS[flag_FUNCWEIGHT]= "FuncWeight";
  flag_VALUES[flag_FUNCWEIGHT] = 1;

  /* Weight of a variable symbol; weight of clause is used to select given */
  flag_STRINGS[flag_VARWEIGHT] = "VarWeight";
  flag_VALUES[flag_VARWEIGHT]  = 1;

  /* Prefer the selection of clauses with many variable occurrences */
  flag_STRINGS[flag_PREFVAR]   = "PrefVar";
  flag_VALUES[flag_PREFVAR]    = 0;

  /* Flags for selecting the ordering to use */
  flag_STRINGS[flag_ORD]    = "Ordering";
  flag_VALUES[flag_ORD]     = flag_ORDKBO;

  /* CNF flag, if set optimized skolemization is performed */
  flag_STRINGS[flag_CNFOPTSKOLEM]       = "CNFOptSkolem";
  flag_VALUES[flag_CNFOPTSKOLEM]        = 1;

  /* CNF flag, restricts the number of optimized skolemization proof steps */
  flag_STRINGS[flag_CNFPROOFSTEPS]      = "CNFProofSteps";
  flag_VALUES[flag_CNFPROOFSTEPS]       = 100;

  /* CNF flag, if set renaming is performed using ren_HasBenefit */
  flag_STRINGS[flag_CNFRENAMING]        = "CNFRenaming";
  flag_VALUES[flag_CNFRENAMING]         = 1;

  /* CNF flag, if set renaming is printed */
  flag_STRINGS[flag_CNFPRENAMING]       = "CNFPRenaming";
  flag_VALUES[flag_CNFPRENAMING]        = 0;

  /* CNF flag, if set renaming variant subformulae are replaced by the same
   * literal */
  flag_STRINGS[flag_CNFRENMATCH]   	= "CNFRenMatch";
  flag_VALUES[flag_CNFRENMATCH]    	= 1;

  /* CNF flag, if set renaming is performed upon quantied formulae using
   * ren_Quantifers */
  flag_STRINGS[flag_CNFRENQUANT]   	= "CNFRenQuant";
  flag_VALUES[flag_CNFRENQUANT]    	= 0;

  /* CNF flag, if set renaming is performed upon quantied formulae,
   * conjunctions and disjunctions */
  flag_STRINGS[flag_CNFRENOPS]   	= "CNFRenOps";
  flag_VALUES[flag_CNFRENOPS]    	= 0;

  /* CNF flag, if set strong skolemization is performed */
  flag_STRINGS[flag_CNFSTRSKOLEM]       = "CNFStrSkolem";
  flag_VALUES[flag_CNFSTRSKOLEM]        = 1;

  /* dfg2otter flag, if set input options for otter are generated */
  flag_STRINGS[flag_TDFG2OTTEROPTIONS]  = "TDfg2OtterOptions";
  flag_VALUES[flag_TDFG2OTTEROPTIONS]   = 0;

  /* Special flag to indicate syntax of extended modal logic is used and
   * translation is required, set during parsing */
  flag_STRINGS[flag_EML]  = "EML";
  flag_VALUES[flag_EML]   = 0;

  /* Autonomous mode, as yet not functional */
  flag_STRINGS[flag_EMLAUTO]         = "EMLAuto";
  flag_VALUES[flag_EMLAUTO]          = 1;

  /* If set, propositional formulae are converted to relational formulae
   * before they are translated to FOL */
  flag_STRINGS[flag_EML2REL]         = "EMLtoRel";
  flag_VALUES[flag_EML2REL]          = 0;

  /* Special EML flag, determines the translation method used, the
   * default is the relational translation method */
  flag_STRINGS[flag_EMLTR]  = "EMLTranslation";
  flag_VALUES[flag_EMLTR]   = flag_EMLTRREL;

  /* Is set, the nde expressions are omitted in the functional
   * translation */
  flag_STRINGS[flag_EMLKD]  = "EMLKD";
  flag_VALUES[flag_EMLKD]   = 0;

  /* If set, translates  tr(dia(phi),s) = nde(s) /\ exists x tr(phi,[s x])  --
   * nde / quantifier expression,
   * otherwise  tr(dia(phi),s) = exists x nde(s) /\ tr(phi,[s x])  --
   * quantifier / nde expression.  
   * Only meaningful when the flag for the functional translation method
   * is set. */
  flag_STRINGS[flag_EMLFUNCNDEQ]       = "EMLFuncNdeQ";
  flag_VALUES[flag_EMLFUNCNDEQ]        = 1;

  /* If set the functional translation into fluted logic is used */
  flag_STRINGS[flag_EMLFFLUTED]      = "EMLFuncFluted";
  flag_VALUES[flag_EMLFFLUTED]       = 1;

  /* If set sorts for terms are used */
  flag_STRINGS[flag_EMLFFSORTS]      = "EMLFFSorts";
  flag_VALUES[flag_EMLFFSORTS]       = 0;
}


void flag_Free(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  MEMORY:  Frees memory for the default store.
***************************************************************/
{
  flag_FreeStore(flag_DEFAULTVALUES);
}


void flag_Print(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  Prints the values of all flags
  MEMORY:  None.
***************************************************************/
{
  int  i;
  char name[30];
  
  printf("\nlist_of_settings(SPASS).");

  for(i = 0; i < flag_MAXFLAG; i=i+3) {
    sprintf(name,"set_flag(%s,%d).", flag_STRINGS[i], flag_VALUES[i]);
    printf("\n %-30s",name);
    if (i+1 < flag_MAXFLAG) {
      sprintf(name,"set_flag(%s,%d).", flag_STRINGS[i+1], flag_VALUES[i+1]);
      printf(" %-30s",name);
      if (i+2 < flag_MAXFLAG) {
	sprintf(name," set_flag(%s,%d).", flag_STRINGS[i+2], flag_VALUES[i+2]);
	printf(" %-30s",name);
      }
    }
  }

  printf("\nend_of_list.\n");
}


void flag_FPrint(FILE* File)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  Prints the values of all flags
  MEMORY:  None.
***************************************************************/
{
  int  i;
  char name[30];
  
  fprintf(File,"list_of_settings(SPASS).");
  fprintf(File,"{*");

  for(i = 0; i < flag_MAXFLAG; i=i+3) {
    sprintf(name,"set_flag(%s,%d).", flag_STRINGS[i], flag_VALUES[i]);
    fprintf(File,"\n %-30s",name);
    if (i+1 < flag_MAXFLAG) {
      sprintf(name,"set_flag(%s,%d).", flag_STRINGS[i+1], flag_VALUES[i+1]);
      fprintf(File," %-30s",name);
      if (i+2 < flag_MAXFLAG) {
	sprintf(name," set_flag(%s,%d).", flag_STRINGS[i+2], flag_VALUES[i+2]);
	fprintf(File," %-30s",name);
      }
    }
  }
  fprintf(File,"*}");
  fprintf(File,"\nend_of_list.\n");
}


BOOL flag_Lookup(const char* String)
/**************************************************************
  INPUT:   A string <String>.
  RETURNS: TRUE iff <String> is the string of a known flag
  EFFECT:  None.
  MEMORY:  None.
***************************************************************/
{
  int i;

  for(i = 0; i < flag_MAXFLAG; i++)
    if (strcmp(flag_STRINGS[i], String) == 0)
      return TRUE;

  return FALSE;
}


int flag_Id(const char* String)
/**************************************************************
  INPUT:   A string <String>.
  RETURNS: the identification of the flag <String> if it exists
           -1 otherwise
  EFFECT:  None.
  MEMORY:  None.
***************************************************************/
{
  int i;

  for(i = 0; i < flag_MAXFLAG; i++)
    if (strcmp(flag_STRINGS[i], String) == 0)
      return i;

  return -1;
}


void flag_PrintSpecialTranslationMethod()
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  Prints which translation methods are used
  MEMORY:  None.
***************************************************************/
{
  int flag;

  flag = flag_GetFlagValue(flag_EMLTR);
  printf("\n Translation method: %s",flag_GetTranslationMethod(flag));

  if (flag == flag_EMLTRFUNC || flag == flag_EMLTROPTFUNC) {
    if (flag_GetFlagValue(flag_EMLFFLUTED))
      printf("%s"," into fluted logic");
    if (!flag_GetFlagValue(flag_EMLFFSORTS))
      printf("%s"," encoding sorts in predicate names");
  }
} 

