/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                 FLAGS OF SPASS                         * */
/* *                                                        * */
/* *  $Module:   FLAGS                                      * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.74 $                                          * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 17:46:59 $                                      * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: flags.h,v $ */

#ifndef _FLAGS_
#define _FLAGS_

#include "memory.h"

/**************************************************************/
/* Data Structures and Constants                              */
/**************************************************************/

#define flag_CLEAN -5

typedef enum { flag_ORDKBO, flag_ORDRPOS } flag_ORDTYPE;

typedef enum { flag_EMLTRREL, flag_EMLTRFUNC, flag_EMLTROPTFUNC } flag_EMLTRTYPE;

typedef enum { flag_AUTO, flag_STDIN, flag_INTERACTIVE, flag_FLOTTER,
	       flag_SOS,

	       flag_CPARA,      flag_CORDPARA,   flag_CSUPE,
	       flag_CSUPP,

               flag_SPLITS,     flag_MEMORY,     flag_TIMELIMIT,
	       flag_DOCSST,     flag_DOCDST,     flag_DOCPROOF, 
	       flag_FORMULAE2CLAUSE,             flag_DOCFORMULAEONLY,
	       flag_DOCSPLIT,   flag_LOOPS,      flag_PSUB,      
	       flag_PREW,       flag_PCON,       flag_PTAUT,
	       flag_POBV,       flag_PSSI,       flag_PSST,
	       flag_PCLR,       flag_PUNC,
               flag_PDER,       flag_PGIVEN,
	       flag_PKEPT,      flag_PPROBLEM,   flag_PEMPTYCLAUSE,
	       flag_PSTATISTIC, flag_FPMODEL,    flag_FPDFGPROOF,
	       flag_PFLAGS,     flag_POPTSKOLEM, flag_PSTRSKOLEM,
	       flag_PAPPLYDEFS, flag_PDIAGNOSE,

	       flag_SELECT,     flag_RINPUT,     flag_SORTS,
               flag_QUANTEXCH,
	       flag_SATINPUT,   flag_WDRATIO,    flag_FULLRED,
	       flag_FUNCWEIGHT, flag_VARWEIGHT,  flag_PREFVAR,
	       flag_APPLYDEFS,

	       flag_ORD,

	       flag_CNFOPTSKOLEM, flag_CNFSTRSKOLEM, flag_CNFPROOFSTEPS, 
	       flag_CNFRENAMING,  flag_CNFPRENAMING,
	       flag_CNFRENMATCH,  flag_CNFRENQUANT,  flag_CNFRENOPS,
	       
	       flag_IEMS,       flag_ISOR,   
	       flag_IEQR, 
	       flag_IEQF,       flag_IMPM,       flag_ISPR, 
	       flag_IOPM,       flag_ISPM,
	       flag_ISPL,       flag_IGER, 
	       flag_ISHY,       flag_IOHY,
	       flag_IGEF,
	       flag_IUNR,       flag_IBUR,       flag_IDEF,

	       flag_RFREW,      flag_RBREW,                
	       flag_RFCLR,      flag_RBCLR,
	       flag_ROBV,       flag_RUNC,
	       flag_RTAUT,      flag_RSST,       flag_RSSI,
	       flag_RDST,       flag_RFSUB,      flag_RBSUB,      
	       flag_RCON,         
	       
	       flag_TDFG2OTTEROPTIONS,

	       flag_EML, 	flag_EMLAUTO,    flag_EML2REL,

	       flag_EMLTR,      flag_EMLKD,      flag_EMLFUNCNDEQ,

               flag_EMLFFLUTED, 	         flag_EMLFFSORTS,

               flag_MAXFLAG } FLAG_FLAGS;     /* flag_MAXFLAG is a final Dummy */

typedef int *FLAGSTORE;

extern FLAGSTORE flag_VALUES;
extern FLAGSTORE flag_DEFAULTVALUES;
extern char* flag_STRINGS[flag_MAXFLAG];

/**************************************************************/
/* Macros                                                     */
/**************************************************************/

static __inline__ FLAGSTORE flag_NewStore()
{
  return (memory_Malloc(flag_MAXFLAG * sizeof(int)));
}

static __inline__ void flag_FreeStore(FLAGSTORE Store)
{
  memory_Free(Store,flag_MAXFLAG * sizeof(int));
}

static __inline__ void flag_InitStoreByDefaults(FLAGSTORE Store)
{
  int i;
  for(i=0;i<flag_MAXFLAG;i++)
    Store[i] = flag_DEFAULTVALUES[i];  
}

static __inline__ void flag_CleanStore(FLAGSTORE Store)
{
  int i;
  for(i=0;i<flag_MAXFLAG;i++)
    Store[i] = flag_CLEAN;  
}

static __inline__ BOOL flag_StoreValueIsClean(FLAGSTORE Store, FLAG_FLAGS FlagId)
{
  return (Store[FlagId] == flag_CLEAN);
}

static __inline__ void flag_TransferSetFlags(FLAGSTORE Store1, FLAGSTORE Store2)
{
  int i;
  for(i=0;i<flag_MAXFLAG;i++)
    if (!flag_StoreValueIsClean(Store1,i))
      Store2[i] = Store1[i];
}

static __inline__ FLAGSTORE flag_GetInstalledStore()
{
  return flag_VALUES;
}

static __inline__ void flag_InstallStore(FLAGSTORE Store)
{
  flag_VALUES = Store;
}

static __inline__ void flag_InstallDefaultStore()
{
  flag_VALUES = flag_DEFAULTVALUES;
}

static __inline__ BOOL flag_IsInstalledStore(FLAGSTORE Store)
{
  return (flag_VALUES == Store);
}

static __inline__ FLAGSTORE flag_DefaultStore()
{
  return flag_DEFAULTVALUES;
}

static __inline__ void flag_SetFlagValue(FLAG_FLAGS FlagId, int Value)
{
#ifdef DEBUG
  if (flag_VALUES == flag_DEFAULTVALUES) {
    fputs("\n\nAttempt to modify default flag values.\n", stderr);
    misc_DumpCore();
  }
  if (FlagId < 0 || FlagId >= flag_MAXFLAG) {
    fputs("\n\nRange of flags exceeded in flag_Name\n", stderr);
    misc_DumpCore();
  }
#endif
  flag_VALUES[FlagId] = Value;
}

static __inline__ int flag_GetFlagValue(FLAG_FLAGS FlagId)
{
  return flag_VALUES[FlagId];
}

static __inline__ char* flag_GetFlagName(FLAG_FLAGS FlagId)
{
  return flag_STRINGS[FlagId];
}

static __inline__ int flag_StoreGetFlagValue(FLAGSTORE Store, FLAG_FLAGS FlagId)
{
  return Store[FlagId];
}


static __inline__ BOOL flag_ValueIsClean(FLAG_FLAGS FlagId)
{
  return (flag_VALUES[FlagId] == flag_CLEAN);
}

static __inline__ char* flag_Name(FLAG_FLAGS Flag)
{
#ifdef DEBUG
  if (Flag <0 || Flag >= flag_MAXFLAG) {
    fputs("\n\nRange of flags exceeded in flag_Name\n", stderr);
    misc_DumpCore();
  }
#endif
 
  return(flag_STRINGS[Flag]);
}


static __inline__ void flag_ClearInferenceRules()
{
  flag_VALUES[flag_IEMS] = 0;
  flag_VALUES[flag_ISOR] = 0;
  flag_VALUES[flag_IEQR] = 0;
  flag_VALUES[flag_IEQF] = 0;
  flag_VALUES[flag_IMPM] = 0;
  flag_VALUES[flag_ISPR] = 0;
  flag_VALUES[flag_ISPL] = 0;
  flag_VALUES[flag_IGER] = 0;
  flag_VALUES[flag_IGEF] = 0;
  flag_VALUES[flag_IBUR] = 0;
}


static __inline__ void flag_ClearReductionRules()
{
  flag_VALUES[flag_RFREW] = 0;
  flag_VALUES[flag_RBREW] = 0;
  flag_VALUES[flag_RFCLR] = 0;
  flag_VALUES[flag_RBCLR] = 0;
  flag_VALUES[flag_ROBV]  = 0;
  flag_VALUES[flag_RTAUT] = 0;
  flag_VALUES[flag_RSST]  = 0;
  flag_VALUES[flag_RDST]  = 0;
  flag_VALUES[flag_RSSI]  = 0;
  flag_VALUES[flag_RFSUB] = 0;
  flag_VALUES[flag_RBSUB] = 0;
  flag_VALUES[flag_RCON]  = 0;
  flag_VALUES[flag_RUNC]  = 0;
}

static __inline__ void flag_PrintReductionRules()
{
  int i;
  printf("\n Reductions: ");
  for(i=flag_RFREW;i<=flag_RCON;i++)
    if (flag_VALUES[i])
      printf("%s ",flag_STRINGS[i]);
} 

static __inline__ void flag_PrintInferenceRules()
{
  int i;
  printf("\n Inferences: ");
  for(i=flag_IEMS;i<=flag_IBUR;i++)
    if (flag_VALUES[i])
      printf("%s ",flag_STRINGS[i]);
} 

static __inline__ void flag_PrintSpecialTranslationFlags()
{
  int i;
  printf("\n Translation flags: ");
  for(i=flag_EML;i<=flag_EMLFFSORTS;i++)
    if (flag_VALUES[i])
      printf("%s ",flag_STRINGS[i]);
} 

static __inline__ char* flag_GetTranslationMethod(flag)
{
  if (flag == flag_EMLTRREL) 
    return "relational";
  else if (flag == flag_EMLTRFUNC)
    return "functional";
  else if (flag == flag_EMLTROPTFUNC)
    return "optimised functional";
  else 
    return "unknown";
} 


/**************************************************************/
/* Functions                                                  */
/**************************************************************/

void  flag_Init(void);
void  flag_Free(void);
void  flag_Print(void);
void  flag_PrintSpecialTranslationMethod(void);
void  flag_FPrint(FILE*);
BOOL  flag_Lookup(const char*);
int   flag_Id(const char*);

#endif






