/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                       HASHING                          * */
/* *                                                        * */
/* *  $Module:   HASHARRAY                                  * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.57 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/07/20 20:56:37 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: hasharray.c,v $ */

#include "hasharray.h"


/**************************************************************/
/* Functions                                                  */
/**************************************************************/

HASH hsh_Create(void)
/**************************************************************
    RETURNS: A new, empty hasharray
***************************************************************/{
  HASH h;
  NAT  l;
  h = (LIST*) memory_Malloc(sizeof(LIST) * hsh__SIZE);
  for(l=0; l < hsh__SIZE; l++)
    h[l] = list_Nil();
  return h;
}

void hsh_Reset(HASH H)
/**************************************************************
    INPUT:   A hasharray
    EFFECT:  Deletes all information stored in the array but keeps
             the array itself.
             Keys and data items are not deleted !
***************************************************************/
{
  int  i;
  LIST Scan, Pair;
  for (i = 0; i < hsh__SIZE; i++) {
    for(Scan = H[i]; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      Pair = list_Car(Scan);
      list_Delete(list_PairSecond(Pair));
      list_PairFree(Pair);
    }
    list_Delete(H[i]);
    H[i] = list_Nil();
  }
}

void hsh_Delete(HASH H)
/**************************************************************
    INPUT:   A hasharray
    EFFECT:  Deletes all information stored in the array and 
             the array itself.
             Keys and data items are not deleted !
***************************************************************/            
{
  hsh_Reset(H);
  memory_Free(H, sizeof(LIST) * hsh__SIZE);
}

LIST hsh_GetAllEntries(HASH H)
/**************************************************************
    INPUT:   A hasharray
    RETURNS: A new list of all data items stored in the hasharray
***************************************************************/
{
  LIST Scan, Result;
  NAT i;
  Result = list_Nil();
  for(i = 0; i < hsh__SIZE; i++) {
    for(Scan = H[i]; !list_Empty(Scan); Scan = list_Cdr(Scan))
      Result = list_Nconc(Result, list_Copy(list_PairSecond(list_Car(Scan))));
  }
  return Result;
}

void hsh_Check(HASH H)
/**************************************************************
    INPUT:   A hasharray
    EFFECT:  Traverses the whole array and the lists to find dangling pointers
***************************************************************/
{
  LIST Scan, Scan2, Pair;
  NAT i;
  unsigned long Key;
  for (i = 0; i < hsh__SIZE; i++) {
    for(Scan = H[i]; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      Pair = list_Car(Scan);
      Key = (unsigned long) list_PairFirst(Pair);
      for(Scan2 = list_PairSecond(Pair); !list_Empty(Scan2); Scan2 = list_Cdr(Scan2)) {
	POINTER Value;
	char Z;
	Value = list_Car(Scan2);
	Z = * ((char*) Value);
      }
    }
  }
}
