/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                MISCELLANEOUS                           * */
/* *                                                        * */
/* *  $Module:   MISC                                       * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.46 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 17:47:01 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: misc.c,v $ */

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "misc.h"
#include "memory.h"
#include "errno.h"
#include <math.h>

/**************************************************************/
/* Functions                                                  */
/**************************************************************/

void misc_DumpCoreOut(const char* String)
/**************************************************************
  INPUT:   A  string.
  RETURNS: Nothing.
  EFFECT:  Prints <String> and then dumps a core.
***************************************************************/
{
  printf("\n %s \n", String);
  misc_DumpCore();
}


BOOL misc_StringIsNumber(const char* String)
/**************************************************************
  INPUT:   A string.
  RETURNS: TRUE iff the string solely consists of number characters.
  EFFECT:  None.
***************************************************************/
{
  int i;
  
  if (String == NULL || String[0] == '\0')
    return FALSE;
  
  i = 0;

  while (String[i] != '\0')
    if (String[i] >= '0' && String[i] <= '9')
      i++;
    else
      return FALSE;

  return TRUE;  
}


int misc_ReturnValue(void)
{
  return 0;
}


int misc_Max(int a, int b)
{
  if (a > b)
    return a;
  else
    return b;
}

FILE* misc_OpenFile(char* Name, char* Mode)
{
  FILE* File;

  File = fopen(Name,Mode);

  if (File == (FILE*)NULL) {
    printf("\n\tError in opening file %s for %s !\n\n", Name, 
	   (Mode[0] == 'r' ? "reading" :
	    (Mode[0] == 'w' ? "writing" : "i/o operations")));
    misc_Error();
  }  

  return File;
}

void misc_CloseFile(FILE* File, char* Name)
{
  int Result;

  Result = fclose(File);

  if (Result != 0) {
    printf("\n\tError in closing file %s !\n\n", Name);
    misc_Error();
  }  
}


char* misc_StringCopy(const char* String)
/**************************************************************
  INPUT:   A string.
  RETURNS: A copy of the string.
  EFFECT:  The memory for the copy is allocated by the memory module.
***************************************************************/
{
  char *copy;

  copy = memory_Malloc(strlen(String)+1);
  strcpy(copy, String);
  return copy;
}

void  misc_StringFree(char* String)
/**************************************************************
  INPUT:   A string.
  RETURNS: Nothing.
  EFFECT:  Frees the memory used by the string.
***************************************************************/
{
  memory_Free(String, strlen(String)+1);
}

char* misc_IntToString(int Number)
/**************************************************************
  INPUT:   An integer number.
  RETURNS: The number as a string.
  EFFECT:  Memory is allocated for the string.
***************************************************************/
{
  char* result;
  NAT   size = 1;

  if (Number > 9)
    size = (NAT)log10(Number) + 1;
  size++; /* for '\0' */

  result = memory_Malloc(sizeof(char) * size);
  sprintf(result, "%d", Number);
  return result;
}

int misc_StringToInt(const char* String)
/**************************************************************
  INPUT:   A string that should represent a number
           that is a sequence of characters that are digits.
  RETURNS: The value of the number.
  EFFECT:  This function tries to convert the <String> to a
           number. If the number is too big (> INT_MAX),
           too small (< INT_MIN)or the string couldn't be
           converted completely, an error message is written
           to stderr and the program exits.
***************************************************************/
{
  long  number;
  char* end;

  end    = (char*)String;     /* We don't access the string itself */
  errno  = 0;
  number = strtol(String, &end, 10);
  if (errno != 0 || number > INT_MAX || number < INT_MIN) {
    /* Number is too large or buffer can't be converted completely */
    fputs("\nString isn't a number or number to large: ", stderr);
    fputs(String, stderr);
    misc_Error();
  }
  return number;
}
