/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *             SPASS OPTIONS HANDLING                     * */
/* *                                                        * */
/* *  $Module:   OPTIONS                                    * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.63 $                                          * */
/* $State: Exp $                                            * */
/* $Date: 1998/07/20 20:56:38 $                                      * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: options.h,v $ */


#ifndef _OPTIONS_
#define _OPTIONS_

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#include "flags.h"
#include "list.h"


/**************************************************************/
/* Data Structures and Constants                              */
/**************************************************************/

/* option type */
typedef int  OPTTYPE;

#define	opts_NOARGTYPE		0 /* no argument       */
#define opts_REQARGTYPE	        1 /* required argument */
#define opts_OPTARGTYPE         2 /* optional argument */

/* option id   */
typedef int OPTID ;
#define opts_IDFIRST            0 /* option id to start with      */

/* struct for decaration of options */
struct OPTDECL_HELP {
  char*    clname;      /* option name in the command line        */
  OPTTYPE  type;        /* argument type: required, optional, non */
};

typedef struct OPTDECL_HELP OPTDECL;

#define opts_ENDMARKER "--"     /* double hyphen: marks end of all options */
#define opts_DEFAULTOPTARG  "1" /* default value of options with optional arguments */

/**************************************************************
 from the getopt.h file 
 **************************************************************/
struct OPTION
{
  char *name;

  /* has_arg can't be an enum because some compilers complain about
     type mismatches in all the code that assumes it is an int.  */
  int has_arg;
  int *flag;
  int val;
};


/**************************************************************/
/* Functions                                                  */
/**************************************************************/

void   opts_Init();
void   opts_Free();
OPTID  opts_DeclareVector(OPTDECL Decls[]);
OPTID  opts_Declare(char*, OPTTYPE);
BOOL   opts_Read(int argc, char* argv[]);
int    opts_Indicator();

BOOL   opts_IsSet(OPTID);
BOOL   opts_GetIntValueByName(char*, int*);
BOOL   opts_GetValueByName(char*, char**);
BOOL   opts_GetValue(OPTID Id, char** s);
BOOL   opts_GetIntValue(OPTID Id, int* i) ;

char*  opts_ClName(OPTID Id);
OPTID  opts_IdFirst(void);
OPTID  opts_Id(char* ClName);
BOOL   opts_IdIsNull(OPTID Id);

/* specials for SPASS */
void   opts_Transfer(FLAGSTORE);
void   opts_SetFlags(FLAGSTORE);
void   opts_DeclareSPASSFlagsAsOptions();
void   opts_PrintSPASSNames();

#endif



