/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                 SCANNER FOR PROLOG SYNTAX              * */
/* *                                                        * */
/* *  $Module:   PROLOG                                     * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.42 $                                         * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 18:04:59 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: proscanner.l,v $ */

%{

#include <ctype.h>
#include "memory.h"
#include "term.h"
#include "proparser.h"

extern NAT pro_LINENUMBER;  /* defined in proparser.y */

static __inline__ char* pro_StringCopy(void)
{
  char *copy;
  copy = (char*) memory_Malloc(yyleng+1);
  strcpy(copy, yytext);
  return copy;
}

%}

%x comment

%option noyywrap

%%

":-"				return PRO_IF;
"?-"				return PRO_QUESTION;
[0-9]+				{ pro_lval.string = pro_StringCopy();
                                  return PRO_NUM;
                                }
[_A-Z][_a-zA-Z0-9]*		{ pro_lval.string = pro_StringCopy();
				  return PRO_VAR;
				}
[a-z][_a-zA-Z0-9]*		{ pro_lval.string = pro_StringCopy();
                                  return PRO_ID;
                                }
"%"[^\n]*                       /* one-line comment */
"/*"				BEGIN(comment);

<comment>[^*\n]*		/* eat anything that's not a '*' */
<comment>"*"+[^*/\n]*		/* eat up '*'s not followed by '/'s */
<comment>\n			++pro_LINENUMBER;
<comment>"*"+"/"                BEGIN(INITIAL);
[ \t]+				/* ignore whitespace */
"\n"				++pro_LINENUMBER;
[()\[\],\.]			return yytext[0];
!				{ fprintf(stderr, "\nline %d: Cut operator is not allowed\n", pro_LINENUMBER);
				  misc_Error();
                                }
.				{ fputs("\nIllegal character '", stderr);
				  if (isprint((int)yytext[0]))
				    putc(yytext[0], stderr);
				  else
				    fprintf(stderr, "\\x%x", (unsigned int) yytext[0]);
				  fprintf(stderr, "' in line %d\n", pro_LINENUMBER);
                                  misc_Error();
                                }

%%
