/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *         RECURSIVE PATH ORDERING WITH STATUS            * */
/* *                                                        * */
/* *  $Module:   RPOS                                       * */
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.73 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/07/20 20:56:39 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: rpos.h,v $ */

#ifndef _RPOS_
#define _RPOS_

#include "misc.h"
#include "order.h"
#include "pqueue.h"
#include "foldfg.h"

/**************************************************************/
/*  Data Structures                                           */
/**************************************************************/

typedef struct RPOS_TERMHELP{
  TERM                  term;
  struct RPOS_TERMHELP* parent;
  LIST                  children;
  NAT                   ord;         /* Order of ground terms */
  NAT                   size;        /* Number of symbols in term */
  NAT                   count;       /* How many subterms have been visited */
  LIST                  newSubList;  /* Used only for terms with mul symbol */
  LIST                  smaller;     /* List of smaller RPOS_TERMs */
  LIST                  equal;       /* List of equal RPOS_TERMs */
  BOOL                  visited;     /* Used for non-ground comparisons */
} RPOS_TERMNODE, *RPOS_TERM;


/**************************************************************/
/*  Function Prototypes                                       */
/**************************************************************/

BOOL       rpos_Equal(TERM, TERM);
ord_RESULT rpos_GreaterEqual(TERM, TERM);
ord_RESULT rpos_Compare(TERM, TERM);

BOOL       rpos_ContEqual(CONTEXT, TERM, CONTEXT, TERM);
ord_RESULT rpos_ContGreaterEqual(CONTEXT, TERM, CONTEXT, TERM);
ord_RESULT rpos_ContCompare(CONTEXT, TERM, CONTEXT, TERM);

void       rpos_TermPrint(RPOS_TERM);           /* For debugging */
void       rpos_ListPrint(LIST);                /* For debugging */

/**************************************************************/
/*  Inline Functions                                          */
/**************************************************************/

static __inline__ BOOL rpos_Greater(TERM T1, TERM T2)
{
  return ord_IsGreaterThan(rpos_GreaterEqual(T1, T2));
}

static __inline__ BOOL rpos_ContGreater(CONTEXT C1, TERM T1,
					CONTEXT C2, TERM T2)
{
  return ord_IsGreaterThan(rpos_ContGreaterEqual(C1, T1, C2, T2));
}

#endif
