#include "stat.h"

LITSTATS stat_STORE;
POL      stat_POL = BOTH;
int      stat_POS = 6;

static BOOL stats_greater(LITSTAT* Ls1, LITSTAT* Ls2) 
{
  if (stat_POS >= 0) {
    return (Ls1->numocc > Ls2->numocc);
  } else {
    return ((Ls1->numocc > Ls2->numocc) ||
	    ((Ls1->numocc == Ls2->numocc) &&
	     (Ls1->polarity == NEG)));
  };
}

static LITSTAT* find_lit(TERM Lit, LITSTATS List)
{
  LITSTAT *LitStat;
  while (!list_Empty(List)) {
    LitStat = list_Car(List);
    if (LitStat->literal == Lit) {
      return LitStat;
    } else {
      List = list_Cdr(List);
    }
  }
  return (LITSTAT*)NULL;
}

static LITSTAT* create_litstat() {
  return (LITSTAT*)memory_Malloc(sizeof(LITSTAT));
}

static POL polarity_equiv(POL Pol1, POL Pol2)
{
  if ((Pol1 == POS) && (Pol2 == POS)) {
    return POS;
  } else if ((Pol1 == NEG) && (Pol2 == NEG)) {
    return NEG;
  } else {
    return BOTH;
  }
}

static LITSTATS update_litstats(TERM Lit, POL Pol, int Index, CLAUSE Clause, LITSTATS List)
{
  LITSTAT *LitStatP, *InStats;
  BOOL Prop;
  int Inc;
  
  Prop       = (clause_LiteralArity(Clause) == 0);
  if (clause_Length(Clause) == 1) {
    Inc = 1000;
  } else {
    Inc = pow(2,-clause_Length(Clause))*pow(2,3);
  };
  InStats    = find_lit(Lit,List);
  if (InStats == NULL) {
    LitStatP = create_litstat();
    LitStatP->literal       = Lit;
    LitStatP->polarity      = Pol;
    LitStatP->polarities    = Pol;
    LitStatP->numocc        = Inc;
    LitStatP->owningClause  = Clause;
    LitStatP->propositional = Prop;
    LitStatP->litindex      = Index;
    List = list_Nconc(list_List(LitStatP),List);
  } else {
    InStats->polarities = polarity_equiv(InStats->polarities,Pol);
    InStats->numocc = InStats->numocc+Inc;
    if (Prop || (flag_GetFlagValue(flag_SELECT) == 4)) {
       if ((clause_Length(Clause) < clause_Length(InStats->owningClause)) ||
	   ((clause_Length(Clause) == clause_Length(InStats->owningClause)) &&
	    (stat_POS <= 0) &&
	    (Pol == NEG))) { 
	 /* if (clause_Length(Clause) < clause_Length(InStats->owningClause)) { */
	InStats->owningClause  = Clause;
	InStats->litindex      = Index;
	InStats->polarity      = Pol;
	InStats->propositional = Prop;
      }
    }
  }
  return List;
}

CLAUSE stat_compute(LIST List)
{
  LITSTATS LitStats = list_Nil();
  LITSTAT  *LitStatP;
  LITERAL  Literal;
  POL      Pol;
  CLAUSE   Clause;

  int      i;
  
  while (!list_Empty(List)) {
    Clause = list_Car(List);
    if (clause_ContainsVariables(Clause)) {
      /* This clause contains variables and its literals will not
         contribute to the literal statitics */
    } else {
      if (clause_IsEmptyClause(Clause)) {
	return Clause;
      } else {
	if (clause_Length(Clause) < 2) {
	  /* printf("\nFound unit clause ");
	     clause_Print(Clause);
	     printf(" ending computation of statistics"); */
	  return Clause;
	} else {
	  /* printf("\nAdding ");
	     clause_Print(Clause);
	     printf("to the statistics: "); */
	  /* This is a ground clause with at least two literals */
	  if (!clause_ContainsNDLiteral(Clause) &&
	      clause_LiteralArity(Clause) >= 0) {
	    /* The clause does not contain a ND literal, that is, it is not part
	       of a definition and all literals have the same arity */
	    for (i = clause_FirstLitIndex(); i < clause_Length(Clause); i++) {
	      Literal  = clause_GetLiteral(Clause, i);
	      Pol      = clause_LiteralIsNegative(Literal) ? NEG : POS;
	      LitStats = update_litstats(clause_LiteralAtom(Literal),Pol,i,Clause,LitStats);
	    };
	  };
	};
      };
    };
    List = list_Cdr(List);
  };
  LitStats = list_Sort(LitStats,stats_greater);
  stat_STORE = LitStats;
  if (list_Empty(LitStats)) {
    return (CLAUSE)NULL;
  } else {
    LitStatP  = list_Car(LitStats);
    if (LitStatP->polarity == POS) {
      stat_POS--;
    } else {
      stat_POS = 6;
    };
#ifdef DDEBUG
    printf("\nSelected literal [%d] ",stat_POS),
    term_Print(LitStatP->literal);
    printf(" at position %d in clause ",LitStatP->litindex);
    clause_Print(LitStatP->owningClause);
    printf("\noccurring %d times in the usable clause set",LitStatP->numocc);
#endif
    return LitStatP->owningClause;
  };
}



