/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                     SYMBOLS                            * */
/* *                                                        * */
/* *  $Module:   SYMBOL                                     * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.67 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 17:47:08 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: symbol.c,v $ */

#include "symbol.h"

SIGNATURE *symbol_SIGNATURE;
PRECEDENCE symbol_PRECEDENCE,symbol_DEFAULTPRECEDENCE;

SYMBOL symbol_STANDARDVARCOUNTER;
SYMBOL symbol_INDEXVARCOUNTER;

const int symbol_MASK         = 3;  /* 0000011 */
const int symbol_TYPEMASK     = 3;  /* 0000011 */
const int symbol_STATMASK     = 4;  /* 0000100 */
const int symbol_TYPESTATMASK = 7;  /* 0000111 */


const int symbol_TYPEBITS      = 2;
const int symbol_STATBITS      = 1;
const int symbol_TYPESTATBITS  = 3;

const int symbol_SIGTYPES = 4;

const int symbol_TABLEBLANKS = 2;
const int symbol_TABLETYPE  =  8;
const int symbol_TABLEARITY =  9;
const int symbol_TABLESTAT  =  8;
const int symbol_TABLESTRING = 36;

const char symbol_SKFNAME[]="skf";
const char symbol_SKCNAME[]="skc";
const char symbol_SKPNAME[]="SkP";
const char symbol_SKANAME[]="SkC";
const int  symbol_SKLENGTH = 3;

int  symbol_ACTINDEX;
int  symbol_ACTSKOLEMFINDEX;
int  symbol_ACTSKOLEMCINDEX;
int  symbol_ACTSKOLEMAINDEX;
int  symbol_ACTSKOLEMPINDEX;
int  symbol_ORDERING;
LIST symbol_FREEDSYMBOLS;
char *symbol_VARSTRING;

/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *  SYMBOL CREATION                                       * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

int symbol_GetIncreasedOrderingCounter(void)
{
  return symbol_ORDERING++;
}

NAT symbol_MaxStringLength()
/**************************************************************
  INPUT:   None.
  RETURNS: The maximal number of characters needed for a string in the signature.
***************************************************************/
{
  int       Index;
  NAT       Result,Length;
  SIGNATURE Entry;

  Result = 0;
  
  for(Index=1;Index<symbol_ACTINDEX;Index++)
    if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
      Entry  = symbol_Signature(Index);
      Length = Entry->length;
      if (Length > Result)
	Result = Length;
    }

  return Result;
}


static SYMBOL symbol_SignatureCreate(char* String, int Type,
				     int Arity, int Status)
/**************************************************************
  INPUT:   A pointer to a string, a type, the arity and the status
  RETURNS: The symbol containing the passed parameters.
  USAGE:   intern private.
  SUMMARY: Establishes a new symbol in the symbol table and returns the
	   internal representation (pointer and type).
***************************************************************/
{
  SIGNATURE Entry;
  
  if (!symbol_HASSIGNATURE) {
    printf("\n\tsymbol_SignatureCreate: Module was initialized"
	   "with no signature.\n");
    misc_Error();
  }
  
  if (Type < 0 || Type >= symbol_SIGTYPES) {
    printf("\n\tsymbol_SignatureCreate: %d is not a Type\n", Type);
    misc_Error();
  }
  
  if ((int)symbol_ACTINDEX >= symbol__MAXSIGNATURE &&
      list_Empty(symbol_FREEDSYMBOLS)) {
    printf("\n\tsymbol_SignatureCreate: No more symbols available\n");
    misc_Error();
  }
  
  if (strlen(String)>=symbol__SYMBOLMAXLEN) {
    printf("\n\tsymbol_SignatureCreate: String too long\n");
    misc_Error();
  }
    
  Entry           = symbol_GetSignature();
  Entry->weight   = 1;
  symbol_PRECEDENCE[symbol_ACTINDEX] = symbol_ORDERING++;
  Entry->props    = 0;
  Entry->name     = String;
  Entry->length   = strlen(String);
  Entry->arity    = Arity;
  
  if (list_Empty(symbol_FREEDSYMBOLS)) {
    Entry->info = symbol_SignatureSymbol(symbol_ACTINDEX, Type, Status);
    symbol_SetSignature(symbol_ACTINDEX++, Entry);
  }
  else {
    int Index;
    
    Index               = (int)list_Car(symbol_FREEDSYMBOLS);
    symbol_FREEDSYMBOLS = list_PointerDeleteElement(symbol_FREEDSYMBOLS,
						    (POINTER)Index);
    Entry->info = symbol_SignatureSymbol(Index,  Type, Status);
    symbol_SetSignature(Index, Entry);
  }
  
  return Entry->info;
}


SYMBOL symbol_CreateFunction(const char* String, int Arity, int Status)
/**************************************************************
  INPUT:   A pointer to a string and an arity and the status the 
           created function is supposed to have.
  RETURNS: A new symbol for a new function.
  SUMMARY: Creates a new function symbol.
  EFFECTS: Inserts the new function in the symbol table
***************************************************************/
{
  if (Arity==0)
    return symbol_SignatureCreate(
	  strcpy(memory_Malloc(symbol__SYMBOLMAXLEN),String), 
	  symbol_CONSTANT, 
          Arity,
	  Status);
  else
    return symbol_SignatureCreate(
	  strcpy(memory_Malloc(symbol__SYMBOLMAXLEN),String), 
	  symbol_FUNCTION, 
          Arity,
	  Status);
}

SYMBOL symbol_CreateSkolemFunction(int Arity)
/**************************************************************
  INPUT:   The arity of a function to be created.
  RETURNS: A new skolem symbol having the required arity.
  SUMMARY: Creates a new function symbol.
  EFFECTS: Inserts the new function in the symbol table
***************************************************************/
{
  char   newname[10];
  SYMBOL Result;

  if (Arity==0)
    sprintf(newname,"%s%d",symbol_SKCNAME,symbol_ACTSKOLEMCINDEX++);
  else
    sprintf(newname,"%s%d",symbol_SKFNAME,symbol_ACTSKOLEMFINDEX++);

  Result = symbol_CreateFunction(newname, Arity,symbol_STATLEX);

  symbol_AddProperty(Result,SKOLEM);

  return Result;
}


SYMBOL symbol_CreatePredicate(const char* String, int Arity, int Status)
/**************************************************************
  INPUT:   A pointer to a string and an arity and the status
           the created predicate is supposed to have.
  RETURNS: A new symbol for a new predicate.
  SUMMARY: Creates a new predicate symbol.
  EFFECTS: Inserts the new predicate symbol into the symbol table.
***************************************************************/
{ 
  return symbol_SignatureCreate(
	    strcpy(memory_Malloc(symbol__SYMBOLMAXLEN), String),
	    symbol_PREDICATE, 
            Arity,
	    Status);
}

SYMBOL symbol_CreateSkolemPredicate(int Arity)
/**************************************************************
  INPUT:   A pointer to a string and an arity and the status
           the created predicate is supposed to have.
  RETURNS: A new symbol for a new predicate.
  SUMMARY: Creates a new predicate symbol.
  EFFECTS: Inserts the new predicate symbol into the symbol table.
***************************************************************/
{ 
  char   newname[10];

  if (Arity==0)
    sprintf(newname,"%s%d",symbol_SKANAME,symbol_ACTSKOLEMAINDEX++);
  else
    sprintf(newname,"%s%d",symbol_SKPNAME,symbol_ACTSKOLEMPINDEX++);

  return symbol_CreatePredicate(newname, Arity, symbol_STATLEX);
}


SYMBOL symbol_CreateJunctor(const char* String, int Arity, int Status)
/**************************************************************
  INPUT:   A pointer to a string and an arity and the status
           the created junctor is supposed to have.
  RETURNS: A new symbol for a new junctor.
  SUMMARY: Creates a new junctor symbol.
  EFFECTS: Inserts the new junctor symbol into the symbol table.
***************************************************************/
{ 
  return symbol_SignatureCreate(
            strcpy(memory_Malloc(symbol__SYMBOLMAXLEN), String),
	    symbol_JUNCTOR, 
            Arity,
	    Status);
}


BOOL symbol_IsSymbol(SYMBOL Symbol)
/**************************************************************
  INPUT:    A symbol.
  RETURNS: TRUE if the symbols is  a variable or contained
           in the symbol table.
***************************************************************/
{
  return 
    (!symbol_HASSIGNATURE ||
     ((!symbol_Equal(Symbol, symbol__NULL)) &&
      ((symbol_IsVariable(Symbol) && Symbol<symbol_MaxVars()) || 
       (symbol_IsSignature(Symbol) && symbol_Index(Symbol)<symbol_ACTINDEX))));
}


void symbol_Delete(SYMBOL Symbol)
/**************************************************************
  INPUT:   A symbol.
  RETURNS: Nothing.
  SUMMARY: Deletes the symbol from the symbol table and frees its memory.
***************************************************************/
{
#ifdef DEBUG
  if (!symbol_HASSIGNATURE) {
    printf("\nsymbol_Delete: Module was initialized with no signature.\n");
    misc_Error();
  }
  if (!symbol_IsSymbol(Symbol)) {
    printf("symbol_Delete: %d is not a symbol",Symbol);
    misc_Error();
  }
#endif

  if (!symbol_IsVariable(Symbol)) {
    int       Index;
    SIGNATURE Entry;
    
    Index               = symbol_Index(Symbol);
    symbol_FREEDSYMBOLS = list_Cons((POINTER)Index,symbol_FREEDSYMBOLS);
    Entry               = symbol_Signature(Index);
    symbol_FreeSignature(Entry);
  }
}

    
void symbol_FreeAllSymbols(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECTS: Frees all generated symbols
***************************************************************/
{
  if (symbol_HASSIGNATURE) {
    int Index;
    
    for(Index=1;Index<symbol_ACTINDEX;Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
	symbol_FreeSignature(symbol_Signature(Index));
      }
    
    memory_Free(symbol_SIGNATURE, sizeof(SIGNATURE) * symbol__MAXSIGNATURE);
    memory_Free(symbol_DEFAULTPRECEDENCE, sizeof(int) * symbol__MAXSIGNATURE);
  }
  
  memory_Free(symbol_VARSTRING, symbol__SYMBOLMAXLEN);
  list_Delete(symbol_FREEDSYMBOLS);
}

void symbol_Init(BOOL Signature)
/**************************************************************
  INPUT:   A flag indicating whether a signature is used or not.
  RETURNS: None.
  USAGE:   extern public.
  SUMMARY: Initializes the Symbol Module.
  EFFECTS: Initializes global variables.
  CAUTION: MUST BE CALLED BEFORE ANY OTHER Symbol-FUNCTION.
***************************************************************/
{ 
  symbol_ResetSkolemIndex();

  if (Signature) {
    symbol_SIGNATURE          = (SIGNATURE*)memory_Malloc(sizeof(SIGNATURE)   * symbol__MAXSIGNATURE);
    symbol_DEFAULTPRECEDENCE  = (PRECEDENCE)memory_Malloc(sizeof(int) * symbol__MAXSIGNATURE);
    symbol_PRECEDENCE         = symbol_DEFAULTPRECEDENCE;
  }

  symbol_STANDARDVARCOUNTER  = symbol_GetInitialStandardVarCounter();
  symbol_INDEXVARCOUNTER     = symbol_GetInitialIndexVarCounter();
  
  symbol_ACTINDEX            = 1;
  symbol_ORDERING            = 1;
  symbol_VARSTRING           = memory_Malloc(symbol__SYMBOLMAXLEN);
  
  symbol_HASSIGNATURE        = Signature;
  symbol_FREEDSYMBOLS        = list_Nil();
}

void symbol_ReinitGenericNameCounters()
/**************************************************************
  INPUT:   None.
  RETURNS: None.
  USAGE:   extern public.
  EFFECTS: Scans the symbol signature for generic names and if
           there exists some, the corresponding counters are increased.
***************************************************************/
{
  if (symbol_HASSIGNATURE) {
    int          Index,Num;
    SYMBOL       Info;
    SIGNATURE    Entry;
    char         *Name,*Subname;
  
    for(Index=1;Index<symbol_ACTINDEX;Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
	Entry  = symbol_Signature(Index);
	Info   = Entry->info;
	Name   = Entry->name;
	if (strlen(Name)>symbol_SKLENGTH) {
	  Subname = Name + symbol_SKLENGTH;
	
	  switch (symbol_Type(Info)) {
	  
	  case symbol_CONSTANT:
	    if (strncmp(Name,symbol_SKCNAME,symbol_SKLENGTH) == 0 && 
		misc_StringIsNumber(Subname)) {
	      Num = atoi(Subname);
	      if (Num >= symbol_ACTSKOLEMCINDEX)
		symbol_ACTSKOLEMCINDEX = Num + 1;
	    }
	    break;
	  case symbol_FUNCTION:
	    if (strncmp(Name,symbol_SKFNAME,symbol_SKLENGTH) == 0  &&
		misc_StringIsNumber(Subname)) {
	      Num = atoi(Subname);
	      if (Num >= symbol_ACTSKOLEMFINDEX)
		symbol_ACTSKOLEMFINDEX = Num + 1;
	    }
	    break;
	  case symbol_PREDICATE:
	    if (Entry->arity == 0) {
	      if (strncmp(Name,symbol_SKANAME,symbol_SKLENGTH) == 0  &&
		  misc_StringIsNumber(Subname)) {
		Num = atoi(Subname);
		if (Num >= symbol_ACTSKOLEMAINDEX)
		  symbol_ACTSKOLEMAINDEX = Num + 1;
	      }
	    }
	    else {
	      if (strncmp(Name,symbol_SKPNAME,symbol_SKLENGTH) == 0  &&
		  misc_StringIsNumber(Subname)) {
		Num = atoi(Subname);
		if (Num >= symbol_ACTSKOLEMPINDEX)
		  symbol_ACTSKOLEMPINDEX = Num + 1;
	      }
	    }

	  }
	}
      }
  }
}

  
/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *  SYMBOL SEARCHING                                      * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

SYMBOL symbol_Lookup(const char* String)
/**********************************************************
  INPUT:   A pointer to a string.
  RETURNS: If a symbol with name String exists in the signature, the symbol;
           0 otherwise
********************************************************/
{
  if (symbol_HASSIGNATURE) {
    int Index;
    for(Index = 1; Index < symbol_ACTINDEX; Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS, (POINTER)Index)) {
	if (strcmp(String, symbol_Signature(Index)->name) == 0)
	  return symbol_Signature(Index)->info;
      }
  }
  
  return 0;
}


/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *  SPECIALS                                              * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

void symbol_LowerSignature()
/**************************************************************
  INPUT:   None
  RETURNS: void
  EFFECT:  Any predicate or function symbol in the signature that
           starts with a capital letter is prefixed with "ss"
***************************************************************/
{
  int       Index;
  SIGNATURE Entry;
  SYMBOL    Info;
  char*     String;

  for(Index=1;Index<symbol_ACTINDEX;Index++)
    if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
      Entry  = symbol_Signature(Index);
      Info   = Entry->info;
	
      if (symbol_IsPredicate(Info) || symbol_IsFunction(Info)) {
	String = Entry->name;
	if ('A' < String[0] && String[0] < 'Z') {
	  char* New;
	  New = (char *)memory_Malloc(symbol__SYMBOLMAXLEN);
	  strcpy(&(New[2]),String);
	  New[0] = 115;                   /* prefix "s" */
	  New[1] = 115;
	  Entry->name = New;
	  memory_Free(String,symbol__SYMBOLMAXLEN);
	}
      }
    }
}

void symbol_Dump(void)
/**************************************************************
  INPUT:   None
  RETURNS: void
  EFFECT:  Dumps all existing symbols to stdout.
  USAGE:   extern public.
***************************************************************/
{
  if (symbol_HASSIGNATURE) {
    int Index;
    printf("Dump:");
    for(Index = 1; Index < symbol_ACTINDEX; Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
	printf("\n\t %4d:%4d:%4d:%4d:%4d:%s:%d",
	       Index,
	       symbol_Signature(Index)->info,
	       symbol_Signature(Index)->weight,
	       symbol_PRECEDENCE[Index],
	       symbol_Signature(Index)->props,
	       symbol_Signature(Index)->name,
	       symbol_Signature(Index)->length);
      }
  }
}


void symbol_PrintPrecedence(void)
/**************************************************************
  INPUT:   None
  RETURNS: void
  EFFECT:  Prints the current precedence of the symbol module
***************************************************************/
{
  SIGNATURE ActSig;

  if (symbol_HASSIGNATURE) {
    LIST Symbols,Scan;
    int Index;
    Symbols = list_Nil();
    for(Index = 1; Index < symbol_ACTINDEX; Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index) &&
	  (symbol_IsPredicate(symbol_Signature(Index)->info) ||
	   symbol_IsFunction(symbol_Signature(Index)->info))) 
	Symbols = list_Cons((POINTER)(symbol_Signature(Index)->info),Symbols);
    Symbols = list_Sort(Symbols, symbol_PrecedenceGreater);
    for(Scan = Symbols; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      ActSig = symbol_Signature(symbol_Index((SYMBOL)list_Car(Scan)));
      if (list_Empty(list_Cdr(Scan)))
	printf(" %s", ActSig->name);
      else
	printf(" %s >", ActSig->name);
    }
    list_Delete(Symbols);
  }
}

void symbol_FPrintPrecedence(FILE *File)
/**************************************************************
  INPUT:   None
  RETURNS: void
  EFFECT:  Prints the current precedence as a setting command to <File>
***************************************************************/
{
  SIGNATURE ActSig;

  if (symbol_HASSIGNATURE) {
    LIST Symbols,Scan;
    int Index;
    Symbols = list_Nil();
    for(Index = 1; Index < symbol_ACTINDEX; Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index) &&
	  (symbol_IsPredicate(symbol_Signature(Index)->info) ||
	   symbol_IsFunction(symbol_Signature(Index)->info))) 
	Symbols = list_Cons((POINTER)(symbol_Signature(Index)->info),Symbols);
    Symbols = list_Sort(Symbols, symbol_PrecedenceGreater);
    Index   = 0;
    fprintf(File,"set_precedence(");
    for(Scan = Symbols; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      ActSig = symbol_Signature(symbol_Index((SYMBOL)list_Car(Scan)));
      if (list_Empty(list_Cdr(Scan)))
	fprintf(File,"%s", ActSig->name);
      else
	fprintf(File,"%s,", ActSig->name);
      if (Index>15) {
	Index = 0;
	fprintf(File,"\n\t");
      }
      else
	Index++;
    }
    fprintf(File,").");
    list_Delete(Symbols);
  }
}


/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *  SYMBOL OUTPUT	                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

void symbol_Print(SYMBOL Symbol)
/**************************************************************
  INPUT:   A symbol.
  RETURNS: None.
  USAGE:   extern private.
  SUMMARY: Prints a symbol to stdout. 
***************************************************************/
{ 
#ifdef DEBUG
  if (!symbol_IsSymbol(Symbol) ||
      list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)symbol_Index(Symbol))) {
    printf("\n\nIllegal input in symbol_Print: %d\n", Symbol);
    misc_DumpCore();
  }
#endif

  if (symbol_Equal(symbol_Null(),Symbol))
      printf("NULL");
  else if (symbol_IsVariable(Symbol)) {
    SYMBOL NormSymbol;

    NormSymbol = symbol_NormVar(Symbol);

    if (symbol_IsStandardVariable(Symbol)) {
      if (Symbol <= 6)
	/* U, V, W, X, Y, Z */
	sprintf(symbol_VARSTRING,"%c", 84 + NormSymbol);
      else
	/* X1, X2, X3, ... */
	sprintf(symbol_VARSTRING,"X%d", NormSymbol - 6);
    }
    else if (symbol_IsIndexVariable(Symbol))
      /* I1, I2, I3, ... */
      sprintf(symbol_VARSTRING,"I%d", NormSymbol);

    printf("%s",symbol_VARSTRING);
  }
  else if (symbol_HASSIGNATURE)
    printf("%s", symbol_Name(Symbol));
  else
    printf("%d", Symbol);
}


void symbol_FPrint(FILE* File, SYMBOL Symbol)
/**************************************************************
  INPUT:   A file and a symbol.
  RETURNS: None.
  USAGE:   extern private.
  SUMMARY: Prints a symbol to the file. 
***************************************************************/
{ 
#ifdef DEBUG
  if (!symbol_IsSymbol(Symbol)) {
    printf("\n\nIllegal input in symbol_FPrint\n");
    misc_DumpCore();
  }
#endif

  if (symbol_IsVariable(Symbol)) {
    SYMBOL NormSymbol;

    NormSymbol = symbol_NormVar(Symbol);

    if (symbol_IsStandardVariable(Symbol)) {
      if (Symbol <= 6)
	/* U, V, W, X, Y, Z */
	sprintf(symbol_VARSTRING,"%c", 84 + NormSymbol);
      else
	/* X1, X2, X3, ... */
	sprintf(symbol_VARSTRING,"X%d", NormSymbol - 6);
    }
    else if (symbol_IsIndexVariable(Symbol))
      /* I1, I2, I3, ... */
      sprintf(symbol_VARSTRING,"I%d", NormSymbol);

    fprintf(File,"%s",symbol_VARSTRING);
  }
  else if (symbol_HASSIGNATURE)
    fprintf(File, "%s", symbol_Name(Symbol));
  else
    fprintf(File, "%d", Symbol);
}


void symbol_FPrintOtter(FILE* File, SYMBOL Symbol)
/**************************************************************
  INPUT:   A file and a symbol.
  RETURNS: None.
  USAGE:   extern private.
  SUMMARY: Prints a symbol in Otter format to stdout. 
***************************************************************/
{ 
#ifdef DEBUG
  if (!symbol_IsSymbol(Symbol)) {
    printf("\n\nIllegal input in symbol_FPrintOtter\n");
    misc_DumpCore();
  }
#endif

  if (symbol_IsVariable(Symbol)) {
    SYMBOL NormSymbol;

    NormSymbol = symbol_NormVar(Symbol);

    if (symbol_IsStandardVariable(Symbol)) {
      if (Symbol <= 6)
	/* u, v, w, x, y, z */
	sprintf(symbol_VARSTRING,"%c", 116 + NormSymbol);
      else
	/* x1, x2, x3, ... */
	sprintf(symbol_VARSTRING,"x%d", NormSymbol - 6);
    }
    else if (symbol_IsIndexVariable(Symbol))
      /* I1, I2, I3, ... */
      sprintf(symbol_VARSTRING,"I%d", NormSymbol);

    fprintf(File,"%s",symbol_VARSTRING);
  }
  else 
    if (symbol_HASSIGNATURE) {
      if (symbol_IsConstant(Symbol))
	fprintf(File, "c%s", symbol_Name(Symbol));
      else
	if (symbol_IsFunction(Symbol))
	  fprintf(File, "f%s", symbol_Name(Symbol));
	else
	  if (symbol_IsPredicate(Symbol))
	    fprintf(File, "P%s", symbol_Name(Symbol));
	  else
	    fprintf(File, "%s", symbol_Name(Symbol));
    }
    else
      fprintf(File, "%d", Symbol);
}


void symbol_PrintLn(SYMBOL Symbol)
/**************************************************************
  INPUT:   A symbol.
  RETURNS: None.
  USAGE:   extern private.
  SUMMARY: Prints a symbol and a newline to stdout. 
***************************************************************/
{ 
  symbol_Print(Symbol);
  printf("\n");
  
}


void symbol_PrintAll(void)
/**************************************************************
  INPUT:   None.
  RETURNS: None.
  USAGE:   extern public.
  SUMMARY: Prints all symbols to stdout.
***************************************************************/
{
  if (symbol_HASSIGNATURE) {
    int          Index;
    SYMBOL       Info;
    SIGNATURE    Entry;
    unsigned int TypePos, ArityPos, StatPos, StringPos, EndPos, ActPos;

    TypePos   = symbol_TABLEBLANKS;
    ArityPos  = TypePos + symbol_TABLETYPE + 1;
    StatPos   = ArityPos + symbol_TABLEARITY + 1;
    StringPos = StatPos + symbol_TABLESTAT + 1;
    EndPos    = StringPos + symbol_TABLESTRING + 1;

    printf("\n\n");
    misc_PrintChar(symbol_TABLEBLANKS,' '); printf("|");
    misc_PrintChar(symbol_TABLETYPE,'-'); printf("|");
    misc_PrintChar(symbol_TABLEARITY,'-'); printf("|");
    misc_PrintChar(symbol_TABLESTAT,'-'); printf("|");
    misc_PrintChar(symbol_TABLESTRING,'-'); printf("|\n");
    misc_PrintChar(symbol_TABLEBLANKS,' '); printf("|");
    ActPos = TypePos + 1 + printf(" Type");
    misc_PrintChar((ArityPos - ActPos),' '); printf("|");
    ActPos = ArityPos + 1 + printf(" Arity");
    misc_PrintChar((StatPos - ActPos),' '); printf("|");
    ActPos = StatPos + 1 + printf(" Stat");
    misc_PrintChar((StringPos - ActPos),' '); printf("|");
    ActPos = StringPos + 1 + printf(" String");
    misc_PrintChar((EndPos - ActPos),' '); printf("|\n");
    misc_PrintChar(symbol_TABLEBLANKS,' '); printf("|");
    misc_PrintChar(symbol_TABLETYPE,'-'); printf("|");
    misc_PrintChar(symbol_TABLEARITY,'-'); printf("|");
    misc_PrintChar(symbol_TABLESTAT, '-'); printf("|");
    misc_PrintChar(symbol_TABLESTRING,'-'); printf("|\n");

  
    for(Index=1;Index<symbol_ACTINDEX;Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
	misc_PrintChar(symbol_TABLEBLANKS,' '); printf("|");
	ActPos = TypePos + 1;
	Entry  = symbol_Signature(Index);
	Info   = Entry->info;
	
	switch (symbol_Type(Info)) {
	  
	case symbol_CONSTANT:
	  ActPos = ActPos + printf(" Con");
	  misc_PrintChar((ArityPos - ActPos),' '); printf("|");
	  ActPos = ArityPos + 1 + 
	      printf(" %3d", Entry->arity);
	  misc_PrintChar((StatPos - ActPos),' ');
	  printf("|");
	  ActPos = StatPos + 1 +
	      printf("   ");
	  misc_PrintChar((StringPos - ActPos),' ');
	  break;
	  
	case symbol_FUNCTION:
	  ActPos = ActPos + printf(" Fun");
	  misc_PrintChar((ArityPos - ActPos),' '); printf("|");
	  ActPos = ArityPos + 1 +
	      printf(" %3d", Entry->arity);
	  misc_PrintChar((StatPos - ActPos),' ');
	  printf("|");
	  switch (symbol_Stat(Info)) {
	  case symbol_STATLEX:
	    ActPos = StatPos + 1 +
		printf(" Lex");
	    break;
	  case symbol_STATMUL:
	    ActPos = StatPos + 1 +
		printf(" Mul");
	    break;
	  }
	  misc_PrintChar((StringPos - ActPos),' ');
	  break;
	  
	case symbol_PREDICATE:
	  ActPos = ActPos + printf(" Pre");
	  misc_PrintChar((ArityPos - ActPos),' '); printf("|");
	  ActPos = ArityPos + 1 +
	      printf(" %3d", Entry->arity);
	  misc_PrintChar((StatPos - ActPos),' ');
	  printf("|");
	  ActPos = StatPos + 1 +
	      printf("   ");
	  misc_PrintChar((StringPos - ActPos),' ');
	  break;
	  
	case symbol_JUNCTOR:
	  ActPos = ActPos + printf(" Jun");
	  misc_PrintChar((ArityPos - ActPos),' '); printf("|");
	  ActPos = ArityPos + 1 +
	      printf(" %3d", Entry->arity);
	  misc_PrintChar((StatPos - ActPos),' ');
	  printf("|");
	  ActPos = StatPos + 1 +
	      printf("   ");
	  misc_PrintChar((StringPos - ActPos),' ');
	  break;
	}
	
	printf("|");
	ActPos = StringPos + 1 + 
	    printf(" %s", symbol_Signature(Index)->name);
	misc_PrintChar((EndPos - ActPos),' ');
	printf("|\n");
      }
    misc_PrintChar(symbol_TABLEBLANKS,' '); printf("|");
    misc_PrintChar(symbol_TABLETYPE,'-'); printf("|");
    misc_PrintChar(symbol_TABLEARITY,'-'); printf("|");
    misc_PrintChar(symbol_TABLESTAT,'-'); printf("|");
    misc_PrintChar(symbol_TABLESTRING,'-'); printf("|\n");
  }
}

