/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                     TABLEAUS                           * */
/* *                                                        * */
/* *  $Module:   TABLEAU                                    * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.77 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 18:01:43 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

  
/* $RCSfile: tableau.h,v $ */

#ifndef _TABLEAU_
#define _TABLEAU_
  

#include "list.h"   
#include "clause.h" 

typedef struct  TABLEAU_HELP {
  LIST   Clauses;               /* all clauses generated on the split level */
  CLAUSE SplitClause;           /* this levels split clause */

  CLAUSE LeftSplitClause;       /* first clause generated by split  */
  LIST   RightSplitClauses;     /* other clauses generated by split */

  struct TABLEAU_HELP* LeftBranch;  /* branch corresponding to first split clause  */
  struct TABLEAU_HELP* RightBranch; /* branch corresponding to other split clauses */

#ifdef USE_CG
  int  Label;
#endif

} TABLEAU_NODE;

typedef  TABLEAU_NODE *TABLEAU;

typedef struct TABPATH_HELP {
  TABLEAU* Path;
  int Length;     
  int MaxLength;  
  
} TABPATH_NODE;

typedef  TABPATH_NODE* TABPATH;

/* tableau output formats */
typedef enum { DAVINCI, XVCG } GRAPHFORMAT;

TABPATH tab_PathCreate(int, TABLEAU);
void    tab_PathDelete(TABPATH);
TABPATH tab_PathInit(TABLEAU);
BOOL    tab_PathContainsClause(TABPATH, CLAUSE);
void    tab_AddClauseOnItsLevel(CLAUSE, TABPATH);
void    tab_Print(TABLEAU);
void    tab_AddSplitAtCursor(TABPATH, BOOL);
BOOL    tab_IsClosed(TABLEAU);
TABLEAU tab_PruneClosedBranches(TABLEAU, LIST*);
int     tab_Depth(TABLEAU);
void    tab_SetSplitLevels(TABLEAU T);
TABLEAU tab_RemoveIncompleteSplits(TABLEAU, LIST*);
void    tab_PathDelete(TABPATH);
void    tab_Delete(TABLEAU);
void    tab_ToClauseList(TABLEAU, LIST*);
CLAUSE  tab_FirstEmptyClause(TABLEAU);
void    tab_GetEarliestEmptyClauses(TABLEAU, LIST*);
void    tab_WriteTableau(TABLEAU, char*,GRAPHFORMAT);
void    tab_CheckEmpties(TABLEAU);
void    tab_GetAllEmptyClauses(TABLEAU, LIST*);

#ifdef USE_CG
TABLEAU tab_FindLabel(TABLEAU, int);
void    tab_LabelNodes(TABLEAU);
#endif

/* macros */
static __inline__ int  tab_PathLength(TABPATH TabPath)
{
  return(TabPath->Length);
}

static __inline__ TABLEAU tab_PathNthNode(TABPATH TabPath, int n)
{
#ifdef DEBUG
  if (n > TabPath->MaxLength) {
    printf("\nerror in tab_PathNthNode: path length is %d,\nnode %d was required.\n", 
	   tab_PathLength(TabPath),n);
    misc_Error(); }
#endif  

  return(TabPath->Path[n]);
}

static __inline__ TABPATH tab_PathPush(TABLEAU Tab, TABPATH TabPath)
{
  TabPath->Length++;
  TabPath->Path[TabPath->Length] = Tab;
  
#ifdef DEBUG
  if (TabPath->Length > TabPath->MaxLength) {
    printf("\ntableau: max path length %d exceeded\n", TabPath->MaxLength);
    misc_Error(); }
#endif
  
  return(TabPath);
}

static __inline__ TABLEAU tab_EmptyTableau()
{
 return((TABLEAU)NULL);
}

static __inline__ TABPATH tab_PathPop(TABPATH TabPath)

{
  TabPath->Path[TabPath->Length] = tab_EmptyTableau();
  TabPath->Length--;

#ifdef DEBUG
  if (TabPath->Length < 0) {
    printf("\ntableau: popping from empty path.\n");
    misc_Error(); }
#endif
  
  return(TabPath);
}

static __inline__ BOOL tab_PathEmpty(TABPATH P)
{
  return(tab_PathLength(P) == 0);
}

static __inline__ TABLEAU tab_CreateNode()
{
  TABLEAU Node;

  Node = (TABLEAU)memory_Malloc(sizeof(TABLEAU_NODE));
  Node->RightBranch = (TABLEAU)NULL;
  Node->LeftBranch  = (TABLEAU)NULL;
  Node->SplitClause = (CLAUSE)NULL;
  Node->LeftSplitClause   = (CLAUSE)NULL;
  Node->RightSplitClauses = list_Nil();

  Node->Clauses     = list_Nil();
 
#ifdef USE_LABEL
  Node->Label          = 0;
#endif
  
  return(Node);
}

static __inline__ TABPATH tab_PathPrefix(int Level, TABPATH TabPath)
{
  TabPath->Length = Level;
  return(TabPath);
}

static __inline__ TABLEAU tab_PathTop(TABPATH Path)
{
  return(tab_PathNthNode(Path, tab_PathLength(Path)));
}


static __inline__ BOOL tab_IsEmpty(TABLEAU Tab)
{
 return(Tab == tab_EmptyTableau());
}

static __inline__ TABLEAU tab_RightBranch(TABLEAU Tab)
{
 return(Tab->RightBranch);
}

static __inline__ TABLEAU tab_LeftBranch(TABLEAU Tab)
{
 return(Tab->LeftBranch);
}

static __inline__ void tab_SetRightBranch(TABLEAU Tab, TABLEAU SubTab)
{
  Tab->RightBranch = SubTab;
} 

static __inline__ void tab_SetLeftBranch(TABLEAU Tab, TABLEAU SubTab)
{
  Tab->LeftBranch = SubTab;
} 

static __inline__ BOOL tab_RightBranchIsEmpty(TABLEAU Tab)
{
  return(Tab->RightBranch == tab_EmptyTableau());
}

static __inline__ BOOL tab_LeftBranchIsEmpty(TABLEAU Tab)
{
  return(Tab->LeftBranch == tab_EmptyTableau());
}

static __inline__ CLAUSE tab_SplitClause(TABLEAU Tab)
{
  return(Tab->SplitClause);
}

static __inline__ void tab_SetSplitClause(TABLEAU Tab, CLAUSE C)
{
  Tab->SplitClause = C;
}

static __inline__ BOOL tab_HasSplit(TABLEAU T)
{
  return(tab_SplitClause(T) != clause_Null());
}

static __inline__ void tab_AddClause(CLAUSE C,TABLEAU T)
{
  T->Clauses = list_Cons(C,T->Clauses);
}

static __inline__ LIST tab_Clauses(TABLEAU T)
{
  return(T->Clauses);
}

static __inline__ void tab_SetClauses(TABLEAU T, LIST Clauses)
{
  T->Clauses = Clauses;
}

static __inline__ CLAUSE tab_LeftSplitClause(TABLEAU T)
{
  return(T->LeftSplitClause);
}

static __inline__ void tab_SetLeftSplitClause(TABLEAU T, CLAUSE C)
{
  T->LeftSplitClause = C;
}


static __inline__ LIST tab_RightSplitClauses(TABLEAU T)
{
  return(T->RightSplitClauses);
}


static __inline__ void tab_SetRightSplitClauses(TABLEAU T, LIST L)
{
  T->RightSplitClauses = L;
}

static __inline__ void tab_AddRightSplitClause(TABLEAU T, CLAUSE C)
{
  T->RightSplitClauses = list_Cons(C, T->RightSplitClauses);
}

static __inline__ BOOL tab_IsLeaf(TABLEAU T)
{
  return(tab_RightBranchIsEmpty(T) && tab_LeftBranchIsEmpty(T));
}

#endif
