/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *              TOP MODULE OF SPASS                       * */
/* *                                                        * */
/* *  $Module:   TOP                                        * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.91 $                                          * */
/* $State: Exp $                                            * */
/* $Date: 1998/11/30 20:32:18 $                                      * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: top.c,v $ */

/*** MAINLOOP *************************************************/


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "dfg.h"
#include "defs.h"
#include "ia.h"
#include "rules-inf.h"
#include "rules-sort.h"
#include "rules-split.h"
#include "rules-red.h"
#include "analyze.h"
#include "clock.h"
#include "clause.h"
#include "options.h"
#include "context.h"
#include "cnf.h"
#include "approx.h"
#include "search.h"
#include "hasharray.h"
#include "eml.h"
#include <errno.h>

#if (defined(SPASS_SIGNALS))
#include <signal.h>

void top_SigHandler(int Signal)
{
    if (Signal == SIGSEGV || Signal == SIGBUS) {
        printf("\nSPASS is going to crash. This is probably caused by a\n");
        printf("bug in SPASS.  Please send a copy of the input file(s) to\n");
        printf("weidenb@mpi-sb.mpg.de, and someone will try to fix the problem.\n\n");
    }

    if (opts_IsSet(top_RemoveFileOptId))
      remove(top_InputFile);
}
#endif

/**************************************************************/
/* Types and Variables                                        */
/**************************************************************/

char *top_InputFile;

static OPTID top_RemoveFileOptId;

typedef enum {top_PROOF, top_COMPLETION, top_RESOURCE} top_SEARCHRESULT;

/**************************************************************/
/* Clause Selection Functions                                 */
/**************************************************************/




static CLAUSE top_SelectClauseDepth(LIST List)
{
  CLAUSE Result;
  NAT    Vars,NewVars,Weight,Depth,NewDepth;

  Result = (CLAUSE)list_Car(List);
  Depth  = clause_Depth(Result);
  Weight = clause_Weight(Result);
  Vars   = clause_NumberOfVarOccs(Result);
  List   = list_Cdr(List);

  while (!list_Empty(List)) {
    NewDepth =  clause_Depth(list_Car(List));
    if (NewDepth <= Depth) {
      if (NewDepth < Depth || clause_Weight(list_Car(List)) < Weight) {
	Depth  = NewDepth;
	Result = (CLAUSE)list_Car(List);
	Weight = clause_Weight(Result);
	Vars   = clause_NumberOfVarOccs(list_Car(List));
      }
      else
	if (clause_Weight(list_Car(List)) == Weight) {
	  NewVars = clause_NumberOfVarOccs(list_Car(List));
	  if (flag_GetFlagValue(flag_PREFVAR)) {
	    if (Vars < NewVars) {
	      Depth  = NewDepth;
	      Result = (CLAUSE)list_Car(List);
	      Weight = clause_Weight(Result);
	      Vars   = NewVars;
	    }
	  }
	  else
	    if (Vars > NewVars) {
	      Depth  = NewDepth;
	      Result = (CLAUSE)list_Car(List);
	      Weight = clause_Weight(Result);
	      Vars   = NewVars;
	    }
	}
    }
    List = list_Cdr(List);
  }

  return Result;
}

static CLAUSE top_NoSelectedLit(LIST List)
{
  CLAUSE Result,Clause;
  int    CArity;
  
  Result = (CLAUSE)NULL;

  while (!list_Empty(List)) {
    Clause = list_Car(List);
    CArity = clause_LiteralArity(Clause);

    if (((CArity > 0) &&
         !clause_HasSelectedLiteral(Clause)) ||
	((CArity == 0) &&
	 (clause_Length(Clause) <= 1))) {
      if (Result == (CLAUSE)NULL) {
	Result = Clause;
      } else {
	if (clause_Length(Result) > clause_Length(Clause)) {
	  Result = Clause;
	}
      }
    };
    List = list_Cdr(List);
  };
  return(Result);
}


static CLAUSE top_SelectNDR(LIST List)
{
  CLAUSE Result,Clause;
  BOOL   Prop, Grd, Monadic, NonMonadic, Dyadic, NonDyadic;
  int    Changed;
  
  Result = (CLAUSE)NULL;

  while (!list_Empty(List)) {
    Prop       = FALSE;
    Grd        = FALSE;
    Monadic    = FALSE;
    NonMonadic = FALSE;
    Clause = list_Car(List);
    Changed = clause_ContainsFolAtom(Clause,&Prop,&Grd,&Monadic,&NonMonadic,&Dyadic,&NonDyadic);

    if (Prop && Monadic && !NonMonadic) {
      Result = Clause;
      return(Result);
    };
    List = list_Cdr(List);
  };
  return(Result);
}

static CLAUSE top_SelectMinimalWeightClause(LIST List)
{
  CLAUSE Result;
  NAT    Vars,NewVars,Weight;

  Result = (CLAUSE)list_Car(List);
  Weight = clause_Weight(Result);
  Vars   = clause_NumberOfVarOccs(Result);
  List   = list_Cdr(List);

  while (!list_Empty(List)) {
    if (clause_Weight(list_Car(List)) == Weight) {
      NewVars = clause_NumberOfVarOccs(list_Car(List));
      if (flag_GetFlagValue(flag_PREFVAR)) {
	if (Vars < NewVars) {
	  Result = (CLAUSE)list_Car(List);
	  Weight = clause_Weight(Result);
	  Vars   = NewVars;
	}
      }
      else
	if (Vars > NewVars) {
	  Result = (CLAUSE)list_Car(List);
	  Weight = clause_Weight(Result);
	  Vars   = NewVars;
	}
    }
    else
      return Result;
    List = list_Cdr(List);
  }
  return Result;
}

static CLAUSE top_SelectMinimalWeightDefClause(LIST List)
{
  CLAUSE Result;
  NAT    Vars,NewVars,Weight;

  Result = (CLAUSE)NULL;
  Vars   = 0;
  Weight = 0;

  while (!list_Empty(List)) {
    if (((clause_HasSelectedLiteral(list_Car(List))) &&
	 (clause_ContainsVariables(list_Car(List)))) ||
	(clause_Length(list_Car(List)) < 2)) {
      if (Result == (CLAUSE)NULL) {
	Result = (CLAUSE)list_Car(List);
	Weight = clause_Weight(Result);
	Vars   = clause_NumberOfVarOccs(Result);
      } else {
	if (clause_Weight(list_Car(List)) == Weight) {
	  NewVars = clause_NumberOfVarOccs(list_Car(List));
	  if (flag_GetFlagValue(flag_PREFVAR)) {
	    if (Vars < NewVars) {
	      Result = (CLAUSE)list_Car(List);
	      Weight = clause_Weight(Result);
	      Vars   = NewVars;
	    }
	  }
	  else
	    if (Vars > NewVars) {
	      Result = (CLAUSE)list_Car(List);
	      Weight = clause_Weight(Result);
	      Vars   = NewVars;
	    }
	}
	else
	  return Result;
      }
    }
    List = list_Cdr(List);
  }
  return Result;
}


/*static CLAUSE top_SelectClauseDepth(LIST List)
{
  CLAUSE Result;
  int    Min, Depth;

  Result = (CLAUSE)list_Car(List);
  Depth  = clause_Depth(Result);
  Min    = Depth * clause_Weight(Result);
  List   = list_Cdr(List);

  if (Depth == 0)
    return Result;

  while (!list_Empty(List)) {
    Depth =  clause_Depth(list_Car(List));
    if (Min > Depth * clause_Weight(list_Car(List))) {
      Result = list_Car(List);
      if (Depth == 0)
	return Result;
      Min    = clause_Depth(Result) * clause_Weight(Result);
    }
    List = list_Cdr(List);
  }

  return Result;
}*/

PROOFSEARCH top_ProofSearch(PROOFSEARCH Search, LIST ProblemClauses, FLAGSTORE InputStore)
{
  LIST   Scan,Usables,WorkedOffs,SubsortClauses,EmptyClauses,Derivables,UsedEmptyClauses;
  CLAUSE GivenClause,SplitClause,HighestLevelEmptyClause;
  int    Counter,Splits;

  HighestLevelEmptyClause = (CLAUSE)NULL;
  UsedEmptyClauses        = list_Nil();
  EmptyClauses            = list_Nil();
  Usables                 = list_Nil();
  WorkedOffs              = list_Nil();
  SubsortClauses          = list_Nil();
  Derivables              = list_Nil();
  Counter                 = 1;

  clock_InitCounter(clock_REDUCTION);
  clock_InitCounter(clock_BACKTRACK);
  clock_InitCounter(clock_INFERENCE);

  ana_AnalyzeProblem(ProblemClauses); /* Important ! Recomputes Weight ! */
  if (flag_GetFlagValue(flag_PDIAGNOSE))
    ana_AnalyzeClauses(ProblemClauses); /* Diagnostics (experimental) */

  if (flag_GetFlagValue(flag_AUTO)) {
    symbol_InstallPrecedence(prfs_Precedence(Search));
    flag_InstallStore(prfs_Store(Search));
    ana_AutoConfiguration(ProblemClauses);
    flag_TransferSetFlags(InputStore,prfs_Store(Search)); /* File and input flags have higher precedence */
  }
  
  list_Apply(clause_OrientAndReInit,ProblemClauses);

  if (flag_GetFlagValue(flag_PPROBLEM)) {
    printf("\n--------------------------SPASS-START-----------------------------");
    printf("\nInput Problem:\n");
    clause_ListPrint(ProblemClauses);
    ana_Print();
  }

  if (flag_GetFlagValue(flag_SORTS) != 0) {
    BOOL Strong;
    Strong     = (flag_GetFlagValue(flag_SORTS) == 2);
    for(Scan=ProblemClauses;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      GivenClause = (CLAUSE)list_Car(Scan);
      clause_SetSortConstraint(GivenClause,Strong);
    }
  }

  if (flag_GetFlagValue(flag_RINPUT)) {
    clock_StartCounter(clock_REDUCTION);
    EmptyClauses = red_ReduceInput(Search, ProblemClauses, flag_GetFlagValue(flag_DOCPROOF));
    clock_StopAddPassedTime(clock_REDUCTION);
    if (list_Empty(EmptyClauses) && flag_GetFlagValue(flag_SATINPUT))
      EmptyClauses = red_SatInput(Search, flag_GetFlagValue(flag_DOCPROOF));
  }
  else {
    for(Scan=ProblemClauses;!list_Empty(Scan);Scan=list_Cdr(Scan))
      prfs_InsertUsableClause(Search, list_Car(Scan));
  }
  Derivables = list_Nil();

  if (ana_SortRestrictions()) {
    if (flag_GetFlagValue(flag_RSST))
      prfs_SetStaticSortTheory(Search,sort_ApproxStaticSortTheory(prfs_UsableClauses(Search)));
    if (flag_GetFlagValue(flag_RDST))
      prfs_SetApproximatedDynamicSortTheory(Search,
					    sort_ApproxDynamicSortTheory(prfs_UsableClauses(Search)));
    prfs_SetDynamicSortTheory(Search,sort_TheoryCreate());
  }

  if (flag_GetFlagValue(flag_SOS)) {
    Derivables = list_Copy(prfs_UsableClauses(Search));
    for (Scan=Derivables;!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (!clause_GetFlag(list_Car(Scan), CONCLAUSE)) {
	clause_SelectLiteral(list_Car(Scan));
	/* prfs_MoveUsableWorkedOff(Search,list_Car(Scan)); */
      };
    list_Delete(Derivables);
  }

  if (flag_GetFlagValue(flag_PPROBLEM)) {
    printf("\nProcessed Problem:\n");
    printf("\nWorked Off Clauses:\n");
    clause_ListPrint(prfs_WorkedOffClauses(Search));
    printf("\nUsable Clauses:\n");
    clause_ListPrint(prfs_UsableClauses(Search));
  }

  Derivables  = list_Nil();
  SplitClause = (CLAUSE)NULL;
  Splits      = flag_GetFlagValue(flag_SPLITS);

  while ((list_Empty(EmptyClauses) || !prfs_SplitStackEmpty(Search)) && 
	 prfs_Loops(Search) != 0 &&
	 (!list_Empty(prfs_UsableClauses(Search)) || !list_Empty(EmptyClauses)) &&
	 (flag_GetFlagValue(flag_TIMELIMIT) < 0 || 
	  flag_GetFlagValue(flag_TIMELIMIT) > clock_GetSeconds(clock_OVERALL))) {
      
    if (!list_Empty(EmptyClauses)) {

      clock_StartCounter(clock_BACKTRACK);
      Derivables = split_Backtrack(Search,HighestLevelEmptyClause);
      clock_StopAddPassedTime(clock_BACKTRACK);
      prfs_IncBacktrackedClauses(Search,list_Length(Derivables));
      
      if (prfs_SplitStackEmpty(Search)) {
	Derivables = list_Nconc(EmptyClauses,Derivables);
      }
      else {
	LIST l;
	for(l = EmptyClauses; l != list_Nil(); l = list_Cdr(l))
	  if ((CLAUSE) list_Car(l) != HighestLevelEmptyClause)
	    clause_Delete((CLAUSE) list_Car(l));
	prfs_InsertDocProofClause(Search, HighestLevelEmptyClause);
	/* Keep HighestLevelEmptyClause for finding the terms required
           for the proof. */
	if (flag_GetFlagValue(flag_DOCPROOF))
	  UsedEmptyClauses = list_Cons(HighestLevelEmptyClause, UsedEmptyClauses);
	list_Delete(EmptyClauses);
	if (flag_GetFlagValue(flag_DOCSPLIT))
	  printf("\n\t Split Backtracking level %d:",prfs_ValidLevel(Search));
      }
      EmptyClauses            = list_Nil();
      HighestLevelEmptyClause = (CLAUSE)NULL;
    }
    else {			/* not (EMPTY_CLAUSE). */

#ifdef DEBUG
      if (!prfs_Check(Search)) {
	fputs("\n\nIllegal state of search in mainloop.\n", stderr);
	misc_DumpCore();
      }
      if (!ana_Equations())
	red_CheckSplitSubsumptionCondition(Search);
#endif 

      /* OLD CLAUSE SELCTION SCHEME
      if ((SplitClause == (CLAUSE)NULL) && (Counter%flag_GetFlagValue(flag_WDRATIO)) == 0)
        GivenClause = top_SelectClauseDepth(prfs_UsableClauses(Search));
      else
	GivenClause = top_SelectMinimalWeightClause(prfs_UsableClauses(Search));
      */
      if (flag_GetFlagValue(flag_SELECT) == 4) {
	GivenClause = top_SelectMinimalWeightDefClause(prfs_UsableClauses(Search));
      } else {
	if (flag_GetFlagValue(flag_SELECT) == 3) {
	  GivenClause = top_NoSelectedLit(prfs_UsableClauses(Search));
	} else {
	  GivenClause = (CLAUSE)NULL;
	};
      };
      if (GivenClause == (CLAUSE)NULL) {
	GivenClause = stat_compute(prfs_UsableClauses(Search));
	if (GivenClause == (CLAUSE)NULL) {
	  GivenClause = top_SelectNDR(prfs_UsableClauses(Search));
	  if (GivenClause == (CLAUSE)NULL) {
	    if (!list_Empty(prfs_UsableClauses(Search))) {
#ifdef DDEBUG
	      printf("\nFalling back to normal selection\n");
#endif
	      if ((SplitClause == (CLAUSE)NULL) && (Counter%flag_GetFlagValue(flag_WDRATIO)) == 0)
		GivenClause = top_SelectClauseDepth(prfs_UsableClauses(Search));
	      else
		GivenClause = top_SelectMinimalWeightClause(prfs_UsableClauses(Search));
	    } else {
#ifdef DDEBUG
	      printf("\nNot selecting a clause!");
#endif
	    }
	  } else {
#ifdef DDEBUG
	    printf("\nUsed top_SelectNDR to select clause");
	    clause_Print(GivenClause);
#endif
	  };
	} else {
#ifdef DDEBUG
	  printf("\nUsed stat_compute to select clause");
	  clause_Print(GivenClause);
#endif
	};
      } else {
#ifdef DDEBUG
	  printf("\nUsed top_NoSelectedLit to select clause");
	  clause_Print(GivenClause);
#endif
      };

      Counter++;
      
      if (flag_GetFlagValue(flag_FULLRED)) 
	Derivables = list_Nil();
      else {
	clock_StartCounter(clock_REDUCTION);
	Derivables = red_BackReduction(Search,GivenClause,flag_GetFlagValue(flag_DOCPROOF),red_ALL);
	clock_StopAddPassedTime(clock_REDUCTION);
      }
      
      
      if (Splits != 0 && 
	  !clause_ContainsVariables(GivenClause) &&
	  /* Additional condition: No ND literal in the clause 
	  !clause_ContainsNDLiteral(GivenClause) &&  */
	  /* Additional condition: All literal have the same arity, that is, GivenClause is not a definition */
	  (clause_LiteralArity(GivenClause) >= 0) && 
	  (SplitClause = prfs_PerformSplitting(Search,GivenClause)) != (CLAUSE)NULL) {
	Splits--;
	Derivables = list_Cons(SplitClause,Derivables);
      }
      else { 
	clause_SelectLiteral(GivenClause);
	clause_SetSpecialFlags(GivenClause,ana_SortDecreasing());
	prfs_MoveUsableWorkedOff(Search,GivenClause);
	clock_StartCounter(clock_INFERENCE);
	Derivables = list_Nconc(Derivables,inf_DerivableClauses(Search,GivenClause));
	clock_StopAddPassedTime(clock_INFERENCE);
      }

      if (flag_GetFlagValue(flag_PGIVEN)) {
	printf("\n\tGiven clause[top]: ");
	clause_Print(GivenClause); 
	fflush(stdout);
      }            

      prfs_IncDerivedClauses(Search,list_Length(Derivables));
    }

    if (flag_GetFlagValue(flag_PDER))
      for(Scan=Derivables;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	printf("\nDerived[top]: "); 
	clause_SelectLiteral(list_Car(Scan));
	clause_Print(list_Car(Scan));
      }

    Derivables = split_ExtractEmptyClauses(Derivables,&EmptyClauses);

    if (flag_GetFlagValue(flag_FULLRED)) { 
      clock_StartCounter(clock_REDUCTION);
      EmptyClauses = list_Nconc(EmptyClauses,
				red_CompleteReductionOnDerivedClauses(Search, Derivables, 
								      flag_GetFlagValue(flag_DOCPROOF)));
      clock_StopAddPassedTime(clock_REDUCTION);
    }
    else {
      clock_StartCounter(clock_REDUCTION);
      EmptyClauses = list_Nconc(EmptyClauses,
				red_ReductionOnDerivedClauses(Search, Derivables, 
							      flag_GetFlagValue(flag_DOCPROOF), red_WORKEDOFF));
      clock_StopAddPassedTime(clock_REDUCTION);
    }
      

    if (!list_Empty(EmptyClauses)) {
      HighestLevelEmptyClause = split_SmallestSplitLevelClause(EmptyClauses);
      if (flag_GetFlagValue(flag_PEMPTYCLAUSE)) {
	printf("\nEmptyClause: ");
	clause_Print(HighestLevelEmptyClause);
      }
    }
    prfs_DecLoops(Search);
    /*prfs_Print(Search);*/
  }

  prfs_SetEmptyClauses(Search,EmptyClauses);
  prfs_SetUsedEmptyClauses(Search,UsedEmptyClauses);

  return Search;
}

void top_Flotter(int argc, char* argv[], LIST InputClauses)
{
  FILE *Output;
  LIST Scan;
  
  if (argc < opts_Indicator()+2)
    Output = stdout;
  else
    Output = misc_OpenFile(argv[opts_Indicator()+1],"w");

  fprintf(Output, "begin_problem(Unknown).\n\n");
  dfg_FPrintDescription(Output);
  fprintf(Output, "\n\n");  
  fprintf(Output, "list_of_symbols.\n");
  fol_FPrintDFGSignature(Output);
  fprintf(Output, "end_of_list.\n\n");
  fprintf(Output, "list_of_clauses(axioms, cnf).\n");
  
  for (Scan=InputClauses;!list_Empty(Scan);Scan=list_Cdr(Scan))
    if (!clause_GetFlag(list_Car(Scan),CONCLAUSE))
      clause_FPrintDFG(Output,list_Car(Scan),FALSE);

  fprintf(Output, "end_of_list.\n\n");
  fprintf(Output, "list_of_clauses(conjectures, cnf).\n");
  
  for (Scan=InputClauses;!list_Empty(Scan);Scan=list_Cdr(Scan))
    if (clause_GetFlag(list_Car(Scan),CONCLAUSE))
      clause_FPrintDFG(Output,list_Car(Scan),FALSE);
 
  fprintf(Output, "end_of_list.\n\n");
  fprintf(Output, "\nend_problem.\n\n");

  
  printf("\nFLOTTER needed\t"); 
  clock_PrintTime(clock_INPUT);
  printf(" for the input.");
  printf("\n\t\t");
  clock_PrintTime(clock_CNF);
  printf(" for the  CNF translation.");
  

  if (Output != stdout)
    misc_CloseFile(Output,argv[opts_Indicator()+1]);
}

BOOL top_CalledFlotter(char* Call)
{
  int length;
  length = strlen(Call);
  return (strcmp((Call + (length > 7 ? length - 7 : 0)),"FLOTTER") ==0);
}



/**************************************************************/
/* Main Function                                              */
/**************************************************************/

int main(int argc, char* argv[])
{
  LIST              InputClauses,Scan,Axioms,Conjectures, Sorts, QueryClauses, 
                    LabelClauses, QueryPair, ProblemClauses, Labellist, Sortlabellist, 
                    Symblist;
  int               Splits;
  TERM              conformula;
  PROOFSEARCH       Search, FlotterSearch;
  FLAGSTORE         InputStore;
  FILE*             InputStream; 
  HASH              TermLabelToClauselist, ClauseToTermLabellist;
  top_SEARCHRESULT  Result;
  
#if (defined(SPASS_SIGNALS))
  signal(SIGINT, top_SigHandler);
  signal(SIGSEGV, top_SigHandler);
  signal(SIGBUS, top_SigHandler);
#endif

  clock_Init();
  clock_StartCounter(clock_OVERALL);
  memory_Init(-1);
  symbol_Init(TRUE);
  stack_Init();
  hash_Init();
  sort_Init();
  term_Init();
  fol_Init(TRUE);
  eml_Init();
  approx_Init();
  cont_Init();
  unify_Init();
  flag_Init();
  subs_Init();
  clause_Init();
  red_Init();
  dp_Init();
  opts_Init();

  /* Build proof search object to store definitions in */
  Search     = prfs_Create();
  InputStore = flag_NewStore();

  
  /* declare all options */
  opts_DeclareSPASSFlagsAsOptions();
  top_RemoveFileOptId = opts_Declare("rf", opts_NOARGTYPE);

  if (!opts_Read(argc, argv)) 
    return EINVAL;

  if (argc < opts_Indicator()+1 && !flag_GetFlagValue(flag_STDIN)) { /* No input file, no stdin input */
    printf("\n\t          %s %s",argv[0],misc_VERSION);
    if (top_CalledFlotter(argv[0]) || flag_GetFlagValue(flag_FLOTTER))
      printf("\n\t       Usage: FLOTTER [options] [<input-file>] [<output-file>]\n\n");
    else
      printf("\n\t       Usage: SPASS [options] [<input-file>] \n\n");
    printf("Possible options:\n\n");
    opts_PrintSPASSNames(); 
    return EINVAL;
  }
  FlotterSearch = NULL;

  Axioms = Conjectures = Sorts = Labellist = Sortlabellist = list_Nil();
  if (flag_GetFlagValue(flag_STDIN)) {
    top_InputFile = (char*)NULL;
    InputStream   = stdin;
  } else {
    top_InputFile = argv[opts_Indicator()];
    InputStream = misc_OpenFile(top_InputFile, "r");
  }
  
  clock_StartCounter(clock_INPUT);
  flag_CleanStore(InputStore);
  flag_InstallStore(InputStore);
  /* Add now flags from file to InputStore */
  InputClauses = dfg_DFGParser(InputStream, &Axioms, &Conjectures, &Sorts); 

  /* Add/overwrite with command line flags */
  opts_SetFlags(InputStore);
  flag_TransferSetFlags(InputStore,prfs_Store(Search));
  flag_InstallStore(prfs_Store(Search));
  
  cnf_Init(); /* Depends on Strong Skolemization Flag */
  if (flag_GetFlagValue(flag_DOCPROOF))
    prfs_AddDocProofSharingIndex(Search);

  /* DocProof is required for interactive mode */
  if (flag_GetFlagValue(flag_INTERACTIVE)) {
    flag_SetFlagValue(flag_FORMULAE2CLAUSE, 1);
    flag_SetFlagValue(flag_DOCPROOF, 1);
  }

  /* If flag_FORMULAE2CLAUSE is enabled, DOCPROOF must be enabled, too */
  if((!flag_GetFlagValue(flag_DOCPROOF)) 
     && (flag_GetFlagValue(flag_FORMULAE2CLAUSE))) {
    printf("\n\t          SPASS %s",misc_VERSION);
    fputs("Mapping between formulae and clauses is only possible with DOCPROOF", stderr);
    return EINVAL;
  }

  /* flag_FORMULAE2CLAUSE is always enabled if DOCPROOF is enabled !? */
  /* Only for queries flag_FORMULAE2CLAUSE is disabled */
  if(flag_GetFlagValue(flag_DOCPROOF))
    flag_SetFlagValue(flag_FORMULAE2CLAUSE, 1);

  /* Create necessary hasharrays */
  if(flag_GetFlagValue(flag_FORMULAE2CLAUSE)) {
    TermLabelToClauselist = hsh_Create();
    ClauseToTermLabellist = hsh_Create();
  }
  else {
    TermLabelToClauselist = NULL;
    ClauseToTermLabellist = NULL;
  }
  
  /* Build conjecture formula and negate it */
  dfg_StripLabelsFromList(Conjectures); 
  if (!list_Empty(Conjectures)) {
    if (list_Empty(list_Cdr(Conjectures))) 
      conformula  = term_Create(fol_Not(), Conjectures); 
    else 
      conformula  = term_Create(fol_Not(), list_List(term_Create(fol_And(),
								 Conjectures))); 
    /* No label at this point */
    Conjectures = list_List(list_PairCreate(NULL, conformula));
  }

  clock_StopPassedTime(clock_INPUT);

  if (top_InputFile) {
    misc_CloseFile(InputStream,top_InputFile);
    if (opts_IsSet(top_RemoveFileOptId))
      remove(top_InputFile);
  }
  opts_Free();

  clock_StartCounter(clock_CNF);

  if (list_Empty(InputClauses)) {
    NAT Termcount;

    Termcount  = 0;
    
    /* Create labels for formulae without them */
    for(Scan = Axioms; !list_Empty(Scan); Scan = list_Cdr(Scan), Termcount++) {
      if (list_PairFirst(list_Car(Scan)) == NULL) {
	char L[100];
	char* Label;
	sprintf(L, "Term%d", Termcount);
	Label = memory_Malloc(sizeof(char) * (strlen(L) + 1));
	sprintf(Label, "%s", L);
	list_Rplaca(list_Car(Scan), Label);
	if(flag_GetFlagValue(flag_DOCPROOF)) {
	  printf("\nAdded label %s for term \n   ", Label);
	  term_Print((TERM) list_PairSecond(list_Car(Scan)));
	}
      }
    }
    for(Scan = Sorts; !list_Empty(Scan); Scan = list_Cdr(Scan), Termcount++) {
      char L[100];
      char* Label;
      sprintf(L, "_SORT_%d", Termcount);
      Label = memory_Malloc(sizeof(char) * (strlen(L) + 1));
      sprintf(Label, "%s", L);
      list_Rplaca(list_Car(Scan), Label);
      Sortlabellist = list_Cons(Label, Sortlabellist);
    }
    Axioms = list_Nconc(Axioms, Sorts);

    if (flag_GetFlagValue(flag_EML)) {
      clock_StartCounter(clock_TRANSL);
      eml_TranlateAndConvertToCnf(&Axioms,&Conjectures,&InputClauses, &Labellist,
				TermLabelToClauselist, ClauseToTermLabellist,
				flag_GetFlagValue(flag_INTERACTIVE),&Symblist);
      clock_StopPassedTime(clock_TRANSL);
    }

    if(flag_GetFlagValue(flag_APPLYDEFS)) {
      BOOL complete;
      def_ExtractDefsFromTermlist(Search, Axioms, Conjectures);
      def_FlattenDefinitionsDestructive(Search);
      Scan=prfs_Definitions(Search);
      while (!list_Empty(Scan)) {
	Conjectures = def_ApplyDefToTermlist((DEF) list_Car(Scan), Conjectures, &complete,
			       TRUE); /* Destructive */
	if(complete == TRUE) {
	  /* Every occurrence could be expanded */
	  /* Remove defintion and expand it in all other definitions */
	  def_FlattenWithOneDefinitionDestructive(Search, (DEF) list_Car(Scan));
	  Scan=prfs_Definitions(Search);
	}
	else 
	  Scan = list_Cdr(Scan);
      }
    }

    /* We must keep the list of symbols because they can't be freed in
       cnf_Flotter */
    Symblist = list_Nil();

    /* Axioms is list of pairs, conjectures is list of terms */
    /* A ProofSearch object is only returned and the symbols kept in Symblist
       if flag_INTERACTIVE is set */
    FlotterSearch = cnf_Flotter(Axioms,Conjectures,&InputClauses, &Labellist,
				TermLabelToClauselist, ClauseToTermLabellist,
				flag_GetFlagValue(flag_INTERACTIVE),&Symblist);

    InputClauses   = clause_ListSortWeighed(InputClauses);
    clause_SetCounter(1);
    for(Scan=InputClauses;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      clause_NewNumber(list_Car(Scan));
    }
  }
  else {
    dfg_DeleteFormulaPairList(Axioms);
    dfg_DeleteFormulaPairList(Sorts);
    dfg_DeleteFormulaPairList(Conjectures);
    if(flag_GetFlagValue(flag_APPLYDEFS)) {
      /* Extract list of definitions */
      def_ExtractDefsFromClauselist(Search, InputClauses);
      def_FlattenDefinitionsDestructive(Search);
      for(Scan=prfs_Definitions(Search); !list_Empty(Scan); Scan=list_Cdr(Scan))
	InputClauses = def_ApplyDefToClauselist(Search, (DEF) list_Car(Scan), InputClauses, TRUE);
    }
  }

  clock_StopPassedTime(clock_CNF);

  if (top_CalledFlotter(argv[0]) || flag_GetFlagValue(flag_FLOTTER)) {
    top_Flotter(argc,argv,InputClauses);
    flag_SetFlagValue(flag_TIMELIMIT,0);         /* Exit No Output*/
    flag_SetFlagValue(flag_INTERACTIVE,0);
    flag_SetFlagValue(flag_PPROBLEM,0);
  }

  prfs_SetPrecedence(Search,symbol_CreatePrecedence());
  memory_Restrict(flag_GetFlagValue(flag_MEMORY));

  do {
    LIST deflist;
    ProblemClauses = list_Nil();
    LabelClauses   = list_Nil();
    Result         = top_RESOURCE;

    if (flag_GetFlagValue(flag_INTERACTIVE)) {
      QueryPair      = ia_GetNextRequest(InputStream); /* Pair (<formula,labellist>) */
      /* Get list of clauses derivable from formulae with labels in labellist */
      if (list_Empty(QueryPair))
	break;
      for(Scan=list_PairSecond(QueryPair);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	LIST l;
	l = list_PointerDeleteDuplicates(list_Copy(hsh_GetWithCompareFunc(TermLabelToClauselist, 
									  list_Car(Scan),
									  cnf_LabelEqual,
									  hsh_StringHashKey)));
	LabelClauses = list_Nconc(l,LabelClauses);
      }
      /* Get list of clauses derivable from sorts */
      for(Scan=Sortlabellist; !list_Empty(Scan); Scan=list_Cdr(Scan)) {
	LIST l;
	l = list_PointerDeleteDuplicates(list_Copy(hsh_GetWithCompareFunc(TermLabelToClauselist, 
									  list_Car(Scan),
									  cnf_LabelEqual,
									  hsh_StringHashKey)));
	LabelClauses = list_Nconc(l,LabelClauses);
      }	

      /* For labelclauses copies are introduced */
      /* So an update to the clause to term mapping is necessary */
      for(Scan=LabelClauses; !list_Empty(Scan); Scan=list_Cdr(Scan)) {
	CLAUSE C;
	LIST l;
	C = (CLAUSE) list_Car(Scan);
	l = list_Copy(hsh_Get(ClauseToTermLabellist, C));
	l = list_DeleteDuplicates(l, cnf_LabelEqual);
	list_Rplaca(Scan, clause_Copy(C));
	hsh_PutList(ClauseToTermLabellist, list_Car(Scan), l);
      }
      QueryClauses   = cnf_QueryFlotter(FlotterSearch, 
					list_PairFirst(QueryPair),
					&Symblist);
      ProblemClauses = list_Nconc(QueryClauses,LabelClauses);
      for(Scan=list_PairSecond(QueryPair); !list_Empty(Scan); Scan= list_Cdr(Scan))
	memory_Free((char*) list_Car(Scan), strlen((char*) list_Car(Scan)) + 1);
      list_Delete(list_PairSecond(QueryPair));
      list_PairFree(QueryPair);
      clock_InitCounter(clock_OVERALL);
      clock_StartCounter(clock_OVERALL);
    }
    else {
      ProblemClauses = InputClauses;
      InputClauses   = list_Nil();
    }


    prfs_SetSplitCounter(Search,flag_GetFlagValue(flag_SPLITS));
    prfs_SetLoops(Search,flag_GetFlagValue(flag_LOOPS));
    prfs_SetBacktrackedClauses(Search,0);

    Search = top_ProofSearch(Search,ProblemClauses,InputStore);  

    if ((flag_GetFlagValue(flag_TIMELIMIT) < 0 ||
	 flag_GetFlagValue(flag_TIMELIMIT) > clock_GetSeconds(clock_OVERALL)) &&
	prfs_Loops(Search) != 0) {
      if (list_Empty(prfs_EmptyClauses(Search)))
	Result = top_COMPLETION;
      else
	Result = top_PROOF;
    }
   
    if (flag_GetFlagValue(flag_TIMELIMIT) != 0) {  /* Stop SPASS immediately */
      printf("\nMSPASS %s ", misc_VERSION);
      switch (Result) {
      case top_RESOURCE: 
	if (prfs_Loops(Search) != 0)
	  printf("\nSPASS beiseite: Ran out of time.");
	else
	  printf("\nSPASS beiseite: Maximal number of loops exceeded.");
	break;
      case top_PROOF:
	printf("\nSPASS beiseite: Proof found.");
	break;
      default:        /* Completion */
	printf("\nSPASS beiseite: Completion found.");
      }
      printf("\nProblem: %s ", 
	     (top_InputFile != (char*)NULL ? top_InputFile : "Read from stdin."));
      if (flag_GetFlagValue(flag_PSTATISTIC)) {
	clock_StopPassedTime(clock_OVERALL);
	printf("\nSPASS derived %d clauses, backtracked %d clauses "
	       "and kept %d clauses.",prfs_DerivedClauses(Search),prfs_BacktrackedClauses(Search),
	       prfs_KeptClauses(Search));
	if (flag_GetFlagValue(flag_SPLITS) < 0) {
	  Splits = -1-prfs_SplitCounter(Search);
	} else {
	  Splits = flag_GetFlagValue(flag_SPLITS)-prfs_SplitCounter(Search);
	};
	printf("\nSPASS performed %d splitting inference steps.",Splits);
	printf("\nSPASS allocated %d KBytes.", memory_DemandedBytes()/1024);
	printf("\nSPASS spent\t"); clock_PrintTime(clock_OVERALL);
	printf(" on the problem.");
	printf("\n\t\t"); clock_PrintTime(clock_INPUT);
	printf(" for the input.");
	printf("\n\t\t"); clock_PrintTime(clock_CNF);
	printf(" for the FLOTTER CNF translation");
        if(flag_GetFlagValue(flag_EML)) {
	  printf(", of which");
	  printf("\n\t\t"); clock_PrintTime(clock_TRANSL);
	  printf(" for the translation from %s to FOL", flag_Name(flag_EML));
	}
	printf(".");
	printf("\n\t\t"); clock_PrintTime(clock_INFERENCE);
	printf(" for inferences.");
	printf("\n\t\t"); clock_PrintTime(clock_BACKTRACK);
	printf(" for the backtracking.");
	printf("\n\t\t"); clock_PrintTime(clock_REDUCTION);
	printf(" for the reduction.\n");
      }
      if (Result != top_PROOF && flag_GetFlagValue(flag_FPMODEL)) {
	FILE *Output;
	char name[100];
	if (Result == top_COMPLETION)
	  sprintf(name,"%s.model",(top_InputFile != (char*)NULL ? top_InputFile : "SPASS"));
	else
	  sprintf(name,"%s.clauses",(top_InputFile != (char*)NULL ? top_InputFile : "SPASS"));
	Output = misc_OpenFile(name,"w");
	if (Result == top_COMPLETION)
	  fprintf(Output, "begin_problem(Completion_by_SPASS).\n\n");
	else
	  fprintf(Output, "begin_problem(Clauses_generated_by_SPASS).\n\n");
	dfg_FPrintDescription(Output);
	fprintf(Output, "list_of_symbols.\n");
	fol_FPrintDFGSignature(Output);
	fprintf(Output, "end_of_list.\n\n");
	clause_FPrintCnfDFG(Output, (flag_GetFlagValue(flag_FPMODEL) == 2), 
			    list_Nil(), prfs_WorkedOffClauses(Search));
	clause_FPrintCnfDFG(Output, (flag_GetFlagValue(flag_FPMODEL) == 2), 
			    list_Nil(), prfs_UsableClauses(Search));
	fprintf(Output, "list_of_settings(SPASS).\n{*\n");
	fol_FPrintPrecedence(Output);
	fprintf(Output, "\n*}\nend_of_list.\n\n");
	fprintf(Output, "end_problem.\n\n");
	misc_CloseFile(Output,name);
	if (Result == top_COMPLETION)
	  printf("\nCompletion printed to: %s\n",name);
	else
	  printf("\nClauses printed to: %s\n",name);
      }
    
      if (flag_GetFlagValue(flag_DOCPROOF) && Result != top_RESOURCE) {
	if (Result == top_COMPLETION) {
	  printf("\n\n The saturated set of worked-off clauses is :\n");
	  clause_ListPrint(prfs_WorkedOffClauses(Search));
	}
	else {
	  LIST UsedClauses, UsedTerms;
	  if (!top_InputFile)
	    top_InputFile = "SPASS";
	  UsedClauses = dp_PrintProof(Search, prfs_EmptyClauses(Search), top_InputFile);
	  /* Find terms required for proof */
	  UsedTerms = list_Nil();

	  for(Scan=UsedClauses; !list_Empty(Scan); Scan = list_Cdr(Scan))
	    if (clause_IsFromInput((CLAUSE) list_Car(Scan))) {
	      LIST L;
	      L = hsh_Get(ClauseToTermLabellist, list_Car(Scan));
	      L = list_Copy(L);
	      L = list_DeleteDuplicates(L, cnf_LabelEqual);
	      UsedTerms = list_Nconc(UsedTerms, L);
	    }
	  list_Delete(UsedClauses);
	  UsedTerms = list_DeleteDuplicates(UsedTerms, cnf_LabelEqual);
	  fputs("\nFormulae used in the proof :", stdout);
	  for(Scan = UsedTerms; !list_Empty(Scan); Scan = list_Cdr(Scan)) 
	    if(!(strncmp((char*) list_Car(Scan), "_SORT_", 6) == 0))
	      printf(" %s", (char*) list_Car(Scan));
	  fputs("\n",stdout);
	  list_Delete(UsedTerms);
	}
      }
    }
    
    /* Delete mapping for the clause copies of the labelclauses */
    for(Scan=LabelClauses; !list_Empty(Scan); Scan=list_Cdr(Scan))
      hsh_DelItem(ClauseToTermLabellist, list_Car(Scan));
    
    list_Delete(ProblemClauses);
 
    fflush(stdout);

    /* Keep definitions */
    deflist = prfs_Definitions(Search);
    prfs_SetDefinitions(Search, list_Nil());
    prfs_Clean(Search);
    prfs_SetDefinitions(Search, deflist);

    symbol_InstallDefaultPrecedence();
    if (flag_GetFlagValue(flag_PPROBLEM))
      printf("\n--------------------------SPASS-STOP------------------------------");
  } while (flag_GetFlagValue(flag_INTERACTIVE) &&
	   (flag_GetFlagValue(flag_TIMELIMIT) != 0));

  for(Scan=InputClauses; !list_Empty(Scan); Scan=list_Cdr(Scan))
    clause_OrientAndReInit((CLAUSE) list_Car(Scan));

  /* Cleanup Flotter data structures */
  if(flag_GetFlagValue(flag_INTERACTIVE)) {
    if (flag_GetFlagValue(flag_DOCPROOF))
      list_Delete(Symblist);
    else 
      list_DeleteWithElement(Symblist,symbol_Delete); 
    /*  symbol_ResetSkolemIndex(); */
    if (FlotterSearch != NULL) 
      prfs_Delete(FlotterSearch);
  }
  if (flag_GetFlagValue(flag_PFLAGS)) {
    printf("\n");
    flag_Print();
  }
  if(flag_GetFlagValue(flag_FORMULAE2CLAUSE)) {
    hsh_Delete(TermLabelToClauselist);
    hsh_Delete(ClauseToTermLabellist);
  }
  for(Scan = Labellist; !list_Empty(Scan); Scan = list_Cdr(Scan))
    memory_Free(list_Car(Scan), strlen((char*)list_Car(Scan)) + 1);
  list_Delete(Labellist);
  list_Delete(Sortlabellist);
  cnf_Free();

  prfs_Delete(Search);
  clause_DeleteClauseList(InputClauses);
  flag_FreeStore(InputStore);

  /*symbol_PrintAll();*/
  sort_Free();
  unify_Free();
  flag_Free();
  cont_Free();
  /*symbol_Dump();*/
  fol_Free();
  symbol_FreeAllSymbols();
  dfg_Free();
  if (flag_GetFlagValue(flag_EML))
    eml_Free();
#ifdef DEBUG
  memory_Print();
#endif
  printf("\n");
  return 0;
}
