/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *             GLOBAL SYSTEM VECTOR                       * */
/* *                                                        * */
/* *  $Module:   VECTOR                                     * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.44 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/07/20 20:56:41 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: vector.h,v $ */


#ifndef _VECTOR_
#define _VECTOR_

#include "misc.h"


/**************************************************************/
/* Type and Variable Definitions                              */
/**************************************************************/

#define vec_SIZE 10000

typedef POINTER VECTOR[vec_SIZE];

extern VECTOR vec_VECTOR;
extern int    vec_MAX;

/**************************************************************/
/* Inline Functions                                           */
/**************************************************************/

/* Stack operations */

static __inline__ void vec_Init(void)
{
  vec_MAX = 0;
}

static __inline__ void vec_Push(POINTER Value)
{
#ifdef DEBUG
  if (vec_MAX >= vec_SIZE) {
    printf("\n Vector Overflow !\n");
    misc_DumpCore();
  }
#endif

  vec_VECTOR[vec_MAX++] = Value;
}


static __inline__ POINTER vec_GetNth(NAT Index)
{
#ifdef DEBUG
  if (Index >= vec_SIZE || Index >= vec_MAX)
    misc_DumpCoreOut("Illegal Vector Access");
#endif

  return vec_VECTOR[Index];
}


static __inline__ void vec_PutNth(NAT Index, POINTER Value)
{
#ifdef DEBUG
  if (Index >= vec_SIZE || Index >= vec_MAX) {
    printf("\n Illegal Vector Access !\n");
    misc_DumpCore();
  }
#endif

  vec_VECTOR[Index] = Value;
}


static __inline__ void vec_Pop(void)
{
  --vec_MAX;
}

static __inline__ POINTER vec_PopResult(void)
{
  return vec_VECTOR[--vec_MAX];
}

static __inline__ void vec_PopToN(int N)
{
  vec_VECTOR[N] = vec_VECTOR[--vec_MAX];
}

static __inline__ void vec_NPop(int N)
{
  vec_MAX -= N;
}

static __inline__ POINTER vec_Top(void)
{
  return vec_VECTOR[vec_MAX-1];
}

static __inline__ POINTER vec_NthTop(int N)
{
  return vec_VECTOR[vec_MAX-(1+N)];
}


static __inline__ void vec_RplacTop(POINTER Value)
{
  vec_VECTOR[vec_MAX-1] = Value;
}

static __inline__ void vec_RplacNthTop(int N, POINTER Value)
{
  vec_VECTOR[vec_MAX-(1+N)] = Value;
}

static __inline__ int vec_ActMax(void)
{
  return vec_MAX;
}

static __inline__ void vec_SetMax(int Index)
{
  vec_MAX = Index;
}

static __inline__ BOOL vec_IsMax(int Index)
{
  return vec_MAX == Index;
}

static __inline__ BOOL vec_IsEmpty(void)
{
  return vec_MAX == 0;
}

/**************************************************************/
/* Prototypes                                                 */
/**************************************************************/

void vec_Swap(int, int);
void vec_PrintSel(int, int, void(*)());
void vec_PrintAll(void(*)());

#endif
